//////////////////////////////////////////////////////////////////////////
/// \file 			MFCCatalogExample.cpp 
/// \brief 			Small example showing catalog use.
/// \date 			2021
/// \copyright 		Metrolab Technology SA
/// \author 		Simon Tellier
// 
//

#include <iostream>
#include <queue>
#include <fstream>		// ifstream to read files
#include <windows.h>	// GetModuleFileName

#include "InstrumentCatalog.h"
#include "MFCManager.h"

using namespace MTL;

class CMyApplication
{
private:
	CGUICatalogInterface					m_Cat;
	CAsyncSnaphsotNotifier					m_ASNotifier;
	std::mutex								m_QLock;
	// This queue is the heart of the interchange mechanism. 
	// The catalog being "interchange-agnostic", it is up to you to 
	// select which type of communication mechanism you want to implement 
	// in your application. In this specific and simple example, we
	// aimed for a queue.
	std::queue< CAsyncNotification *>		m_Q;

	// Instrument to connect to. 
	static std::string m_InstrToConnect;

private:
	void l_PushNewAsyncNotification(CAsyncNotification * pANot)
	{
		std::lock_guard<std::mutex> l_LG(m_QLock);
		m_Q.push(pANot);
	}
	CAsyncNotification * l_GetNotification()
	{
		std::lock_guard<std::mutex> l_LG(m_QLock);
		if (m_Q.empty())
			return nullptr;

		CAsyncNotification * l_pNot = m_Q.front();
		m_Q.pop();
		return l_pNot;
	}

public:
	CMyApplication(CInstrumentCatalog & rCat)
		: m_Cat(rCat), m_ASNotifier(&CMyApplication::l_PushNewAsyncNotification, this)
	{
		//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		// NOTE:
		// You can edit this m_InstrToConnect string according to the instrument you want to connect to. This string must correspond to the UniqueID string of an instrument found in the "New Instrument List".
		// The UniqueID string corresponds to the VISA ressource name. 
		// As example, UniqueID according to different interfaces:
		// - USB connection (as PT2026 instrument), UniqueID looks like: "USB0::0x1BFA::0x07EA::0000001::INSTR" , with 0000001 the 7 digit instrument serial number.
		// - Ethernet connection (as PT2026 instrument), UniqueID looks like: "TCPIP0::pt2026_00001::inst0::INSTR" .
		// - Serial connection (as MFC3045 instrument), UniqueID looks like: "ASRL1::INSTR" , in ASRLx the x corresponds to the serial port COM number.
		m_InstrToConnect = "USB0::0x1BFA::0x07EA::0000001::INSTR";
		// END OF NOTE
		//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
		//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	}

	void SyncControls()
	{
		// We do not need to sync anything with the catalog for this application
	}
	void SyncObservers()
	{
		// We do not need to sync anything with the catalog for this application
	}
	void Subscribe()
	{
		// Proceed to the subscription to each variable of interrest.
		m_Cat.oInstrumentList.Subscribe(m_ASNotifier);
		m_Cat.oProbeArrayInformation.Subscribe(m_ASNotifier);
		m_Cat.oSearchResult.Subscribe(m_ASNotifier);
		m_Cat.oInstrumentStatus.Subscribe(m_ASNotifier);
		m_Cat.oError.Subscribe(m_ASNotifier);
	}
	
	std::string GetInterfaceString(CInstrumentCatalog::eConnectionInterface l_InstrInterface)
	{
		std::string l_InterfaceStr = "Unknown";
		switch (l_InstrInterface)
		{
		case CInstrumentCatalog::eConnectionInterface::kEthernet:
			l_InterfaceStr = "Ethernet";
			break;
		case CInstrumentCatalog::eConnectionInterface::kSerial:
			l_InterfaceStr = "Serial";
			break;
		case CInstrumentCatalog::eConnectionInterface::kUSB:
			l_InterfaceStr = "USB";
			break;

		default:
			break;
		}
		return l_InterfaceStr;
	}

	void DoJob()
	{
		bool l_Operate = true;
		while (l_Operate)
		{
			// Get the current notification.
			CAsyncNotification * l_pNot = l_GetNotification();
			if (nullptr == l_pNot)
			{	
				// No notification currently in the queue. 
				// Sleep for a while.
				std::this_thread::sleep_for(std::chrono::milliseconds(100));
			}
			else
			{
				// Since we have a notification...
				CObserver_untyped & rVar = l_pNot->Observer();
				const CAsyncSnapshotNotification * l_pASNot = static_cast<CAsyncSnapshotNotification *>(l_pNot);

				// Try determining which variable spawned it.

				if (rVar.Is(m_Cat.oInstrumentList)) //New instrument list received
				{
					std::cout << "New Instrument List" << std::endl;					
					const CInstrumentCatalog::tInstrumentList * l_pVal = static_cast<const CInstrumentCatalog::tInstrumentList *>(l_pASNot->DataCopy());
					for (int i = 0; i < l_pVal->size(); i++)
					{
						std::cout << "Instrument num " << i << " : UniqueID = " << l_pVal->at(i).UniqueID << " , Label = " << l_pVal->at(i).Label << " , Inteface = " << GetInterfaceString(l_pVal->at(i).Interface) << std::endl;
					}
					std::cout << std::endl;

					for (CInstrumentCatalog::tInstrumentList::const_iterator it= l_pVal->begin(); it != l_pVal->end(); it++)
					{
						if (m_InstrToConnect == it->UniqueID)
						{
							//Connect to the default instrument if found in the list
							std::cout << "Trying to connect to default Instrument:" << " : UniqueID = " << it->UniqueID << " , Label = " << it->Label << " , Inteface = " << GetInterfaceString(it->Interface) << std::endl;
							m_Cat.cInstrumentCurrent(*it);
						}
					}
				}
				else if (rVar.Is(m_Cat.oProbeArrayInformation))	//New ProbeArray information received
				{
					std::cout << "New Probe-Array Information received" << std::endl;
					const CInstrumentCatalog::sProbeArrayInformation * l_pVal = static_cast<const CInstrumentCatalog::sProbeArrayInformation *>(l_pASNot->DataCopy());
					if (l_pVal->Type != CInstrumentCatalog::eProbeArrayType::kUnknown)
					{
						std::cout << "Probe-Array connected. Start a NMR search operation." << std::endl;
						// Configure instrument and start a NMR Seacrh 
						CInstrumentCatalog::sInstrumentState l_NewState;
						l_NewState.OperatingMode = CInstrumentCatalog::eOperatingMode::kSearch;
						l_NewState.RepeatMode = CInstrumentCatalog::eRepeatMode::kSingle;
						l_NewState.ProbeSelection = CInstrumentCatalog::sMeasureProbeSelection();
						l_NewState.FreqMin = l_pVal->FreqMin;
						l_NewState.FreqMax = l_pVal->FreqMax;
						l_NewState.NbAveragedMeasurements = 0;
						m_Cat.cInstrumentState(l_NewState);
					}
					else
					{
						std::cout << "No Probe-Array connected." << std::endl;
					}
				}
				else if (rVar.Is(m_Cat.oSearchResult)) //New NMR search result received
				{
					const CInstrumentCatalog::sSearchResult * l_pVal = static_cast<const CInstrumentCatalog::sSearchResult *>(l_pASNot->DataCopy());
					std::cout << "New Search Result :" << CInstrumentCatalog::l_Dump(*l_pVal);
					l_Operate = false;
				}
				else if (rVar.Is(m_Cat.oInstrumentStatus)) // New Instrument Status received
				{
					const CInstrumentCatalog::sInstrumentStatus * l_pVal = static_cast<const CInstrumentCatalog::sInstrumentStatus *>(l_pASNot->DataCopy());
					std::cout << CInstrumentCatalog::l_Dump(*l_pVal);
				}
				else if (rVar.Is(m_Cat.oError)) // New Error received
				{
					const CInstrumentCatalog::sError * l_pVal = static_cast<const CInstrumentCatalog::sError *>(l_pASNot->DataCopy());
					std::cout << CInstrumentCatalog::l_Dump(*l_pVal);
					l_Operate = false;
				}

				// Delete notification (and its content)
				delete l_pNot;
			}
		}
	}
};


std::string CMyApplication::m_InstrToConnect;

std::string GetApplicationDir()
{
#ifdef _WIN32
	wchar_t l_ExePath[MAX_PATH];
	GetModuleFileName(NULL, l_ExePath, MAX_PATH);					// Get executable path
	std::wstring l_AppPath(l_ExePath);
	std::string l_AppDir(l_AppPath.begin(), l_AppPath.end());
	return l_AppDir.substr(0, l_AppDir.find_last_of('\\'));			// Remove application name
#endif
}

int main()
{
	std::string l_AppDir = GetApplicationDir();

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// NOTE:
	//
	// These C++ classes are provided by Metrolab and should not be edited.
	CInstrumentCatalog		l_Catalog;
	CMFCManager				l_MFCMan(l_Catalog, l_AppDir + "/../../../../InstrumentCatalog/MFC/ExternalFiles/DefaultCommandScript.scs", l_AppDir + "/../../../../InstrumentCatalog/MFC/ExternalFiles");
	//
	// This C++ class is the custom user application example.
	// You should edit this class for your own application to interact with the C++ instrument catalog
	CMyApplication			l_MyApp(l_Catalog);
	//
	// END OF NOTE
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


	// Synchronize catalog controls
	l_MFCMan.SyncControls();
	l_MyApp.SyncControls();
	// Synchronize catalog observers
	l_MFCMan.SyncObservers();
	l_MyApp.SyncObservers();
	// Subscribe to variables
	l_MFCMan.Subscribe();
	l_MyApp.Subscribe();
	// Operate
	l_MFCMan.Operate();

	l_MyApp.DoJob();

	// Terminate manager operation
	l_MFCMan.Terminate();

	return 0;
}
