///	\file			MFCController.h
///	\date			2020
///	\copyright (c)	Metrolab Technology S.A., Geneva, Switzerland (www.metrolab.com)
///	See the included file License Terms.pdf for the licensing conditions.

#pragma once

// Standard includes
#include <string>
// Personal includes
#include "InstrumentCatalog.h"
#include "VISAInstrument.h"
#include "MFCDevice.h"
#include "Helpers.h"
#include "ThreadedMailbox.h"
#include "MFCManagerCatalog.h"

using namespace MTL;
using namespace MTL::Instrument;
using namespace MTL::ThreadedMailbox;

namespace MTL {

	//--------------------------------------------------------------------------//
	//								MFC CONTROLLER
	//--------------------------------------------------------------------------//
	struct sMFCControllerMailBoxMsg {
		enum eMsgType { kTimerEvent }		MsgType;
		sMFCControllerMailBoxMsg()
			: MsgType(kTimerEvent) {}
		sMFCControllerMailBoxMsg(eMsgType MT)
			: MsgType(MT) {}
	};
	class CMFCController : private CThreadedMailBox<sMFCControllerMailBoxMsg>
	{
	//--------------------------------------------------------------------------//
	//								Friend class
	//--------------------------------------------------------------------------//
	private:
		friend class CMFC2x46Device;
		friend class CMFC3045Device;

	//--------------------------------------------------------------------------//
	//								Attributes
	//--------------------------------------------------------------------------//
	private:
		CMFCControllerInterface						m_ContCat;							// Controller catalog
		CTimer										m_Timer;
		std::atomic<CMFCManagerCatalog::eControllerState>	m_ControllerState;
		CVISAResourceManager *						m_pRsrcManager = nullptr;
		CMFCDevice *								m_pMFCDevice = nullptr;
		CInstrumentCatalog::sInstrumentState		m_InstrState;
		CInstrumentCatalog::sProbeArrayInformation	m_PAInfo;
		std::string									m_DefaultPAConfScriptFilePath;		// Path of the default PA configuration script (if no specific file exists)
		std::string									m_CustomPAConfScriptDirectory;		// Directory containing PA specific configuration scripts
		// Constants
		const int			RSRC_POLLING_PERIOD_MS			= 200;
		const int			PA_POLLING_PERIOD_MS			= 200;

	//--------------------------------------------------------------------------//
	//								Management
	//--------------------------------------------------------------------------//
	private:
		void l_SendLogMessage(const std::string & rMessage, const CInstrumentCatalog::eLogLevel & rLevel, const std::string & rContext, const std::string & rLocation);
		void l_SendErrorMessage(const CMFCDevice::sMFCDeviceError & rErrorMessage);
		void l_SendSearchProgress(const U8 & rSearcHprogress);
		void l_SendSearchResult(const CInstrumentCatalog::sSearchResult & rSearchResults);
		void l_SendMeasResult(const CInstrumentCatalog::sMeasurementResults & rMeasRes);
		bool l_IsAbortRequested();
		void l_CheckConnectionOnWarning(const CMFCDevice::sMFCDeviceError & rOriginError, CMFCDevice::sMFCDeviceError & rErrorOut);

	//--------------------------------------------------------------------------//
	//							Internal Operations
	//--------------------------------------------------------------------------//
	// Lives in the timer thread.
	private:
		void l_PostTimerEvent();										// Called when timer elapses. Adds timer event to queue.
	// Lives in the Controller thread.
	private:
		// Thread task
		void l_ThreadTask(CThreadedMailBox::eEventType EvtType, sMFCControllerMailBoxMsg * pMsg);
		// Mail box event handlers
		void l_TimerEventHandler();
		// Internal operations
		void l_AbortInstrument();
		void l_ScanResources();
		void l_GetRemoteBoxButtonStatus();
		bool l_ScanPA();
		void l_OperateWithInstrument();

	//--------------------------------------------------------------------------//
	//								Utilities
	//--------------------------------------------------------------------------//
	// Lives in the Controller thread.
	private:
		void l_RequestSelfAbort();

	//--------------------------------------------------------------------------//
	//							Constructor / Destructor
	//--------------------------------------------------------------------------//
	// Lives in the caller thread.
	public:
		CMFCController(CMFCManagerCatalog & rCatInstance, const std::string & rDefaultPAConfScriptFilePath, const std::string & rCustomPAConfScriptDir);
		virtual ~CMFCController();

	//--------------------------------------------------------------------------//
	//							Available Operations
	//--------------------------------------------------------------------------//
	public:
		// Lives in the caller thread.
		void Operate();
		void Terminate();	// Blocks while the Controller thread did not stopped working.
		void ScanInstruments(CVISAResourceManager * pRsrcManager);
		void ScanPA(CMFCDevice * pMFCDevice);
		void SetInstrumentState(CMFCDevice * pMFCDevice, CInstrumentCatalog::sInstrumentState InsState);
		void UpdateNormalizationDate();
	}; // class

} // namespace MTL
