///	\file			MFC2x46Device.cpp
///	\date			2020
///	\copyright (c)	Metrolab Technology S.A., Geneva, Switzerland (www.metrolab.com)
///	See the included file License Terms.pdf for the licensing conditions.

// Standard includes
#include <regex>
#include <fstream>		// ifstream to read files
#include <sstream>		// string manipulation

// Personal includes
#include "MFC2x46Device.h"
#include "CPT2026PeripheralROM.h"
#include "Exception.h"

// External tools includes
#include "date.h"

#if defined _WIN32
#define DIR_SEP		'\\'
#else
#define DIR_SEP		'/'
#endif

#define LOG(__msg__, __level__,__location__)		m_rMFCController.l_SendLogMessage(__msg__, __level__, ".MFC2x46Device", __location__)


//------------------------------------------------------------------//
//					Constructors / Destructors
//------------------------------------------------------------------//
CMFC2x46Device::CMFC2x46Device (	CMFCController & rMFCController,
									CVISAResourceManager & rRM, tResourceName Rsrc,
									const std::string & rDefaultPAConfScriptFilePath,
									const std::string & rCustomPAConfScriptDir)
	:
	m_rMFCController(rMFCController),
	m_PT2026(rRM, Rsrc),
	m_DefaultPAConfScriptFilePath(rDefaultPAConfScriptFilePath),
	m_CustomPAConfScriptDirectory(rCustomPAConfScriptDir)
{

}

CMFC2x46Device::~CMFC2x46Device ()
{

}

//--------------------------------------------------------------------------//
//							Utilities
//--------------------------------------------------------------------------//
bool CMFC2x46Device::l_UtcToStrDateISO8601(const std::time_t & rUtcDate, std::string & rStrDate)
{
	try {
		auto l_TimePointUTC = std::chrono::system_clock::from_time_t(rUtcDate);
		rStrDate = date::format("%FT%TZ", date::floor<std::chrono::seconds>(l_TimePointUTC)); // convert date UTC in YYYY-MM-DDTHH:MM:SSZ ISO8601 string format
		if(rStrDate.empty())
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Fail to format time_t UTC date to string", MTL__LOCATION__);
	}
	catch (sMFCDeviceError & rE)
	{
		MTL_Unused(rE)
		rStrDate.clear();
		return false;
	}
	return true;
}

bool CMFC2x46Device::l_ExecuteConfigurationScript (U16 Model, U32 SN, sMFCDeviceError & rError)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	try
	{
		// Build PA specific configuration script path: MODEL_SERIAL.scs
		std::stringstream lssFileName;
		lssFileName << std::setfill('0') << std::setw(4) << Model << "_" << std::setw(8) << SN << ".scs";
		std::string lPAScriptFilePath = m_CustomPAConfScriptDirectory + DIR_SEP + lssFileName.str();

		std::ifstream lScriptFile;
		// Try to open a PA specific script file
		lScriptFile.open(lPAScriptFilePath);
		if (lScriptFile.is_open())
		{
			LOG("Custom command script file loaded: " + lPAScriptFilePath , CInstrumentCatalog::eLogLevel::kInfo, MTL__LOCATION__);
		}
		else
		{
			// There is no PA specific script file, open the default one
			lScriptFile.open(m_DefaultPAConfScriptFilePath);
			if (!lScriptFile.is_open())
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not open Command Script File", MTL__LOCATION__);
			LOG("Default command script file loaded: "+ m_DefaultPAConfScriptFilePath, CInstrumentCatalog::eLogLevel::kInfo, MTL__LOCATION__);
		}

		std::string lFileLine;
		while (std::getline(lScriptFile, lFileLine))
		{
			if (!m_PT2026.WriteCustomCommand(lFileLine))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not write command from script", MTL__LOCATION__);
		}
	}
	catch (sMFCDeviceError & rE)
	{
		LOG("Command script file could be (or not entirely) run. Some instrument parameters may not be set.", CInstrumentCatalog::eLogLevel::kWarning, MTL__LOCATION__);

		rError = rE;
		return false;
	}
	return true;
}

bool CMFC2x46Device::l_ConfigurationChanged (CInstrumentCatalog::sInstrumentState lastcfg, CInstrumentCatalog::sInstrumentState newcfg)
{
	return (lastcfg.OperatingMode != newcfg.OperatingMode ||
			lastcfg.ProbeSelection != newcfg.ProbeSelection ||
			*(reinterpret_cast<U64*>(&lastcfg.FreqMin)) != *(reinterpret_cast<U64*>(&newcfg.FreqMin)) ||		// Compare binary representation of the double values to detect value change
			*(reinterpret_cast<U64*>(&lastcfg.FreqMax)) != *(reinterpret_cast<U64*>(&newcfg.FreqMax)) ||		// Compare binary representation of the double values to detect value change
			lastcfg.NbAveragedMeasurements != newcfg.NbAveragedMeasurements);
}

bool CMFC2x46Device::l_ConfigureForOperation (const CInstrumentCatalog::sInstrumentState & rInstrState, sMFCDeviceError & rError)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	//-----------------------------------------------------
	// According to the current sInstrumentState, set
	//  ProbeSelection
	//  FreqMin
	//  FreqMax
	//  NbAveragedMeasurements

	if (l_ConfigurationChanged(m_LastConfigured, rInstrState))
	{
		LOG("Reconfiguration needed", CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);
		try
		{
			switch (rInstrState.OperatingMode)
			{
			case CInstrumentCatalog::eOperatingMode::kSearch:
				// Set selection mode to restricted probe set in the camera
				if (!m_PT2026.ChannelsSetLimitedListState(true))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not set limited list state", MTL__LOCATION__);
				break;
			default:
				// Set selection mode to entire probe set in the camera
				if (!m_PT2026.ChannelsSetLimitedListState(false))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not set limited list state", MTL__LOCATION__);
				break;
			}

			// Select channel(s)
			switch (rInstrState.ProbeSelection.SelectionMode)
			{
			case CInstrumentCatalog::eMeasureSelectionMode::kProbeArray:
			{
				// Select the camera
				SCPI::tChannelList l_CameraChannelList;
				l_CameraChannelList.push_back(m_PAChannel);
				if (!m_PT2026.ChannelsSelect(l_CameraChannelList))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not select camera", MTL__LOCATION__);
				// In camera measurement, the number of results to fetch are defined by the number of measurement heads
				switch (rInstrState.OperatingMode)
				{
				case CInstrumentCatalog::eOperatingMode::kSearch:
					m_NbOfMeasurementsToFetch = m_NbOfSearchHeads;
					break;
				default:
					m_NbOfMeasurementsToFetch = m_NbOfMeasHeads;
					break;
				}
			}
				break;
			case CInstrumentCatalog::eMeasureSelectionMode::kSingleProbe:
			{
				// Select the single camera head
				// Compute channel of the single head
				SCPI::tChannel l_SingleHeadChannel = m_PAChannel;
				l_SingleHeadChannel.push_back(rInstrState.ProbeSelection.ProbeNumber);
				//  Compute channel list of the single head
				SCPI::tChannelList l_SingleHeadChannelList;
				l_SingleHeadChannelList.push_back(l_SingleHeadChannel);
				//  Select the single head
				if (!m_PT2026.ChannelsSelect(l_SingleHeadChannelList))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not select camera", MTL__LOCATION__);
				// In single probe measurement, there is only one result to fetch
				m_NbOfMeasurementsToFetch = 1;
			}
				break;
			default:
				// Unhandled case
				MTL_Assert(false);
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kFatal, "Unhandled selection mode", MTL__LOCATION__);
				break;
			}

			// Set search limits
			if (!m_PT2026.ParmSearchSetLimits(rInstrState.FreqMin, rInstrState.FreqMax))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not set search limits", MTL__LOCATION__);

			// Set measurement averaging
			sMeasurementAveraging<uParm> l_MeasAvg;
			if (rInstrState.NbAveragedMeasurements <= 1)
			{
				l_MeasAvg.Type = eMeasurementAveragingType::kMeasNone;
				l_MeasAvg.NoPoints = 1;
			}
			else if (rInstrState.NbAveragedMeasurements >= 1500)
			{
				l_MeasAvg.Type = eMeasurementAveragingType::kMeasRepeat;
				l_MeasAvg.NoPoints = 1500;
			}
			else
			{
				l_MeasAvg.Type = eMeasurementAveragingType::kMeasRepeat;
				l_MeasAvg.NoPoints = rInstrState.NbAveragedMeasurements;
			}
			if (!m_PT2026.ParmAveragingMeasurementSet(l_MeasAvg))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not set measurement averaging", MTL__LOCATION__);
		}
		catch (sMFCDeviceError & rE)
		{
			//if (nullptr != m_pPT2026)
			//	std::vector<sError> l_2026ErrorList = m_pPT2026->ErrorList();
			rError = rE;
			return false;
		}
	}
	else {
		LOG("No reconfiguration needed", CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);
	}
	m_LastConfigured = rInstrState;
	return true;
}

// Compute the average of the measurements made during the search operation on the set of selected heads.
void CMFC2x46Device::l_ComputeAveraging (const std::vector<tFlux> & rFluxList, F64 & rReturnedAveraging, F64 & rReturnedStdDev)
{
	// Compute the mean of the measurements.
	U8 l_ValidCount = 0;
	rReturnedAveraging	= 0.0l;
	rReturnedStdDev		= 0.0l;

	try
	{
		// Check content
		if (rFluxList.empty())
			throw false;

		// Compute averaging
		for (std::vector<tFlux>::const_iterator l_it = rFluxList.begin(); l_it != rFluxList.end(); l_it++)
		{
			// Make sure that the measurement taken to compute the mean are valid.
			// An invalid measurement is marked with a mean value being equal to NaN.
			if (!std::isnan(*l_it))
			{
				// Sum up all the NMR frequencies measured.
				rReturnedAveraging += *l_it;
				// Update the number of valid heads.
				l_ValidCount++;
			}
		}
		// If we had some valid heads, compute the average.
		if (l_ValidCount > 0)
			rReturnedAveraging = rReturnedAveraging / l_ValidCount;
		else
			rReturnedAveraging = DPqNaN;

		// We already know how many valid heads are available.
		if (l_ValidCount > 1)
		{
			for (std::vector<tFlux>::const_iterator l_it = rFluxList.begin(); l_it != rFluxList.end(); l_it++)
			{
				// Make sure that the measurement taken to compute the mean are valid.
				// An invalid measurement is marked with a mean value being equal to NaN.
				if (!std::isnan(*l_it))
				{
					F64 l_Value = *l_it - rReturnedAveraging;
					rReturnedStdDev += l_Value * l_Value;
				}
			}
			// Compute the standard deviation in ppm from the computed mean.
			rReturnedStdDev = sqrt(rReturnedStdDev / (l_ValidCount - 1)) / rReturnedAveraging * 1e6;
		}
		else
			rReturnedStdDev = DPqNaN;
	}
	catch (bool & rE)
	{
		MTL_Unused(rE)
		rReturnedAveraging	= DPqNaN;
		rReturnedStdDev		= DPqNaN;
	}
}

bool CMFC2x46Device::AbortInstrument ()
{
	bool l_ReturnSucceeded = m_PT2026.Abort(true);

	// Flush RemoteBoxStart button status bit
	bool l_RemoteBoxPressed(false);
	RemoteBoxButtonStatus(l_RemoteBoxPressed);
	return l_ReturnSucceeded;
}

bool CMFC2x46Device::IsInstrumentConnected()
{
	std::string l_IDN;
	return (m_PT2026.GetIdentification(l_IDN));
}

//--------------------------------------------------------------------------//
//							Available Operations
//--------------------------------------------------------------------------//
// connect to PT2026 main unit and get instrument information
CMFC2x46Device::sMFCDeviceError CMFC2x46Device::Connect (	CInstrumentCatalog::sInstrumentInformation & rInstrInfo)
{
	try
	{
		m_ChanList.clear();
		if (!m_PT2026.Connect(5000))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Cannot connect to requested resource", MTL__LOCATION__);
		if (!m_PT2026.Abort(true))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Cannot abort instrument", MTL__LOCATION__);
		std::string l_IDN;
		if (!m_PT2026.GetIdentification(l_IDN))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Cannot retrieve instrument identification", MTL__LOCATION__);
		// Parse identification: according to 488.2 10.14.3, *IDN? format is Manufacturer,Model,SerialNumber,FirmwareLevel
		std::regex l_reg("([^,]+)+");
		std::sregex_token_iterator l_IdnTokIt(l_IDN.begin(), l_IDN.end(), l_reg);
		std::vector<std::string> l_IdnTokens(l_IdnTokIt, std::sregex_token_iterator());
		if (l_IdnTokens.size() < 4)
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Cannot parse identification", MTL__LOCATION__);

		std::string l_ManDateStr, l_CalDateStr;
		time_t l_ManDate, l_CalDateUTC;
		if (!m_PT2026.ReadInformationDates(l_ManDateStr, l_ManDate, l_CalDateStr, l_CalDateUTC))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Cannot read information dates", MTL__LOCATION__);
		if(!l_UtcToStrDateISO8601(l_CalDateUTC, rInstrInfo.CalibrationDate))	// convert time_t UTC to YYYY-MM-DDTHH:MM:SSZ ISO8601 UTC format
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Cannot parse date", MTL__LOCATION__);

		//rInstrInfo.InstrumentConnection is filled in MFCManager
		rInstrInfo.Type = CInstrumentCatalog::eInstrumentType::kPT2026;
		rInstrInfo.SerialNumber = l_IdnTokens[2];
		if (!l_IDN.empty() && l_IDN[l_IDN.length()-1] == '\n') {
			l_IDN.erase(l_IDN.length()-1);
		}
		rInstrInfo.Description = l_IDN;
	}
	catch (sMFCDeviceError & rE)
	{
		return rE;
	}
	return sMFCDeviceError(sMFCDeviceError::eLevel::kNoError, "Instrument Connection done.", MTL__LOCATION__);
}

// Scan FCA and Probe Array connected to main unit and get information
CMFC2x46Device::sMFCDeviceError CMFC2x46Device::ScanPA (CInstrumentCatalog::sProbeArrayInformation & rPAInfo,
														CInstrumentCatalog::sInstrumentConfiguration & rInstrConfig, bool &rPaInfoChanged)
{
	CInstrumentCatalog::sFCAInformation l_FCAInfo;

	try
	{
		CExtendedChannelList l_NewChanList;
		rPaInfoChanged = false;
		if (!m_PT2026.ChannelsGetList(l_NewChanList))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not retrieve channel list", MTL__LOCATION__);
		// If the channel list changed
		if (l_NewChanList != m_ChanList)
		{
			m_LastConfigured = CInstrumentCatalog::sInstrumentState();	// Clear last known configuration
			m_PAChannel.clear();										// Clear last known camera channel
			m_ChanList = l_NewChanList;									// Save new Channel list
			try
			{
				// Read acquisition period
				sPulse<uParm> l_RFPulse;
				if (!m_PT2026.ParmRFPulseGet(l_RFPulse))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Cannot retrieve RF pulse parms", MTL__LOCATION__);
				rInstrConfig.AcquisitionPeriod = l_RFPulse.Period_s;

				// If there is no available channel
				if (m_ChanList.empty())
					throw false;

				std::vector<sChannelIformation> l_FCAChanInfoVect;
				SCPI::tChannelList l_FCAChannelListVect;
				//check in Channels List to find FCA and PA channels
				for (auto & l_ChanIt : m_ChanList)
				{
					// Check that we have a PA
					if (l_ChanIt.Type == kPwCamera)
					{
						m_PAChannel = l_ChanIt.Channel;
						break; //break at first PA channel
					}
					//Check that we have a FCA
					else if (kPwFCA == l_ChanIt.Type || kPwMultiplexer == l_ChanIt.Type) //todo: for next PT2026 firmware ARM version remove kPwMultiplexer == l_ChanIt.Type
					{//todo: For next PT2026 Firmware ARM version, modify FCA type to kPwFCA. An FCA must not anymore be detected as kPwMultiplexer. The PT2026TypeConversions.cpp must be modified as well.
						l_FCAChannelListVect.clear(); //clear history to keep only the last peripheral FCA
						l_FCAChannelListVect.push_back(l_ChanIt.Channel); // fill with FCA channel list
					}
				}

				// get FCA information if exists
				if(!l_FCAChannelListVect.empty())
				{
					if(!m_PT2026.ChannelsGetInfo(l_FCAChannelListVect, l_FCAChanInfoVect))
					{
						l_FCAChanInfoVect.clear();
					}
					else
					{
						if(!l_FCAChanInfoVect.empty())
						{
							l_FCAInfo.Description = l_FCAChanInfoVect.at(0).Designation;
							l_FCAInfo.SerialNumber = std::to_string(l_FCAChanInfoVect.at(0).SN);
							l_FCAInfo.FCAFwVersion = l_FCAChanInfoVect.at(0).FirmwareVersion;
						}
					}
				}

				// If we don't have a camera
				if (m_PAChannel.empty())
					throw false;

				// Select the camera
				SCPI::tChannelList l_CameraChannelList;
				l_CameraChannelList.push_back(m_PAChannel);
				if (!m_PT2026.ChannelsSelect(l_CameraChannelList))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not select camera", MTL__LOCATION__);

				// Set units in kMHz
				if (!m_PT2026.ParmUnitsSet(UNITS))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not set units", MTL__LOCATION__);

				// Retrieve frequency limits
				sSearch<sBoundedParm> l_SearchParms;
				if (!m_PT2026.ParmSearchGet(l_SearchParms))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not retrieve search parms", MTL__LOCATION__);

				// Retrieve PA information
				// Read ROM
				std::vector<U8> l_ROMContent;
				if (!m_PT2026.DeviceROMRead(m_PAChannel, l_ROMContent))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not retrieve PA ROM content", MTL__LOCATION__);
				// Parse ROM
				CPT2026PeripheralROM::Parser l_ROM(l_ROMContent.begin(), l_ROMContent.end());
				if (!l_ROM.Parse())
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not parse PA ROM content", MTL__LOCATION__);
				// Get Identification
				CPT2026PeripheralROM::sIdentification l_Identification;
				if(!l_ROM.HasIdentification(l_Identification))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not parse PA ROM content", MTL__LOCATION__);
				// Get PW Camera Physical Information
				U16 l_ParmAddr;
				if (!l_ROM.ParmAddress(static_cast<U8>(CPT2026PeripheralROM::e9046ROMParm::kPwCamPhysicalInformation), l_ParmAddr))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not parse PA ROM content", MTL__LOCATION__);
				CPT2026PeripheralROM::sPwCamPhysicalInfo l_PwCamPhysInfo;
				if (!l_ROM.ParsePwCamPhysInfo(l_ParmAddr, l_PwCamPhysInfo))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not parse PA ROM content", MTL__LOCATION__);
				// Get Search Head List
				if (!l_ROM.ParmAddress(static_cast<U8>(CPT2026PeripheralROM::e9046ROMParm::kSearchHeads), l_ParmAddr))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not parse PA ROM content", MTL__LOCATION__);
				CPT2026PeripheralROM::tHeadList l_SearchHeadList;
				if (!l_ROM.ParseHeadList(l_ParmAddr, l_SearchHeadList))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not parse PA ROM content", MTL__LOCATION__);
				// Get Meas Head List
				if (!l_ROM.ParmAddress(static_cast<U8>(CPT2026PeripheralROM::e9046ROMParm::kMeasHeads), l_ParmAddr))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not parse PA ROM content", MTL__LOCATION__);
				CPT2026PeripheralROM::tHeadList l_MeasHeadList;
				if (!l_ROM.ParseHeadList(l_ParmAddr, l_MeasHeadList))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not parse PA ROM content", MTL__LOCATION__);
				// Get Multi Head Information
				if (!l_ROM.ParmAddress(static_cast<U8>(CPT2026PeripheralROM::e9046ROMParm::kMultiHeadInformation), l_ParmAddr))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not parse PA ROM content", MTL__LOCATION__);
				CPT2026PeripheralROM::sMultiHeadInfo l_MultiHeadInfo;
				if (!l_ROM.ParseMultiHeadInfo(l_ParmAddr, l_MultiHeadInfo))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not parse PA ROM content", MTL__LOCATION__);
				// Get PW Physical Information Set
				if (!l_ROM.ParmAddress(static_cast<U8>(CPT2026PeripheralROM::e9046ROMParm::kPwPhysicalInformationSet), l_ParmAddr))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not parse PA ROM content", MTL__LOCATION__);
				CPT2026PeripheralROM::sPwPhysicalInformationSet l_PwPhysInfoSet;
				if (!l_ROM.ParsePwPhysInfoSet(l_ParmAddr, l_PwPhysInfoSet))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not parse PA ROM content", MTL__LOCATION__);
				// Get Manufacture Information
				if (!l_ROM.ParmAddress(static_cast<U8>(CPT2026PeripheralROM::e9046ROMParm::kManufactureInformation), l_ParmAddr))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not parse PA ROM content", MTL__LOCATION__);
				CPT2026PeripheralROM::sManufactureInformation l_ManInfo;
				if (!l_ROM.ParseManufactureInformation(l_ParmAddr, l_ManInfo))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not parse PA ROM content", MTL__LOCATION__);

				// Execute configuration script for this PA
				CMFC2x46Device::sMFCDeviceError l_Error;
				if (!l_ExecuteConfigurationScript(l_Identification.Model, l_Identification.SN, l_Error))
					throw l_Error;

				// Prepare new PA Info to be returned
				switch (l_Identification.Model) {
				case 9046:
					rPAInfo.Type = CInstrumentCatalog::eProbeArrayType::kMFC9046;
					break;
				case 9146:
					rPAInfo.Type = CInstrumentCatalog::eProbeArrayType::kMFC9146;
					break;
				default:
					rPAInfo.Type = CInstrumentCatalog::eProbeArrayType::kUnknown;
					break;
				}
				rPAInfo.SerialNumber		= std::to_string(l_Identification.SN);



				// Available frequency range depends on the main unit, the FCA, and the PA, so we retrieve the intersection through the search limits
				rPAInfo.FreqMin			= l_PwCamPhysInfo.LowFreq / 1e6;		// From [Hz] to [MHz]
				rPAInfo.FreqMax			= l_PwCamPhysInfo.HighFreq / 1e6;		// From [Hz] to [MHz]
				rPAInfo.FreqCentral		= l_PwCamPhysInfo.NominalFreq / 1e6;	// From [Hz] to [MHz]
				rPAInfo.GyroRatio		= l_PwCamPhysInfo.GyroRatio / 1e6;		// From [Hz/T] to [MHz/T]
				rPAInfo.Description		= l_ManInfo.Designation;

				m_NbOfSearchHeads = static_cast<U8>(l_SearchHeadList.size());
				m_NbOfMeasHeads = static_cast<U8>(l_MeasHeadList.size());

				U8 l_ProbeNo = 1;
				rPAInfo.ProbeList.clear();
				for (auto & l_it : l_MultiHeadInfo.MultiHeadInfo)
				{
					CPT2026PeripheralROM::sPwPhysicalInformation & l_rPhysInfo = l_PwPhysInfoSet[l_it.PhysInfoSetIndex];
					CInstrumentCatalog::sProbeInformation l_PI;
					l_PI.ProbeNumber	= l_ProbeNo++;
					l_PI.FreqMin		= l_rPhysInfo.MinFreq / 1e6;	// From [Hz] to [MHz]
					l_PI.FreqMax		= l_rPhysInfo.MaxFreq / 1e6;	// From [Hz] to [MHz]
					rPAInfo.ProbeList.push_back(l_PI);
				}

				//update FCA Info (must be done after last throw)
				rPAInfo.FCAInfo = l_FCAInfo;

				//Get MFC9046 probe array Normalization date. If we failed to get it we just send a warning!
				std::match_results<std::string::const_iterator> l_DateMatch;
				// MFC9046 normalization date & time string format ISO8601 UTC is: YYYY-MM-DDTHH:MM:SSZ
				if (std::regex_match(l_PwCamPhysInfo.NormalizationDate, l_DateMatch, std::regex("^([0-9]{4}-[0-9]{2}-[0-9]{2})(T[0-9]{2}:[0-9]{2}:[0-9]{2})(Z?)")))
				{
					rPAInfo.NormalizationDate	= l_DateMatch[1].str() + l_DateMatch[2].str() + "Z";
				}
				else if (std::regex_match(l_PwCamPhysInfo.NormalizationDate, l_DateMatch, std::regex("^([0-9]{4}-[0-9]{2}-[0-9]{2})(.*)"))) //YYYY-MM-DD
				{ // if only normalization date exists (no time). String format date ISO8601 is: YYYY-MM-DD
					rPAInfo.NormalizationDate = l_DateMatch[1].str() + "T00:00:00Z"; // convert "YYYY-MM-DD" to "YYYY-MM-DDTHH:MM:SSZ" UTC ISO8601 format
				}
				else if (std::regex_match(l_PwCamPhysInfo.NormalizationDate, l_DateMatch, std::regex("^([0-9]{2})(-[0-9]{2}-)([0-9]{4})(.*)"))) //DD-MM-YYYY
				{ //if year and day has been inversed
					rPAInfo.NormalizationDate = l_DateMatch[3].str() + l_DateMatch[2].str() + l_DateMatch[1].str() + "T00:00:00Z"; // convert "DD-MM-YYYY" to "YYYY-MM-DDTHH:MM:SSZ" UTC ISO8601 format
				}
				else
				{
					rPAInfo.NormalizationDate ="";
					LOG("Failed to parse Porbe Array normalization date", CInstrumentCatalog::eLogLevel::kWarning, MTL__LOCATION__);
				}

				// Ok to send new PA Info
				rPaInfoChanged = true;
			}
			// There is no error but camera is not available
			catch (bool & rUsableCamera)
			{
				MTL_Unused(rUsableCamera)

				//if FCA has changed
				if((l_FCAInfo.SerialNumber != rPAInfo.SerialNumber) || (l_FCAInfo.Description != rPAInfo.FCAInfo.Description))
					rPaInfoChanged = true;

				// If we previously had a PA
				if (!rPAInfo.ProbeList.empty())
				{
					// Clear PA Info (it also clears FCAInfo)
					rPAInfo = CInstrumentCatalog::sProbeArrayInformation();
					// Clear Camera Channel
					m_PAChannel = SCPI::tChannel();
					m_NbOfSearchHeads = 0;
					m_NbOfMeasHeads = 0;
					m_NbOfMeasurementsToFetch = 0;
					rPaInfoChanged = true;
				}
				//update FCA Info
				rPAInfo.FCAInfo = l_FCAInfo;	// (restore FCA Info if cleared in rPAInfo)
			}
		}
	}
	catch (sMFCDeviceError & rE)
	{
		//if (nullptr != m_pPT2026)
		//	std::vector<sError> l_2026ErrorList = m_pPT2026->ErrorList();

		m_ChanList = CExtendedChannelList(); // clear last channel list
		//if FCAInfo has changed
		if((l_FCAInfo.SerialNumber != rPAInfo.SerialNumber) || (l_FCAInfo.Description != rPAInfo.FCAInfo.Description))
			rPaInfoChanged = true;

		// If we previously had a PA
		if (!rPAInfo.ProbeList.empty())
		{
			// Update PA Info (empty)
			rPAInfo = CInstrumentCatalog::sProbeArrayInformation();
			// Clear Camera Channel
			m_PAChannel = SCPI::tChannel();
			rPaInfoChanged = true;
		}
		//update FCA Info
		rPAInfo.FCAInfo = l_FCAInfo;	// (restore FCA Info if cleared in rPAInfo)

		// If the error is not too bad but instrument is not responding anymore
		std::string l_Identification;
		if (rE.Level > sMFCDeviceError::eLevel::kError
				&& !m_PT2026.GetIdentification(l_Identification))
		{
			// Return an error
			return sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Instrument is not responding", MTL__LOCATION__);
		}
		else
			return rE; // Return received error
	}
	return sMFCDeviceError(sMFCDeviceError::eLevel::kNoError, "Scan PA done.", MTL__LOCATION__);
}

CMFC2x46Device::sMFCDeviceError CMFC2x46Device::Search (const CInstrumentCatalog::sInstrumentState & rInstrState)
{
	sMFCDeviceError l_ReturnError;
	try
	{
		// Check initial conditions
		//  Check that we have a PA
		if (m_PAChannel.empty())
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "No available PA", MTL__LOCATION__);

		sMFCDeviceError l_Error;
		if(!l_ConfigureForOperation(rInstrState, l_Error))
			throw l_Error;

		// Search
		bool l_InitiatedContinuous = (CInstrumentCatalog::eRepeatMode::kContinuous == rInstrState.RepeatMode);
		if (!m_PT2026.Initiate(l_InitiatedContinuous))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not initiate measurements", MTL__LOCATION__);

		uOPERation l_OperCond, l_OperEvent;
		do
		{
			// Read status operation register
			if (!m_PT2026.StatusGet(sStatusRegister(eStatusRegisterSet::kStatusOperation, eStatusRegisterType::kStatusCondition), l_OperCond.RawOPER))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not read status operation register", MTL__LOCATION__);
			if (!m_PT2026.StatusGet(sStatusRegister(eStatusRegisterSet::kStatusOperation, eStatusRegisterType::kStatusEvent), l_OperEvent.RawOPER))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not read status operation register", MTL__LOCATION__);

			// Read STB conntinuously while
			//  - we are not requested to stop
			//  - we don't have our measurement results
			//  - we are still INITiated
			// If system indicates Searching, report search progress.
			for (;
				 !m_rMFCController.l_IsAbortRequested() && l_OperCond.OPERation.INITiated && !l_OperEvent.OPERation.NewMeasAvail;
				 )
			{
				// If we don't have our result yet and we are searching
				if (l_OperCond.OPERation.SWEeping)
				{
					U8 l_SPR;
					F64 l_HallMeas;
					if (!m_PT2026.GetSearchProgress(l_SPR, l_HallMeas))
						throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not read search progress", MTL__LOCATION__);
					m_rMFCController.l_SendSearchProgress(l_SPR);//notify controller for search progress
				}

				std::this_thread::sleep_for(std::chrono::milliseconds(SEARCH_POLLING_PERIOD_MS));

				// Read status operation register
				if (!m_PT2026.StatusGet(sStatusRegister(eStatusRegisterSet::kStatusOperation, eStatusRegisterType::kStatusCondition), l_OperCond.RawOPER))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not read status operation register", MTL__LOCATION__);
				if (!m_PT2026.StatusGet(sStatusRegister(eStatusRegisterSet::kStatusOperation, eStatusRegisterType::kStatusEvent), l_OperEvent.RawOPER))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not read status operation register", MTL__LOCATION__);
			}

			// If we have a result to read, read search result and return it
			if (!m_rMFCController.l_IsAbortRequested())
			{
				// If the search succeeded...
				if (l_OperEvent.OPERation.NewMeasAvail)
				{
					std::vector<tFlux> l_Flux;
					eUnits l_Units;
					if (!m_PT2026.MeasurementsGet(m_NbOfMeasurementsToFetch, l_Flux, l_Units))
						throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not fetch search result", MTL__LOCATION__);

					MTL_Assert(!l_Flux.empty() && l_Units == UNITS);
					if (l_Flux.empty() || l_Units != UNITS)
						throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Invalid search result", MTL__LOCATION__);

					// Compute result
					F64 l_Average, l_StdDev;
					l_ComputeAveraging(l_Flux, l_Average, l_StdDev);

					// Send result
					CInstrumentCatalog::sSearchResult l_SearchResult;
					l_SearchResult.Freq = l_Average;
					l_SearchResult.StdDev = static_cast<F32>(l_StdDev);
					m_rMFCController.l_SendSearchResult(l_SearchResult);
				}
				else
				{//No error but no measurement result available from instrument.
					LOG("No measurement result returned by instrument", CInstrumentCatalog::eLogLevel::kWarning, MTL__LOCATION__);
				}
			}
		} while (!m_rMFCController.l_IsAbortRequested() && l_OperCond.OPERation.INITiated && l_InitiatedContinuous);
		l_ReturnError = sMFCDeviceError(sMFCDeviceError::eLevel::kNoError, "Search operation done.", MTL__LOCATION__);
	}
	catch (sMFCDeviceError & rE)
	{
		//if (nullptr != m_pPT2026)
		//	std::vector<sError> l_2026ErrorList = m_pPT2026->ErrorList();

		// Signal error
		l_ReturnError = rE;
	}
	return l_ReturnError;
}

CMFC2x46Device::sMFCDeviceError CMFC2x46Device::Measure (const CInstrumentCatalog::sInstrumentState & rInstrState)
{
	sMFCDeviceError l_ReturnError;
	try
	{
		// Check initial conditions
		//  Check that we have a PA
		if (m_PAChannel.empty())
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "No available PA", MTL__LOCATION__);

		sMFCDeviceError l_Error;
		if(!l_ConfigureForOperation(rInstrState,l_Error))
			throw l_Error;

		// Measure
		bool l_InitiatedContinuous = (CInstrumentCatalog::eRepeatMode::kContinuous == rInstrState.RepeatMode);
		if (!m_PT2026.Initiate(l_InitiatedContinuous))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not initiate measurements", MTL__LOCATION__);

		uOPERation l_OperCond, l_OperEvent;
		do
		{
			// Read status operation register
			if (!m_PT2026.StatusGet(sStatusRegister(eStatusRegisterSet::kStatusOperation, eStatusRegisterType::kStatusCondition), l_OperCond.RawOPER))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not read status operation register", MTL__LOCATION__);
			if (!m_PT2026.StatusGet(sStatusRegister(eStatusRegisterSet::kStatusOperation, eStatusRegisterType::kStatusEvent), l_OperEvent.RawOPER))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not read status operation register", MTL__LOCATION__);

			// Read STB conntinuously while
			//  - we are not requested to stop
			//  - we don't have our measurement results
			//  - we are still INITiated
			// If system indicates Searching, report search progress.
			for (;
				 !m_rMFCController.l_IsAbortRequested() && l_OperCond.OPERation.INITiated && !l_OperEvent.OPERation.NewMeasAvail;
				 )
			{
				// If we don't have our result yet and we are searching
				if (l_OperCond.OPERation.SWEeping)
				{
					U8 l_SPR;
					F64 l_HallMeas;
					if (!m_PT2026.GetSearchProgress(l_SPR, l_HallMeas))
						throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not read search progress", MTL__LOCATION__);
					m_rMFCController.l_SendSearchProgress(l_SPR);//notify controller for search progress
				}

				std::this_thread::sleep_for(std::chrono::milliseconds(MEAS_POLLING_PERIOD_MS));
				// Read status operation register
				if (!m_PT2026.StatusGet(sStatusRegister(eStatusRegisterSet::kStatusOperation, eStatusRegisterType::kStatusCondition), l_OperCond.RawOPER))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not read status operation register", MTL__LOCATION__);
				if (!m_PT2026.StatusGet(sStatusRegister(eStatusRegisterSet::kStatusOperation, eStatusRegisterType::kStatusEvent), l_OperEvent.RawOPER))
					throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not read status operation register", MTL__LOCATION__);
			}

			// Quit on an abort requested?
			if (!m_rMFCController.l_IsAbortRequested())
			{
				// If we have a result to read, read measurement result and return it
				if(l_OperEvent.OPERation.NewMeasAvail)
				{
					eUnits l_Units;
					sArbitraryMeasurements l_ArbMeas;
					sAllMeasurements l_Measurements;
					l_ArbMeas.Flux = true;
					l_ArbMeas.Deviation = true;
					l_ArbMeas.Uniformity = true;
					l_ArbMeas.Channel = true;
					l_ArbMeas.Timestamp = true;
					l_ArbMeas.Status = true;
					l_ArbMeas.NbValidMeas = true;
					if (!m_PT2026.MeasurementsGet(m_NbOfMeasurementsToFetch, l_ArbMeas, l_Measurements, l_Units))
						throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Could not fetch result", MTL__LOCATION__);

					MTL_Assert(l_Measurements.AdvMeas.Basic.size() == m_NbOfMeasurementsToFetch && l_Units == UNITS);
					if (l_Measurements.AdvMeas.Basic.size() != m_NbOfMeasurementsToFetch || l_Units != UNITS)
						throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Invalid measurement result", MTL__LOCATION__);

					// Prepare result
					CInstrumentCatalog::sMeasurementResults l_MeasResult;
					l_MeasResult.Timestamp = l_Measurements.AdvMeas.Basic.front().Timestamp;
					for (auto & l_MeasIt : l_Measurements.AdvMeas.Basic)
					{
						CInstrumentCatalog::sMeasurement l_Meas;
						l_Meas.ProbeNumber = l_MeasIt.Channel.back();
						l_Meas.Freq = l_MeasIt.Flux_UNITS;
						l_Meas.NbValidMeasurements = l_MeasIt.NbValidMeas;
						l_Meas.StdDev = l_MeasIt.Deviation_ppm;
						l_MeasResult.MeasurementList.push_back(l_Meas);
					}
					//Send measurement result
					m_rMFCController.l_SendMeasResult(l_MeasResult);
				}
				else
				{//No error but no measurement result available from instrument.
					LOG("No measurement result returned by instrument", CInstrumentCatalog::eLogLevel::kWarning, MTL__LOCATION__);
				}
			}
		} while (!m_rMFCController.l_IsAbortRequested() && l_OperCond.OPERation.INITiated && l_InitiatedContinuous);
		l_ReturnError = sMFCDeviceError(sMFCDeviceError::eLevel::kNoError, "Measure operation done.", MTL__LOCATION__);
	}
	catch (sMFCDeviceError & rE)
	{
		//if (nullptr != m_pPT2026)
		//	std::vector<sError> l_2026ErrorList = m_pPT2026->ErrorList();

		l_ReturnError = rE;
	}
	return l_ReturnError;
}

CMFC2x46Device::sMFCDeviceError CMFC2x46Device::NewNormalizationRequest (	const CInstrumentCatalog::sNormalization & rNewNormalizationRequest,
																			CInstrumentCatalog::sNormalization & rReturnedNormalization)
{
	try
	{
		rReturnedNormalization.Access = rNewNormalizationRequest.Access;
		rReturnedNormalization.Table.clear();

		switch (rNewNormalizationRequest.Access)
		{
		case CInstrumentCatalog::eNormalizationAccess::kReadRAM:
		{
			// Read Normalization
			tNormalizationTable l_MFC2x46Normalization;
			if (!m_PT2026.NormalizationGet(m_PAChannel, l_MFC2x46Normalization))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Cannot retrieve normalization table", MTL__LOCATION__);
			// Convert units and fill Normalization table
			for (auto&& Factor_Hz : l_MFC2x46Normalization)
				rReturnedNormalization.Table.push_back(Factor_Hz * 1e-6f);	// From [Hz] to [MHz]
			break;
		}
		case CInstrumentCatalog::eNormalizationAccess::kReadROM:
		{
			// Read Normalization
			// Read ROM
			std::vector<U8> l_ROMContent;
			if (!m_PT2026.DeviceROMRead(m_PAChannel, l_ROMContent))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Cannot retrieve device ROM", MTL__LOCATION__);
			// Parse ROM
			CPT2026PeripheralROM::Parser l_ROM(l_ROMContent.begin(), l_ROMContent.end());
			if (!l_ROM.Parse())
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Cannot parse device ROM", MTL__LOCATION__);
			U16 l_ParmAddr;
			// Get PW Camera Physical Information
			if (!l_ROM.ParmAddress(static_cast<U8>(CPT2026PeripheralROM::e9046ROMParm::kPwCamPhysicalInformation), l_ParmAddr))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Invalid device ROM", MTL__LOCATION__);
			CPT2026PeripheralROM::sPwCamPhysicalInfo l_PwCamPhysInfo;
			if (!l_ROM.ParsePwCamPhysInfo(l_ParmAddr, l_PwCamPhysInfo))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Invalid device ROM", MTL__LOCATION__);

			// Convert units
			for (auto&& Factor_Hz : l_PwCamPhysInfo.NormalizationTable)
				rReturnedNormalization.Table.push_back(Factor_Hz * 1e-6f);	// From [Hz] to [MHz]
			break;
		}
		case CInstrumentCatalog::eNormalizationAccess::kWriteRAM:
		{
			tNormalizationTable l_PT2026Normalization;
			// Convert units
			for (auto&& Factor_MHz : rNewNormalizationRequest.Table)
				l_PT2026Normalization.push_back(Factor_MHz * 1e6f);	// From [MHz] to [Hz]
			// Write Normalization to temporary memory
			if (!m_PT2026.NormalizationSet(m_PAChannel, l_PT2026Normalization))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Cannot set normalization table in RAM", MTL__LOCATION__);
			break;
		}
		case CInstrumentCatalog::eNormalizationAccess::kWriteROM:
		{
			// Write Normalization
			// Read ROM
			std::vector<U8> l_ROMContent;
			if (!m_PT2026.DeviceROMRead(m_PAChannel, l_ROMContent))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Cannot retrieve device ROM", MTL__LOCATION__);
			// Parse ROM
			CPT2026PeripheralROM::Parser l_ROM(l_ROMContent.begin(), l_ROMContent.end());
			if (!l_ROM.Parse())
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Cannot parse device ROM", MTL__LOCATION__);
			// Get Identification
			CPT2026PeripheralROM::sIdentification l_Identification;
			if (!l_ROM.HasIdentification(l_Identification))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Invalid device ROM", MTL__LOCATION__);
			U16 l_ParmAddr;
			// Get Manufacture information
			if (!l_ROM.ParmAddress(static_cast<U8>(CPT2026PeripheralROM::e9046ROMParm::kManufactureInformation), l_ParmAddr))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Invalid device ROM", MTL__LOCATION__);
			CPT2026PeripheralROM::sManufactureInformation l_ManuInfo;
			if (!l_ROM.ParseManufactureInformation(l_ParmAddr, l_ManuInfo))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Invalid device ROM", MTL__LOCATION__);
			// Get Multi Head Information
			if (!l_ROM.ParmAddress(static_cast<U8>(CPT2026PeripheralROM::e9046ROMParm::kMultiHeadInformation), l_ParmAddr))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Invalid device ROM", MTL__LOCATION__);
			CPT2026PeripheralROM::sMultiHeadInfo l_MultiHeadInfo;
			if (!l_ROM.ParseMultiHeadInfo(l_ParmAddr, l_MultiHeadInfo))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Invalid device ROM", MTL__LOCATION__);
			// Get Search Heads
			if (!l_ROM.ParmAddress(static_cast<U8>(CPT2026PeripheralROM::e9046ROMParm::kSearchHeads), l_ParmAddr))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Invalid device ROM", MTL__LOCATION__);
			CPT2026PeripheralROM::tHeadList l_SearchHeadList;
			if (!l_ROM.ParseHeadList(l_ParmAddr, l_SearchHeadList))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Invalid device ROM", MTL__LOCATION__);
			// Get Measurement Heads
			if (!l_ROM.ParmAddress(static_cast<U8>(CPT2026PeripheralROM::e9046ROMParm::kMeasHeads), l_ParmAddr))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Invalid device ROM", MTL__LOCATION__);
			CPT2026PeripheralROM::tHeadList l_MeasHeadList;
			if (!l_ROM.ParseHeadList(l_ParmAddr, l_MeasHeadList))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Invalid device ROM", MTL__LOCATION__);
			// Get PW Camera Physical Information
			if (!l_ROM.ParmAddress(static_cast<U8>(CPT2026PeripheralROM::e9046ROMParm::kPwCamPhysicalInformation), l_ParmAddr))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Invalid device ROM", MTL__LOCATION__);
			CPT2026PeripheralROM::sPwCamPhysicalInfo l_PwCamPhysInfo;
			if (!l_ROM.ParsePwCamPhysInfo(l_ParmAddr, l_PwCamPhysInfo))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Invalid device ROM", MTL__LOCATION__);
			// Get PW Physical Information Set
			if (!l_ROM.ParmAddress(static_cast<U8>(CPT2026PeripheralROM::e9046ROMParm::kPwPhysicalInformationSet), l_ParmAddr))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Invalid device ROM", MTL__LOCATION__);
			CPT2026PeripheralROM::sPwPhysicalInformationSet l_PwPhysInfoSet;
			if (!l_ROM.ParsePwPhysInfoSet(l_ParmAddr, l_PwPhysInfoSet))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Invalid device ROM", MTL__LOCATION__);

			// Set normalization date
			auto l_TimeNow = std::chrono::system_clock::now(); // get current time at UTC
			std::string l_StrDate;
			l_StrDate= date::format("%FT%TZ", date::floor<std::chrono::seconds>(l_TimeNow)); // Date & Time in UTC ISO8601 format: YYYY-MM-DDTHH:MM:SSZ
			if(l_StrDate.empty())
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError, "Fail to format time & date to string", MTL__LOCATION__);
			l_PwCamPhysInfo.NormalizationDate = l_StrDate;

			// Convert units and fill Normalization table
			l_PwCamPhysInfo.NormalizationTable.clear();
			for (auto&& Factor_MHz : rNewNormalizationRequest.Table)
				l_PwCamPhysInfo.NormalizationTable.push_back(Factor_MHz * 1e6F);	// From [MHz] to [Hz]

			// Write new ROM
			std::vector<U8> l_NewROMContent;
			CPT2026PeripheralROM::Builder::Build9046ROM(l_NewROMContent, l_Identification, l_ManuInfo, l_MultiHeadInfo, l_SearchHeadList, l_MeasHeadList, l_PwCamPhysInfo, l_PwPhysInfoSet);
			if (!m_PT2026.DeviceROMWrite(m_PAChannel, l_NewROMContent))
				throw sMFCDeviceError(sMFCDeviceError::eLevel::kError,"Cannot write device ROM", MTL__LOCATION__);
			break;
		}
		default:
			MTL_Assert(false);
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kError,"Software error. Unhandled case", MTL__LOCATION__);
			break;
		}
	}
	catch (sMFCDeviceError & rE)
	{
		return rE;
	}
	return sMFCDeviceError(sMFCDeviceError::eLevel::kNoError, "New normalization request done.", MTL__LOCATION__);
}

CMFC2x46Device::sMFCDeviceError CMFC2x46Device::GetNormalizationDate(CInstrumentCatalog::tISO8601DateTime & rNormDate)
{
	try
	{
		// Retrieve PA information
		// Read ROM
		std::vector<U8> l_ROMContent;
		if (!m_PT2026.DeviceROMRead(m_PAChannel, l_ROMContent))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not retrieve PA ROM content", MTL__LOCATION__);
		// Parse ROM
		CPT2026PeripheralROM::Parser l_ROM(l_ROMContent.begin(), l_ROMContent.end());
		if (!l_ROM.Parse())
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not parse PA ROM content", MTL__LOCATION__);
		// Get PW Camera Physical Information
		U16 l_ParmAddr;
		if (!l_ROM.ParmAddress(static_cast<U8>(CPT2026PeripheralROM::e9046ROMParm::kPwCamPhysicalInformation), l_ParmAddr))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not parse PA ROM content", MTL__LOCATION__);
		CPT2026PeripheralROM::sPwCamPhysicalInfo l_PwCamPhysInfo;
		if (!l_ROM.ParsePwCamPhysInfo(l_ParmAddr, l_PwCamPhysInfo))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not parse PA ROM content", MTL__LOCATION__);

		std::match_results<std::string::const_iterator> l_DateMatch;
		// PT2026 normalization date & time string format ISO8601 UTC is: YYYY-MM-DDTHH:MM:SSZ
		if (std::regex_match(l_PwCamPhysInfo.NormalizationDate, l_DateMatch, std::regex("^([0-9]{4}-[0-9]{2}-[0-9]{2})(T[0-9]{2}:[0-9]{2}:[0-9]{2})(Z?)")))
			rNormDate	= l_DateMatch[1].str() + l_DateMatch[2].str() + "Z";
		else if (std::regex_match(l_PwCamPhysInfo.NormalizationDate, l_DateMatch, std::regex("^([0-9]{4}-[0-9]{2}-[0-9]{2})(.*)"))) //YYYY-MM-DD
		{ // if only normalization date exists (no time info). String format date ISO8601 is: YYYY-MM-DD
			rNormDate = l_DateMatch[1].str() + "T00:00:00Z"; // convert "YYYY-MM-DD" to "YYYY-MM-DDTHH:MM:SSZ" UTC ISO8601 format
		}
		else if (std::regex_match(l_PwCamPhysInfo.NormalizationDate, l_DateMatch, std::regex("^([0-9]{2})(-[0-9]{2}-)([0-9]{4})(.*)"))) //DD-MM-YYYY
		{ //if year and day has been inversed
			rNormDate = l_DateMatch[3].str() + l_DateMatch[2].str() + l_DateMatch[1].str() + "T00:00:00Z"; // convert "DD-MM-YYYY" to "YYYY-MM-DDTHH:MM:SSZ" UTC ISO8601 format
		}
		else
		{
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Cannot parse normalization date", MTL__LOCATION__);
		}
	}
	catch (sMFCDeviceError & rE)
	{
		return rE;
	}
	return sMFCDeviceError(sMFCDeviceError::eLevel::kNoError, "Get normalization date done.", MTL__LOCATION__);
}

CMFC2x46Device::sMFCDeviceError CMFC2x46Device::SetRBBusyLedState (const CInstrumentCatalog::eRemoteBoxBusyLedState & rRBBusyLedState)
{
	try
	{
		eRemoteBusyLEDmode l_MFC2x46RemoteLedState;
		switch (rRBBusyLedState)
		{
		case CInstrumentCatalog::eRemoteBoxBusyLedState::kRemoteBusyLedOff:
			l_MFC2x46RemoteLedState = eRemoteBusyLEDmode::kOff;
			break;
		case CInstrumentCatalog::eRemoteBoxBusyLedState::kRemoteBusyLedMeasuring:
			l_MFC2x46RemoteLedState = eRemoteBusyLEDmode::kOnWhenMeasuring;
			break;
		case CInstrumentCatalog::eRemoteBoxBusyLedState::kRemoteBusyLedOffEndMeas:
			l_MFC2x46RemoteLedState = eRemoteBusyLEDmode::kClearsAfterMeasurement;
			break;
		case CInstrumentCatalog::eRemoteBoxBusyLedState::kRemoteBusyLedOn:
			l_MFC2x46RemoteLedState = eRemoteBusyLEDmode::kOn;
			break;
		case CInstrumentCatalog::eRemoteBoxBusyLedState::kRemoteBusyLedBlinkSlow:
			l_MFC2x46RemoteLedState = eRemoteBusyLEDmode::kBlinkSlow;
			break;
		case CInstrumentCatalog::eRemoteBoxBusyLedState::kRemoteBusyLedBlinkFast:
			l_MFC2x46RemoteLedState = eRemoteBusyLEDmode::kBlinkFast;
			break;
		default:
			MTL_Assert(false);
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kFatal, "Unhandled busy led mode", MTL__LOCATION__);
			break;
		}
		if (!m_PT2026.ChannelSetRemoteLEDMode(l_MFC2x46RemoteLedState))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not set remote box busy led state", MTL__LOCATION__);
	}
	catch (sMFCDeviceError & rE)
	{
		return rE;
	}
	return sMFCDeviceError(sMFCDeviceError::eLevel::kNoError, "Set remote box busy led state done.", MTL__LOCATION__);
}

CMFC2x46Device::sMFCDeviceError CMFC2x46Device::RemoteBoxButtonStatus (bool & rRemoteStartPressed)
{
	try
	{
		uOPERationBIT12 l_OPERBIT12_Event;
		if (!m_PT2026.StatusGet(sStatusRegister(eStatusRegisterSet::kStatusOperationBit12, eStatusRegisterType::kStatusEvent), l_OPERBIT12_Event.RawOPER12))
			throw sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "Could not read OPER:BIT12 event register", MTL__LOCATION__);
		if (l_OPERBIT12_Event.OPERationBIT12.RemoteBoxStart)
			rRemoteStartPressed = true;
		else
			rRemoteStartPressed = false;
	}
	catch (sMFCDeviceError & rE)
	{
		rRemoteStartPressed = false;
		return rE;
	}
	return sMFCDeviceError(sMFCDeviceError::eLevel::kNoError, "Check RemoteBox Status button done.", MTL__LOCATION__);
}

CMFC2x46Device::sMFCDeviceError CMFC2x46Device::ChangeSerialParmsOnConnectedInstr(const CInstrumentCatalog::sMFC3045SerialPortSettings & rNewSerialSetting)
{
	MTL_Unused(rNewSerialSetting)
	return sMFCDeviceError(sMFCDeviceError::eLevel::kWarning, "No serial COM port on MFC2x46 device.", MTL__LOCATION__);
}
