///	\file			MFCManager.cpp
///	\date			2021
///	\copyright (c)	Metrolab Technology S.A., Geneva, Switzerland (www.metrolab.com)
///	See the included file License Terms.pdf for the licensing conditions.

// Standard includes
#include <random>
#include <chrono>
#include <algorithm>
// Personal includes
#include "MFCManager.h"
#include "Helpers.h"
#include "OSDefines.h"
#include "Exception.h"

// External tools includes
#include "date.h"

#define LOG(__msg__, __level__,__location__)		l_Log(CInstrumentCatalog::sLogMessage(__level__, "MFCManager", __msg__, __location__))

#define NO_SEARCH_SPR				0		// No search is being performed
#define SEARCH_ENDED_SPR			100		// Search progress that is notified when search has ended

typedef CException<CMFCManager>	CMFCManagerException;


//--------------------------------------------------------------------------//
//								MFC MANAGER
//--------------------------------------------------------------------------//

//--------------------------------------------------------------------------//
// Reception of catalog events
void CMFCManager::l_CatalogNotificationHandler(CAsyncNotification * pANot)
{
	CObserverNotifier_untyped & rNotifier = pANot->Notifier();
	CObserver_untyped & rVar = pANot->Observer();

	if (rVar.Is(m_Cat.oInstrumentCurrent))
	{
		// Current instrument requested to change
		if (rNotifier.Is(m_ADNotifier))
		{
			CObserver<CInstrumentCatalog::sInstrumentConnection> & rCurrentInstr = reinterpret_cast<CObserver<CInstrumentCatalog::sInstrumentConnection> &>(rVar);
			l_NewInstrumentCurrent(rCurrentInstr.ReadCurrent());
		}
	}
	else if (rVar.Is(m_Cat.oMFC3045SerialPortSettings))
	{
		// New MFC3045 Serial Port settings
		if (rNotifier.Is(m_ADNotifier))
		{
			CObserver<CInstrumentCatalog::sMFC3045SerialPortSettings> & rMFC3045SerialSett = reinterpret_cast<CObserver<CInstrumentCatalog::sMFC3045SerialPortSettings> &>(rVar);
			l_NewMFC3045SerialSettings(rMFC3045SerialSett.ReadCurrent());
		}
	}
	else if (rVar.Is(m_Cat.oInstrumentState))
	{
		// Current instrument state requested to change
		if (rNotifier.Is(m_ADNotifier))
		{
			CObserver<CInstrumentCatalog::sInstrumentState> & rInstrState = reinterpret_cast<CObserver<CInstrumentCatalog::sInstrumentState> &>(rVar);
			l_NewInstrumentState(rInstrState.ReadCurrent());
		}
	}
	else if (rVar.Is(m_Cat.oRemoteBoxBusyLedState))
	{
		// Current remote box busy led state requested to change
		if (rNotifier.Is(m_ADNotifier))
		{
			CObserver<CInstrumentCatalog::eRemoteBoxBusyLedState> & rRBBusyLedState = reinterpret_cast<CObserver<CInstrumentCatalog::eRemoteBoxBusyLedState> &>(rVar);
			l_NewRBBusyLedState(rRBBusyLedState.ReadCurrent());
		}
	}
	else if (rVar.Is(m_Cat.oNormalizationRequest))
	{
		// New Normalization request received
		if (rNotifier.Is(m_ADNotifier))
		{
			CObserver<CInstrumentCatalog::sNormalization> & rNormalizationRequest = reinterpret_cast<CObserver<CInstrumentCatalog::sNormalization> &>(rVar);
			l_NewNormalizationRequest(rNormalizationRequest.ReadCurrent());
		}
	}

	delete pANot;
}

void CMFCManager::l_ControllerNotificationHandler(CAsyncNotification * pASNot)
{
	CObserver_untyped & rVar = pASNot->Observer();

	const CAsyncSnapshotNotification * l_pASNot = static_cast<CAsyncSnapshotNotification *>(pASNot);

	if (rVar.Is(m_ContCat.oState))
	{
		const CMFCManagerCatalog::eControllerState * l_pVal = static_cast<const CMFCManagerCatalog::eControllerState *>(l_pASNot->DataCopy());
		SlotControllerState(*l_pVal);
	}
	else if (rVar.Is(m_ContCat.oInstrumentList))
	{
		const CParsedResourceList * l_pVal = static_cast<const CParsedResourceList *>(l_pASNot->DataCopy());
		SlotInstrumentList(*l_pVal);
	}
	else if (rVar.Is(m_ContCat.oPAInfo))
	{
		const CInstrumentCatalog::sProbeArrayInformation * l_pVal = static_cast<const CInstrumentCatalog::sProbeArrayInformation *>(l_pASNot->DataCopy());
		SlotPAInfo(*l_pVal);
	}
	else if (rVar.Is(m_ContCat.oInstrumentConfiguration))
	{
		const CInstrumentCatalog::sInstrumentConfiguration * l_pVal = static_cast<const CInstrumentCatalog::sInstrumentConfiguration *>(l_pASNot->DataCopy());
		SlotInstrConfig(*l_pVal);
	}
	else if (rVar.Is(m_ContCat.oSearchProgress))
	{
		const U8 * l_pVal = static_cast<const U8 *>(l_pASNot->DataCopy());
		SlotSearchProgress(*l_pVal);
	}
	else if (rVar.Is(m_ContCat.oSearchResult))
	{
		const CInstrumentCatalog::sSearchResult * l_pVal = static_cast<const CInstrumentCatalog::sSearchResult *>(l_pASNot->DataCopy());
		SlotSearchResult(*l_pVal);
	}
	else if (rVar.Is(m_ContCat.oMeasResults))
	{
		const CInstrumentCatalog::sMeasurementResults * l_pVal = static_cast<const CInstrumentCatalog::sMeasurementResults *>(l_pASNot->DataCopy());
		SlotMeasResult(*l_pVal);
	}
	else if (rVar.Is(m_ContCat.oError))
	{
		const CMFCDevice::sMFCDeviceError * l_pVal = static_cast<const CMFCDevice::sMFCDeviceError *>(l_pASNot->DataCopy());
		SlotControllerError(*l_pVal);
	}
	else if (rVar.Is(m_ContCat.oLogMessage))
	{
		const CInstrumentCatalog::sLogMessage * l_pVal = static_cast<const CInstrumentCatalog::sLogMessage *>(l_pASNot->DataCopy());
		SlotControllerLogMessage(*l_pVal);
	}
	else if (rVar.Is(m_ContCat.oRemoteBoxStart))
	{
		SlotRBStart();
	}
	else if (rVar.Is(m_ContCat.oUSBRemoteBoxStart))
	{
		SlotRBStart();
	}

	delete pASNot;
}

void CMFCManager::l_ClearInstrumentInfo()
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	// Fill empty probe array and instrument information and return it to catalog
	LOG("Disconnect current instrument", CInstrumentCatalog::eLogLevel::kInfo, MTL__LOCATION__);
	m_InstrInfo = CInstrumentCatalog::sInstrumentInformation();
	m_Cat.cProbeArrayInformation.rRequestControl(CInstrumentCatalog::sProbeArrayInformation());
	m_Cat.cInstrumentInformation.rRequestControl(CInstrumentCatalog::sInstrumentInformation());
	m_Cat.cInstrumentConfiguration.rRequestControl(CInstrumentCatalog::sInstrumentConfiguration());
}

void CMFCManager::l_NewInstrumentCurrent(const CInstrumentCatalog::sInstrumentConnection & rNewInstr)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);
	LOG(CInstrumentCatalog::l_Dump(rNewInstr).c_str(), CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	// Abort any current operation...
	m_MFCController.Terminate();

	try
	{
		if(nullptr != m_pMFCDevice)
			l_UpdateRBBusyLedState(CInstrumentCatalog::eRemoteBoxBusyLedState::kRemoteBusyLedOff); //try to turn off remote led on previous connected instrument
		// If we are already connected to an instrument, disconnect from it
		delete m_pMFCDevice;
		m_pMFCDevice = nullptr;

		// If the given connection is empty, we do not connect to any instrument
		if (rNewInstr.isEmpty())
		{
			l_ClearInstrumentInfo();
			// Set controller state to scan instrument
			m_MFCController.ScanInstruments(&m_RsrcManager);
		}
		else
		{
			// In any case of new instrument connection, send an empty Probe Array info
			m_Cat.cProbeArrayInformation.rRequestControl(CInstrumentCatalog::sProbeArrayInformation());
			// Look for the requested Serial Port resource
			auto l_InstrIt = m_RsrcListMap.find(rNewInstr.UniqueID);
			if (l_InstrIt == m_RsrcListMap.end())
				throw CMFCManagerException("Cannot find requested resource among available list", MTL__LOCATION__);
			if (rNewInstr.Interface == CInstrumentCatalog::eConnectionInterface::kSerial)
			{
				// Instrument is MFC3045 (MFC3045 main unit + MFC3048 probe array)
				m_pMFCDevice = new CMFC3045Device(m_MFCController, m_RsrcManager, l_InstrIt->second.ParsedResource.Name, m_MFC3045SerialSett);
			}
			else
			{
				// Instrument is MFC2x46 (PT2026 main unit + FCA7046 + MFC9x46 probe array)
				m_pMFCDevice = new CMFC2x46Device(m_MFCController, m_RsrcManager, l_InstrIt->second.ParsedResource.Name, m_DefaultPAConfScriptFilePath, m_CustomPAConfScriptDirectory);
			}
			// Try to connect to instrument and get its information
			CMFCDevice::sMFCDeviceError l_RetError;
			l_RetError = m_pMFCDevice->Connect(m_InstrInfo);
			if(l_RetError.HasError())
			{
				throw CMFCManagerException(l_RetError.Description, l_RetError.Location);
			}
			m_InstrInfo.InstrumentConnection = rNewInstr;

			// Fill instrument information and return it to catalog
			m_Cat.cInstrumentInformation.rRequestControl(m_InstrInfo);

			// Configure remote box busy led mode state on current instrument
			if(!l_UpdateRBBusyLedState(m_RBBusyLedState))
				throw CMFCManagerException("Fail to set remote led mode state", MTL__LOCATION__);

			// Set controller state to Scan probe array
			m_MFCController.ScanPA(m_pMFCDevice);
		}
	}
	catch (CMFCManagerException & rE)
	{
		l_SendError(CInstrumentCatalog::eErrorLevel::kError, rE.message(), rE.context());

		// Fill empty instrument information and return it to catalog
		l_ClearInstrumentInfo();

		// Delete connected instrument if exists
		if(nullptr != m_pMFCDevice)
		{
			delete m_pMFCDevice;
			m_pMFCDevice = nullptr;
		}

		// Set controller state to scan instrument
		m_MFCController.ScanInstruments(&m_RsrcManager);
	}
}

// This method will set rs232 setting to use with MFC3045. If a MFC3045 is already connected, it allows to change instrument serial setting.
void CMFCManager::l_NewMFC3045SerialSettings(const CInstrumentCatalog::sMFC3045SerialPortSettings & rNewSerialSett)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);
	LOG(CInstrumentCatalog::l_Dump(rNewSerialSett).c_str(), CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	CInstrumentCatalog::sInstrumentConnection l_CurrentInstr;
	m_Cat.oInstrumentCurrent.rReadCurrent(l_CurrentInstr);
	try
	{
		// If current connected instrument is MFC3045 and serial settings is different, change it.
		if ((nullptr != m_pMFCDevice) && (l_CurrentInstr.Interface == CInstrumentCatalog::eConnectionInterface::kSerial) && (rNewSerialSett != m_MFC3045SerialSett))
		{
			// Abort any current operation...
			m_MFCController.Terminate();
			// Change instrument serial setting
			LOG("Serial settings change in progress. Operation on current instrument may take a few seconds.", CInstrumentCatalog::eLogLevel::kInfo, MTL__LOCATION__);
			CMFCDevice::sMFCDeviceError l_RetError;
			l_RetError = m_pMFCDevice->ChangeSerialParmsOnConnectedInstr(rNewSerialSett);
			if(l_RetError.HasError())
			{
				l_SendError(l_RetError);
				if (CMFCDevice::sMFCDeviceError::eLevel::kError >= l_RetError.Level)
				{
					throw CMFCManagerException("Serial settings change failed. Instrument serial settings might be corrupted. Set default instrument settings by pressing remote Start button untill end of led blinking while resetting the instrument.", MTL__LOCATION__);
				}
				else //warning
					throw CMFCManagerException("Serial settings change failed.", MTL__LOCATION__);
			}
			// Reconnect to instrument with new serial port settings
			m_MFC3045SerialSett = rNewSerialSett;
			LOG("Serial settings change completed. Reconnection to instrument in progress.", CInstrumentCatalog::eLogLevel::kInfo, MTL__LOCATION__);
			l_NewInstrumentCurrent(l_CurrentInstr);
		}
	}
	catch (CMFCManagerException & rE)
	{
		l_SendError(CInstrumentCatalog::eErrorLevel::kError, rE.message(), rE.context());
		//Disconnect m_pMFCDevice, send empty PA & instrument information and restart MFCController thread with state kScanningInstruments
		if ((nullptr != m_pMFCDevice) && (l_CurrentInstr.Interface == CInstrumentCatalog::eConnectionInterface::kSerial))
			l_NewInstrumentCurrent(CInstrumentCatalog::sInstrumentConnection());
	}
	m_MFC3045SerialSett = rNewSerialSett; // In any case we store new settings for next serial instrument connection
}

void CMFCManager::l_NewInstrumentState(const CInstrumentCatalog::sInstrumentState & rNewInstrState)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);
	LOG(CInstrumentCatalog::l_Dump(rNewInstrState).c_str(), CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	// Abort any current operation...
	m_MFCController.Terminate();

	// Prepare for new operation...
	m_InstrState = rNewInstrState;
	
	// Clear Search Progress
	l_UpdateSPR(NO_SEARCH_SPR);
	
	if (rNewInstrState.OperatingMode == CInstrumentCatalog::eOperatingMode::kIdle)
	{
		// Set controller state
		if (nullptr != m_pMFCDevice)
			m_MFCController.ScanPA(m_pMFCDevice);
		else
			m_MFCController.ScanInstruments(&m_RsrcManager);
	}
	else
	{
		// Start new operation
		m_MFCController.SetInstrumentState(m_pMFCDevice, m_InstrState);
	}
}

void CMFCManager::l_NewNormalizationRequest(const CInstrumentCatalog::sNormalization & rNewNormalizationRequest)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);
	LOG(CInstrumentCatalog::l_Dump(rNewNormalizationRequest).c_str(), CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	// Abort any current operation...
	m_MFCController.Terminate();

	try
	{
		if (nullptr == m_pMFCDevice)
			throw CMFCManagerException("No device connected", MTL__LOCATION__);

		// Try to connect to instrument and get its information
		CMFCDevice::sMFCDeviceError l_RetError;
		CInstrumentCatalog::sNormalization l_ReturnedNormalization(rNewNormalizationRequest);
		CInstrumentCatalog::tISO8601DateTime l_NewNormDate;
		l_RetError = m_pMFCDevice->NewNormalizationRequest(rNewNormalizationRequest, l_ReturnedNormalization);
		if(l_RetError.HasError())
		{
			throw l_RetError;
		}
		if ((rNewNormalizationRequest.Access == CInstrumentCatalog::eNormalizationAccess::kReadRAM) || (rNewNormalizationRequest.Access == CInstrumentCatalog::eNormalizationAccess::kReadROM))
			m_Cat.cNormalizationReturned.rRequestControl(l_ReturnedNormalization); 	// Return read content

		if ((rNewNormalizationRequest.Access == CInstrumentCatalog::eNormalizationAccess::kWriteROM))
		{
			m_MFCController.UpdateNormalizationDate();
		}
	}
	catch (CMFCDevice::sMFCDeviceError & rE)
	{
		l_SendError(rE);
	}
	catch (CMFCManagerException & rE)
	{
		l_SendError(CInstrumentCatalog::eErrorLevel::kError, rE.message(), rE.context());
	}

	// Set controller state
	if (nullptr != m_pMFCDevice)
		m_MFCController.ScanPA(m_pMFCDevice);
	else
		m_MFCController.ScanInstruments(&m_RsrcManager);
}

// Set the new m_RBBusyLedState to be used for next connection or if already connected, changes remote led mode on instrument
void CMFCManager::l_NewRBBusyLedState(const CInstrumentCatalog::eRemoteBoxBusyLedState & rNewRBBusyLedState)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);
	LOG(CInstrumentCatalog::l_Dump(rNewRBBusyLedState).c_str(), CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	// Abort any current operation...
	m_MFCController.Terminate();
	try
	{
		if(nullptr != m_pMFCDevice)
		{
			if(!l_UpdateRBBusyLedState(rNewRBBusyLedState))
				throw CMFCManagerException("Fail to change remote led mode state", MTL__LOCATION__);
		}
		m_RBBusyLedState = rNewRBBusyLedState; // Update m_RBBusyLedState in any case (connected or not)
	}
	catch (CMFCManagerException & rE)
	{
		l_SendError(CInstrumentCatalog::eErrorLevel::kError, rE.message(), rE.context());
	}
	// Set controller state
	if (nullptr != m_pMFCDevice)
		m_MFCController.ScanPA(m_pMFCDevice);
	else
		m_MFCController.ScanInstruments(&m_RsrcManager);
}

//--------------------------------------------------------------------------//
// private method
// Update the m_RBBusyLedState (remote box led state) on the current instrument
bool CMFCManager::l_UpdateRBBusyLedState(const CInstrumentCatalog::eRemoteBoxBusyLedState & rNewRBBusyLedState)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);
	try
	{
		if (nullptr == m_pMFCDevice)
			throw CMFCManagerException("No device connected", MTL__LOCATION__);

		CMFCDevice::sMFCDeviceError l_RetError;
		// Change remote box busy led state mode on connected instrument
		switch (m_InstrInfo.Type)
		{
		case CInstrumentCatalog::eInstrumentType::kPT2026:
		{
			// check PT2026 ARM Firmware version
			std::string l_Desc = m_InstrInfo.Description;
			size_t l_Found = l_Desc.find("arm"); // Arm version is written as "armX.YY", with X = digit for major version & YY = two digits for minor version
			if(std::string::npos!=l_Found)
			{
				U32 l_ArmVerMaj = std::stoul(l_Desc.substr(l_Found +3, 1));
				U32 l_ArmVerMin = std::stoul(l_Desc.substr(l_Found +5, 2));
				if(l_ArmVerMaj >= 1 && l_ArmVerMin >= 22) // Remote box led control is available in PT2026 with ARM firmware version >= 1.22
				{
					l_RetError = m_pMFCDevice->SetRBBusyLedState(rNewRBBusyLedState);
					if(l_RetError.HasError())
					{
						l_SendError(CInstrumentCatalog::eErrorLevel::kWarning, "Failed to set remote box led state.", MTL__LOCATION__);
					}
				}
				else
				{
					// Send a Warning. ARM FW version doesn't support this command
					l_SendError(CInstrumentCatalog::eErrorLevel::kWarning, "Set remote led state not supported. Update main unit FW version.", MTL__LOCATION__);
				}
			}
			break;
		}
		case CInstrumentCatalog::eInstrumentType::kMFC3045:
		{
			l_RetError = m_pMFCDevice->SetRBBusyLedState(rNewRBBusyLedState);
			if(l_RetError.HasError())
			{
				throw l_RetError;
			}
			break;
		}
		default:
			// send warning:
			l_SendError(CInstrumentCatalog::eErrorLevel::kWarning, "Set remote led state not supported. Unknown Instrument type", MTL__LOCATION__);
			break;
		}
	}
	catch (CMFCManagerException & rE)
	{
		l_SendError(CInstrumentCatalog::eErrorLevel::kError, rE.message(), rE.context());
		return false;
	}
	catch (CMFCDevice::sMFCDeviceError & rE)
	{
		l_SendError(rE);
		return false;
	}
	return true;
}

// Notifications to the catalog
void CMFCManager::l_UpdateInstrumentStatus()
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	bool l_Busy = false;

	// Compute State
	switch (m_MFCControllerState)
	{
	case CMFCManagerCatalog::eControllerState::kIdle:
		// Nothing to say but controller stopped a job
		break;
	case CMFCManagerCatalog::eControllerState::kScanningInstruments:
		l_UpdateSPR(NO_SEARCH_SPR);
		m_InstrStatus.State = CInstrumentCatalog::eInstrumentState::kWaitingForPA;
		break;
	case CMFCManagerCatalog::eControllerState::kScanningPA:
		l_UpdateSPR(NO_SEARCH_SPR);
		if (0 == m_Cat.cProbeArrayInformation.ReadCurrent().NbProbes())
			m_InstrStatus.State = CInstrumentCatalog::eInstrumentState::kWaitingForPA;
		else
			m_InstrStatus.State = CInstrumentCatalog::eInstrumentState::kReady;
		break;
	case CMFCManagerCatalog::eControllerState::kSearching:
		m_InstrStatus.State = CInstrumentCatalog::eInstrumentState::kSearching;
		l_Busy = true;
		break;
	case CMFCManagerCatalog::eControllerState::kMeasuring:
		m_InstrStatus.State = CInstrumentCatalog::eInstrumentState::kMeasuring;
		l_Busy = true;
		break;
	}

	// Update instrument State & SearchProgress
	m_Cat.cInstrumentStatus.rRequestControl(m_InstrStatus);

	// Update USB Remote box busy led state
	m_ContCat.cUSBRemoteBusyState.RequestControl(l_Busy);
}

void CMFCManager::l_SendError(const CMFCDevice::sMFCDeviceError & rMFCDeviceError)
{
	switch(rMFCDeviceError.Level)
	{
	case CMFCDevice::sMFCDeviceError::eLevel::kFatal:
		m_Cat.cError.rRequestControl(CInstrumentCatalog::sError(CInstrumentCatalog::eErrorLevel::kFatal, CInstrumentCatalog::FATAL_ERROR_CODE, rMFCDeviceError.Description, rMFCDeviceError.Location));
		break;
	case CMFCDevice::sMFCDeviceError::eLevel::kError:
		m_Cat.cError.rRequestControl(CInstrumentCatalog::sError(CInstrumentCatalog::eErrorLevel::kError, CInstrumentCatalog::ERROR_CODE, rMFCDeviceError.Description, rMFCDeviceError.Location));
		break;
	case CMFCDevice::sMFCDeviceError::eLevel::kWarning:
		m_Cat.cError.rRequestControl(CInstrumentCatalog::sError(CInstrumentCatalog::eErrorLevel::kWarning, CInstrumentCatalog::WARNING_ERROR_CODE, rMFCDeviceError.Description, rMFCDeviceError.Location));
		break;
	case CMFCDevice::sMFCDeviceError::eLevel::kNoError:
		// No error message to send further to instrumentCatalog.
		break;
	}
}

void CMFCManager::l_SendError(CInstrumentCatalog::eErrorLevel Level, std::string Descr, std::string Source)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	switch(Level)
	{
	case CInstrumentCatalog::eErrorLevel::kFatal:
		m_Cat.cError.rRequestControl(CInstrumentCatalog::sError(Level, CInstrumentCatalog::FATAL_ERROR_CODE, Descr, Source));
		break;
	case CInstrumentCatalog::eErrorLevel::kError:
		m_Cat.cError.rRequestControl(CInstrumentCatalog::sError(Level, CInstrumentCatalog::ERROR_CODE, Descr, Source));
		break;
	case CInstrumentCatalog::eErrorLevel::kWarning:
		m_Cat.cError.rRequestControl(CInstrumentCatalog::sError(Level, CInstrumentCatalog::WARNING_ERROR_CODE, Descr, Source));
		break;
	}
}

void CMFCManager::l_SendError(CInstrumentCatalog::sError Error)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);
	m_Cat.cError.rRequestControl(Error);
}

void CMFCManager::l_Log(const CInstrumentCatalog::sLogMessage & rMessage)
{
	m_Cat.cLogMessage.rRequestControl(rMessage);
}

void CMFCManager::l_UpdateSPR(uint8_t SearchProgress)
{
	m_InstrStatus.SearchProgress = SearchProgress;
}

void CMFCManager::l_ThreadTask(CThreadedMailBox::eEventType EvtType, sMFCManagerMailBoxMsg * pMsg)
{
	switch (EvtType)
	{
	case eEventType::kExit:
		break;
	case eEventType::kUserMsg:
		if (nullptr != pMsg)
		{
			if (nullptr != pMsg->pCatNot)
				l_CatalogNotificationHandler(pMsg->pCatNot);

			if (nullptr != pMsg->pContrNot)
				l_ControllerNotificationHandler(pMsg->pContrNot);
		}
		break;
	}
}

void CMFCManager::PostCatalogNotification(CAsyncNotification * pANot)
{
	CThreadedMailBox::l_SendMsg(sMFCManagerMailBoxMsg(pANot, nullptr));
}

void CMFCManager::PostControllerNotification(CAsyncNotification * pASNot)
{
	CThreadedMailBox::l_SendMsg(sMFCManagerMailBoxMsg(nullptr, pASNot));
}



//--------------------------------------------------------------------------//
//							Reception of external events
//--------------------------------------------------------------------------//
void CMFCManager::SlotControllerThreadIsReady()
{
	m_MFCController.ScanInstruments(&m_RsrcManager);
}

void CMFCManager::SlotControllerState(CMFCManagerCatalog::eControllerState State)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	if (CMFCManagerCatalog::eControllerState::kIdle == State)
	{
		switch (m_MFCControllerState)
		{
		case CMFCManagerCatalog::eControllerState::kIdle:
			break;
		case CMFCManagerCatalog::eControllerState::kScanningInstruments:
			break;
		case CMFCManagerCatalog::eControllerState::kScanningPA:
			break;
		case CMFCManagerCatalog::eControllerState::kSearching:
			m_MFCController.ScanPA(m_pMFCDevice);
			break;
		case CMFCManagerCatalog::eControllerState::kMeasuring:
			m_MFCController.ScanPA(m_pMFCDevice);
			break;
		default:
			break;
		}
	}
	m_MFCControllerState = State;

	// Compute new status
	l_UpdateInstrumentStatus();
}

void CMFCManager::SlotInstrumentList(CParsedResourceList NewResourceList)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	//Parse resource list and create user-friendly names when possible
	// USB0::0x1BFA::0xC7EA::PT2026_00012::INSTR	=> PT2026 00012
	// ASRLx::INSTR									=> INSTR - COMx (x correspond to port COM number) , we can't have any instrument information yet
	m_RsrcListMap = l_PrettyResourceNameList(NewResourceList);

	// Convert list type
	m_PrettyInstrList.clear();
	for (auto & l_it : m_RsrcListMap)
	{
		CInstrumentCatalog::sInstrumentConnection l_InsCon;
		l_InsCon.Label = l_it.second.PrettyName;	// "Pretty name"
		l_InsCon.UniqueID = l_it.first;				// VISA resource name is a unique ID
		switch (l_it.second.ParsedResource.Info.Type())
		{
		case eInterfaceType::TCPIP:
			l_InsCon.Interface = CInstrumentCatalog::eConnectionInterface::kEthernet;
			break;
		case eInterfaceType::USB:
			l_InsCon.Interface = CInstrumentCatalog::eConnectionInterface::kUSB;
			break;
		case eInterfaceType::ASRL:
			l_InsCon.Interface = CInstrumentCatalog::eConnectionInterface::kSerial;
			break;
		default:
			l_InsCon.Interface = CInstrumentCatalog::eConnectionInterface::kUnknown;
			break;
		}
		m_PrettyInstrList.push_back(l_InsCon);
	}
	m_Cat.cInstrumentList.rRequestControl(m_PrettyInstrList);
}

void CMFCManager::SlotPAInfo(CInstrumentCatalog::sProbeArrayInformation PAInfo)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	if(nullptr !=m_pMFCDevice)
		m_Cat.cProbeArrayInformation.rRequestControl(PAInfo);
	else // clear PA info if no main unit anymore
		m_Cat.cProbeArrayInformation.rRequestControl(CInstrumentCatalog::sProbeArrayInformation());
	// Compute new status
	l_UpdateInstrumentStatus();
}

void CMFCManager::SlotInstrConfig(CInstrumentCatalog::sInstrumentConfiguration InstrConfig)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	m_Cat.cInstrumentConfiguration.rRequestControl(InstrConfig);
}

void CMFCManager::SlotProbeArrayAngle(CInstrumentCatalog::sProbeArrayAngle Angle)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	m_Cat.cProbeArrayAngle.rRequestControl(Angle);
}

void CMFCManager::SlotRemoteBoxTriggerEvent()
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	m_Cat.cRemoteBoxTriggerEvent.RequestControl(CInstrumentCatalog::eRemoteBoxTrigger::kRemoteBoxTrigger);
}

void CMFCManager::SlotSearchProgress(U8 Progress)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	l_UpdateSPR(Progress);
	l_UpdateInstrumentStatus();
}

void CMFCManager::SlotSearchResult(CInstrumentCatalog::sSearchResult SearRes)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	l_UpdateSPR(SEARCH_ENDED_SPR);	// Set SPR to 100% when we have a result
	l_UpdateInstrumentStatus();
	m_Cat.cSearchResult.rRequestControl(SearRes);
}

void CMFCManager::SlotMeasResult(CInstrumentCatalog::sMeasurementResults MeasRes)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	l_UpdateSPR(SEARCH_ENDED_SPR);	// Set SPR to 100% when we have a result
	l_UpdateInstrumentStatus();
	m_Cat.cMeasurementResults.rRequestControl(MeasRes);
}

void CMFCManager::SlotControllerError(CMFCDevice::sMFCDeviceError error)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	// Whatever the error, broadcast it to instrument catalog
	l_SendError(error);

	switch (m_MFCControllerState)
	{
	case CMFCManagerCatalog::eControllerState::kIdle:
		break;
	case CMFCManagerCatalog::eControllerState::kScanningInstruments:
		if (error.Level <= CMFCDevice::sMFCDeviceError::eLevel::kFatal)
		{
			m_MFCController.Terminate();
			l_SendError(CInstrumentCatalog::eErrorLevel::kFatal, "Fatal error received from MFCController", MTL__LOCATION__);
		}
		break;
	case CMFCManagerCatalog::eControllerState::kScanningPA:
		switch (error.Level)
		{
		case CMFCDevice::sMFCDeviceError::eLevel::kFatal:
			m_MFCController.Terminate();
			l_ClearInstrumentInfo();
			l_SendError(CInstrumentCatalog::eErrorLevel::kFatal, "Fatal error received from MFCController", MTL__LOCATION__);
			break;
		case CMFCDevice::sMFCDeviceError::eLevel::kError:
			//Disconnect m_pMFCDevice, send empty PA & instrument information and restart MFCController thread with state kScanningInstruments
			l_NewInstrumentCurrent(CInstrumentCatalog::sInstrumentConnection());
			break;
		case CMFCDevice::sMFCDeviceError::eLevel::kWarning:
			// Error has already been reported. Nothing to do.
			break;
		default:
			break;
		}
		break;
	case CMFCManagerCatalog::eControllerState::kSearching:
	case CMFCManagerCatalog::eControllerState::kMeasuring:
		m_MFCController.ScanPA(m_pMFCDevice);
		break;
	default:
		break;
	}
}

void CMFCManager::SlotControllerLogMessage(CInstrumentCatalog::sLogMessage Message)
{
	l_Log(Message);
}

void CMFCManager::SlotRBReset()
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

}
void CMFCManager::SlotRBStart()
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	m_Cat.cRemoteBoxTriggerEvent.RequestControl(CInstrumentCatalog::eRemoteBoxTrigger::kRemoteBoxTrigger);
}



//--------------------------------------------------------------------------//
// Class management
CMFCManager::CMFCManager(CInstrumentCatalog & rCatalog, const std::string & rDefaultPAConfScriptFilePath, const std::string & rCustomPAConfScriptDir)
	:
	  m_Cat(rCatalog),
	  m_ContCat(m_ContCatInstance),
	  m_ADNotifier(&CMFCManager::PostCatalogNotification, this),
	  m_ASControllerNotifier(&CMFCManager::PostControllerNotification, this),
	  m_pMFCDevice(nullptr),
	  m_MFCController(m_ContCatInstance, rDefaultPAConfScriptFilePath, rCustomPAConfScriptDir),
	  m_RBController(m_RsrcManager, m_ContCatInstance),
	  m_DefaultPAConfScriptFilePath(rDefaultPAConfScriptFilePath),
	  m_CustomPAConfScriptDirectory(rCustomPAConfScriptDir)
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	// Initialize variables
	m_InstrStatus = CInstrumentCatalog::sInstrumentStatus(CInstrumentCatalog::eInstrumentState::kWaitingForPA , 0);
}
CMFCManager::~CMFCManager()
{
	CMFCManager::Terminate();
}

void CMFCManager::SyncControls()
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	// Initialize controls (synchronize catalog with instrument manager)
	//m_Cat.cInstrumentList.rRequestControl(m_PrettyInstrList);
	m_Cat.cInstrumentInformation.rRequestControl(CInstrumentCatalog::sInstrumentInformation());
	m_Cat.cProbeArrayInformation.rRequestControl(CInstrumentCatalog::sProbeArrayInformation());
	m_Cat.cProbeArrayAngle.rRequestControl(CInstrumentCatalog::sProbeArrayAngle());
	m_Cat.cInstrumentStatus.rRequestControl(m_InstrStatus);
	//	SearchResults in ephemeral
	//	MeasurementResults in ephemeral
	//	Error in ephemeral
}

void CMFCManager::SyncObservers()
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);
	// Read variables (synchronize catalog with instrument manager)
	m_Cat.oInstrumentCurrent.rReadCurrent(m_Instr);
	m_Cat.oMFC3045SerialPortSettings.rReadCurrent(m_MFC3045SerialSett);
	m_Cat.oInstrumentState.rReadCurrent(m_InstrState);
	m_Cat.oRemoteBoxBusyLedState.rReadCurrent(m_RBBusyLedState);
	//	Instrument calibration date is ephemeral
	//	Normalization request is ephemeral
}

void CMFCManager::Subscribe()
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);
	// Subscribe to InstrumentCatalog observers
	m_Cat.oInstrumentCurrent.Subscribe(m_ADNotifier);
	m_Cat.oMFC3045SerialPortSettings.Subscribe(m_ADNotifier);
	m_Cat.oInstrumentState.Subscribe(m_ADNotifier);
	m_Cat.oNormalizationRequest.Subscribe(m_ADNotifier);
	m_Cat.oRemoteBoxBusyLedState.Subscribe(m_ADNotifier);

	// Subscribe to MFCController observers
	m_ContCat.oState.Subscribe(m_ASControllerNotifier);
	m_ContCat.oInstrumentList.Subscribe(m_ASControllerNotifier);
	m_ContCat.oPAInfo.Subscribe(m_ASControllerNotifier);
	m_ContCat.oInstrumentConfiguration.Subscribe(m_ASControllerNotifier);
	m_ContCat.oSearchProgress.Subscribe(m_ASControllerNotifier);
	m_ContCat.oSearchResult.Subscribe(m_ASControllerNotifier);
	m_ContCat.oMeasResults.Subscribe(m_ASControllerNotifier);
	m_ContCat.oError.Subscribe(m_ASControllerNotifier);
	m_ContCat.oLogMessage.Subscribe(m_ASControllerNotifier);
	m_ContCat.oRemoteBoxStart.Subscribe(m_ASControllerNotifier);

	// Subscribe to USBRemoteBoxController observer
	m_ContCat.oUSBRemoteBoxStart.Subscribe(m_ASControllerNotifier);
}

void CMFCManager::Operate()
{
	LOG(MTL__FUNCTION_NAME__, CInstrumentCatalog::eLogLevel::kDebug, MTL__LOCATION__);

	// Initialize Resource Manager
	m_RsrcManager.Initialize();

	// Start controller thread
	m_MFCController.Operate();

	// Start USB remote box controller thread
	m_RBController.Operate();

	// Start Catalog notification thread
	CThreadedMailBox::ThreadStart();

	// Run first controller operation
	m_MFCController.ScanInstruments(&m_RsrcManager);
}

void CMFCManager::Terminate()
{
	// Stop Catalog notification thread
	CThreadedMailBox::ThreadStop();
	// Stop USB Remote Box controller
	m_RBController.Terminate();
	// Stop MFC controller
	m_MFCController.Terminate();
	// Close MFC instrument (=> PT2026 or MFC3045) connection
	delete m_pMFCDevice;
	m_pMFCDevice = nullptr;
}

CMFCManager::tResourceMap CMFCManager::l_PrettyResourceNameList(const CParsedResourceList & rRsrcList)
{
	// Expected resource formats
	//	"USB0::0x1BFA::0x07EA::0000015::INSTR",
	//	"TCPIP0::pt2026_00015::inst0::INSTR",
	//	"TCPIP0::192.168.110.239::inst0::INSTR",
	//	"ASRL12::INSTR"

	CMFCManager::tResourceMap l_ReturnedList;

	std::regex l_reg("([^:]+)+");
	std::sregex_token_iterator l_end;
	for (auto & l_RsrcIt : rRsrcList)
	{
		// Tokenize resource name
		std::sregex_token_iterator l_TokIt(l_RsrcIt.Name.begin(), l_RsrcIt.Name.end(), l_reg);
		std::vector<std::string> l_Tokens(l_TokIt, l_end);
		// Analyze content
		if (l_Tokens.size() >= 2 && l_RsrcIt.Info.Class() == "INSTR" && l_Tokens.back() == "INSTR")
		{
			// Parse USB resource
			if (l_Tokens[0].find("USB") == 0 && l_Tokens.size() == 5 && l_Tokens[1] == "0x1BFA" && l_Tokens[2] == "0x07EA")
			{
				// "PT2026 nnnnn USB"
				std::string l_InstrName = "PT2026 " + l_Tokens[3];
				l_ReturnedList.insert(std::pair<tUniqueID, sPrettyResource>(l_RsrcIt.Name, { l_InstrName, l_RsrcIt }));
			}
			// Parse TCPIP resource
			else if (0 == l_Tokens[0].find("TCPIP") && 4 == l_Tokens.size())
			{
				if(std::string::npos != l_Tokens[1].find("PT2026") || std::string::npos != l_Tokens[1].find("pt2026"))
				{
					// "pt2026_nnnnn"
					std::string l_InstrName = l_Tokens[1];
					l_ReturnedList.insert(std::pair<tUniqueID, sPrettyResource>(l_RsrcIt.Name, { l_InstrName, l_RsrcIt }));
				}
				//otherwise we ignore IP address to avoid having all instruments on network listed in the resource list
			}
			// Parse ASRL resource
			else if (l_Tokens[0].find("ASRL") == 0)
			{
				// "MFC3045 COMn"
				U8 l_COMnumber = static_cast<U8>(l_RsrcIt.Info.Number()); //static_cast<U8>(std::stoul(l_Tokens[0].substr(4)));
				std::string l_InstrName = "INSTR - COM" + std::to_string(l_COMnumber);
				l_ReturnedList.insert(std::pair<tUniqueID, sPrettyResource>(l_RsrcIt.Name, { l_InstrName, l_RsrcIt }));
			}
		}
	}

	return l_ReturnedList;
}
