///	\file			USBRemoteBoxController.cpp
///	\date			2020
///	\copyright (c)	Metrolab Technology S.A., Geneva, Switzerland (www.metrolab.com)
///	See the included file License Terms.pdf for the licensing conditions.

#include "USBRemoteBoxController.h"

using namespace MTL;
using namespace MTL::Instrument;

const std::string	CUSBRemoteBoxController::RSRC_FILTER		= "USB[0-9]+::0x1BFA::0x1F6D::[0-9]+::RAW";
const int			CUSBRemoteBoxController::REFRESH_RATE_MS	= 100;

typedef CException<CUSBRemoteBoxController> CUSBRemoteBoxControllerException;

CUSBRemoteBoxController::CUSBRemoteBoxController(CVISAResourceManager & rRM, CMFCManagerCatalog & rCatInstance)
	: m_rRsrcMan(rRM),
	m_RBCat(rCatInstance),
	m_ASManagerNotifier(&CUSBRemoteBoxController::l_PostManagerNotification, this),
	m_pRemoteBox(nullptr),
	m_Timer(&CUSBRemoteBoxController::l_PostTimerEvent, this)
{
	m_RBCat.oUSBRemoteBusyState.Subscribe(m_ASManagerNotifier);
}

CUSBRemoteBoxController::~CUSBRemoteBoxController()
{
	m_Timer.Stop();
	if (nullptr != m_pRemoteBox)
	{
		m_LEDs.LEDs.ON = false;
		m_LEDs.LEDs.BUSY = false;
		m_pRemoteBox->Update(m_LEDs, m_Buttons);

		delete m_pRemoteBox;
		m_pRemoteBox = nullptr;
	}
}

void CUSBRemoteBoxController::Operate()
{
	CThreadedMailBox::ThreadStart();
	m_Timer.Start(std::chrono::milliseconds(REFRESH_RATE_MS));
}

void CUSBRemoteBoxController::Terminate()
{
	m_Timer.Stop();
	CThreadedMailBox::ThreadStop();
}

void CUSBRemoteBoxController::l_PostTimerEvent()
{
	CThreadedMailBox::l_SendMsg(sUSBRemoteBoxControllerMailBoxMsg(sUSBRemoteBoxControllerMailBoxMsg::eMsgType::kTimerEvent, nullptr));
}

void CUSBRemoteBoxController::l_PostManagerNotification(CAsyncNotification * pASNot)
{
	CThreadedMailBox::l_SendMsg(sUSBRemoteBoxControllerMailBoxMsg(sUSBRemoteBoxControllerMailBoxMsg::eMsgType::kCatalogNotification, pASNot));
}

void CUSBRemoteBoxController::l_Refresh()
{
	try
	{
		// If no remote is currently connected
		if (nullptr == m_pRemoteBox)
		{
			// Find new resources
			CResourceList l_RSL;
			m_rRsrcMan.FindResources(l_RSL, RSRC_FILTER);
			if (!l_RSL.empty())
			{
				m_pRemoteBox = new CUSBRemoteBox(m_rRsrcMan, l_RSL.front());
				if (!m_pRemoteBox->Connect())
					throw CUSBRemoteBoxControllerException("Could not connect to the usb remote box", MTL__LOCATION__);
						
				m_LEDs.LEDs.ON = true;
				m_LEDs.LEDs.BUSY = false;
				if (!m_pRemoteBox->Update(m_LEDs, m_Buttons))
					throw CUSBRemoteBoxControllerException("Could not update remote box state", MTL__LOCATION__);
			}
		}
				
		// Check RB state, if any
		if (nullptr != m_pRemoteBox)
		{
			if (!m_pRemoteBox->Update(m_LEDs, m_Buttons))
				throw CUSBRemoteBoxControllerException("Could not update remote box state", MTL__LOCATION__);

			// Send relevant signals
			if (m_Buttons.BUTs.ResetEvent)
				m_RBCat.cUSBRemoteBoxReset.RequestControl(true);
			if (m_Buttons.BUTs.StartEvent)
				m_RBCat.cUSBRemoteBoxStart.RequestControl(true);
		}
	}
	catch (CUSBRemoteBoxControllerException & rE)
	{
		MTL_Unused(rE);
		if (nullptr != m_pRemoteBox)
		{
			delete m_pRemoteBox;
			m_pRemoteBox = nullptr;
		}
	}
}

void CUSBRemoteBoxController::l_SetBusyState(bool Busy)
{
	try
	{
		if (nullptr != m_pRemoteBox)
		{
			m_LEDs.LEDs.BUSY = Busy;
			if (!m_pRemoteBox->Update(m_LEDs, m_Buttons))
				throw CUSBRemoteBoxControllerException("Could not update remote box state", MTL__LOCATION__);
		}
	}
	catch (CUSBRemoteBoxControllerException & rE)
	{
		MTL_Unused(rE);
		if (nullptr != m_pRemoteBox)
		{
			delete m_pRemoteBox;
			m_pRemoteBox = nullptr;
		}
	}
}

void CUSBRemoteBoxController::l_ThreadTask(eEventType EvtType, sUSBRemoteBoxControllerMailBoxMsg * pMsg)
{
	switch (EvtType)
	{
	case eEventType::kExit:
		break;
	case eEventType::kUserMsg:
		if (nullptr != pMsg)
		{
			switch (pMsg->MsgType)
			{
			case sUSBRemoteBoxControllerMailBoxMsg::eMsgType::kTimerEvent:
				l_Refresh();
				break;
			case sUSBRemoteBoxControllerMailBoxMsg::eMsgType::kCatalogNotification:
				l_AsyncManagerNotificationHandler(pMsg->pCatNot);
				break;
			}
		}
		break;
	}
}

void CUSBRemoteBoxController::l_AsyncManagerNotificationHandler(CAsyncNotification * pASNot)
{
	CObserverNotifier_untyped & rNotifier = pASNot->Notifier();
	CObserver_untyped & rVar = pASNot->Observer();

	const CAsyncSnapshotNotification * l_pASNot = static_cast<CAsyncSnapshotNotification *>(pASNot);

	if (rVar.Is(m_RBCat.oUSBRemoteBusyState))
	{
		const bool * l_pVal = static_cast<const bool *>(l_pASNot->DataCopy());
		l_SetBusyState(*l_pVal);
	}

	delete pASNot;
}
