#include <conio.h>
#include <cerrno>

#include <iostream>
#include <iomanip>
#include <chrono>
#include <thread>

#include <PT2026>

#define PT2026_USB_RESOURCE_PATTERN		"USB[0-9]*::0x1BFA::0x07EA::[0-9]+::INSTR"
#define VISA_TIMEOUT_MS					8000

int main(void)
{
	MTL::Instrument::CVISAResourceManager	l_RM;
	MTL::Instrument::CPT2026Instrument*		l_pInstrument = NULL;;

	using namespace std::chrono_literals;
	using namespace MTL::Instrument::PT2026Types;

	// Initialize our resource manager (mandatory).
	l_RM.Initialize();

	// Try to find a PT2026 or abort when the user hit any key.
	bool _KeyboardKeyPressed = false;
	do
	{
		// Find a PT2026 resource
		MTL::Instrument::CResourceList l_RSL;
		l_RM.FindResources(l_RSL, PT2026_USB_RESOURCE_PATTERN);
		// Create our instrument interface
		if (!l_RSL.empty())
			l_pInstrument = new MTL::Instrument::CPT2026Instrument(l_RM, l_RSL.front());
		
		_KeyboardKeyPressed = _kbhit();

		if ((NULL == l_pInstrument) && !_KeyboardKeyPressed)
			std::this_thread::sleep_for(1000ms);

	} while ((NULL == l_pInstrument) && !_KeyboardKeyPressed);
	
	if (_KeyboardKeyPressed)
	{
		std::cout << "Instrument detection aborted by the user." << std::endl;
		return ECANCELED;
	}

	// Connect to the PT2026 we found 
	l_pInstrument->Connect(VISA_TIMEOUT_MS);

	/////////////////
	// Initialize
	/////////////////

	// Make sure that we are the only one actively operating the instrument
	l_pInstrument->LockInterface();
	// Ensure that we start in known condition by aborting previously started operations.
	l_pInstrument->Abort(true);
	// Define the format in which the measurement shall be returned.
	l_pInstrument->SetFormat(kComFormatInteger);
	// Define the unit in which the measurement shall be returned (Tesla)
	l_pInstrument->ParmUnitsSet(kT);

	////////////////////////////
	// Connect the active probe
	////////////////////////////
	MTL::SCPI::tChannelList l_ChanList;
	// Get the list of channels
	l_pInstrument->ChannelsGetList(l_ChanList);
	// We assume that we have only one probe connected 
	l_pInstrument->ChannelsSelect(l_ChanList);

	/////////////////////////////////////////////////////
	// Set the measurement and search in automatic mode.
	/////////////////////////////////////////////////////
	sSearch<uParm> l_Search;

	// Set the search in fully automated mode.
	l_pInstrument->ParmSearchGet(l_Search);
	l_Search.Mode = kSearchAuto;
	l_pInstrument->ParmSearchSet(l_Search);

	// Set the search limits. 
	// Here, we collect the search limits as defined in the probe.
	// One could have set a smaller search range to accelerate the search operation.
	sSearch<sBoundedParm> l_BoundedSearchParameters;
	l_pInstrument->ParmSearchGet(l_BoundedSearchParameters);
	l_pInstrument->ParmSearchSetLimits(l_BoundedSearchParameters.LowLimit_UNITS.Min, l_BoundedSearchParameters.HighLimit_UNITS.Max);

	// Set the measurement in fully automated mode.
	sMeasure<uParm> l_Measure;
	l_pInstrument->ParmMeasureGet(l_Measure);
	l_Measure.Mode = kMeasureAuto;
	l_pInstrument->ParmMeasureSet(l_Measure);

	/////////////////////////////////
	// Set the measurement averaging 
	////////////////////////////////
	sMeasurementAveraging<uParm> l_MeasAvg;
	l_MeasAvg.Type = kMeasRepeat;
	// This value gives a 1s measurement rate for a pulse period of 50ms
	l_MeasAvg.NoPoints = 20;
	l_pInstrument->ParmAveragingMeasurementSet(l_MeasAvg);

	////////////////////////////////
	// Configure the trigger
	////////////////////////////////
	sInputTrigger<uParm> l_InputTrig;
	l_InputTrig.Source = kInputTrigSrceImmediate;
	l_InputTrig.Count = 1;
	l_pInstrument->ParmTriggerInputSet(l_InputTrig);

	////////////////////////////////////////
	// Configure the service requests 
	// and Initiate a measurement
	///////////////////////////////////////

	// The variables to store the measurements
	sArbitraryMeasurements l_ArbSel;
	sAllMeasurements       l_ArbMeas;
	eUnits                 l_ArbUnits;
	// Set some default values.
	l_ArbSel.Flux = true;
	l_ArbSel.Deviation = true;
	l_ArbSel.Uniformity = true;
	l_ArbSel.Channel = true;
	l_ArbSel.Timestamp = true;
	l_ArbSel.Status = true;

	// Service request readings
	l_pInstrument->EnableEvent(MTL::Instrument::eEventType::ServiceRequest, MTL::Instrument::eEventMechanism::Queue);
	uStatusByte l_StatusSRQ;
	l_StatusSRQ.StatusByte.OSB = 1;
	l_pInstrument->StandardStatusSetEnableRegister(kServiceRequestRegister, 0x00, l_StatusSRQ.RawSTB);
	uOPERation l_Oper;
	l_Oper.OPERation.NewMeasAvail = 1;
	l_pInstrument->StatusSetEnableRegister(kStatusOperation, 0, l_Oper.RawOPER);

	// Start the measurement. This will also force an NMR signal search.
	// 
	l_pInstrument->Initiate(true);

	// Gather the measurements from the instrument and display them " la" PT2025.
	do
	{
		// States which status we are interested into.
		std::vector<sStatusRegister> l_StatusList = { sStatusRegister(kStatusOperation, kStatusCondition), sStatusRegister(kStatusOperation, kStatusEvent) };
		std::vector<U16> l_StatRes;

		l_pInstrument->StatusGet(l_StatusList, l_StatRes);

		if (l_pInstrument->WaitOnAllEvents(200))
		{
			uOPERation l_OpEv = { l_StatRes[1] };

			// A new measurement is available.
			if (l_OpEv.OPERation.NewMeasAvail)
			{
				// Collect the measurement.
				l_pInstrument->MeasurementsGet(1, l_ArbSel, l_ArbMeas, l_ArbUnits);
				// Display it " la" PT2025.
				std::cout << "L" << std::fixed << std::showpoint << std::setprecision(8) << l_ArbMeas.AdvMeas.Basic[0].Flux_UNITS << "T" << std::endl;
			}
		}
		else
		{
			uOPERation l_OperCond = { l_StatRes[0] };

			// We are in search mode.
			if (l_OperCond.OPERation.SWEeping)
			{
				U8 l_SPR;
				F64 l_HallMeas;
				l_pInstrument->GetSearchProgress(l_SPR, l_HallMeas);
				// Compute an equivalent frequency to behave as if we were operating a PT2025. The actual frequency
				// might be very different as the search operation on the PT2026 doesn't consist in a linear sweep.
				F64 _CurrentUnlockedFrequency = l_SPR / 100.0 * (l_BoundedSearchParameters.HighLimit_UNITS.Max - l_BoundedSearchParameters.LowLimit_UNITS.Min) + l_BoundedSearchParameters.LowLimit_UNITS.Min;
				// Display it " la" PT2025.
				std::cout << "N" << std::fixed << std::showpoint << std::setprecision(8) << _CurrentUnlockedFrequency << "T" << std::endl;
			}
			std::this_thread::sleep_for(200ms);

		}
		// Leave the loop if any key on the keyboard is hit.
	} while (!_kbhit());

	////////////////////////////
	// Leave the application 
	////////////////////////////
	l_pInstrument->Abort(true);
	l_pInstrument->UnlockInterface();
	l_pInstrument->~CPT2026Instrument();
}

