// Standard includes
#include <sstream>
#include <iomanip>
#include <locale>
// Personal includes
#include "PT2026.h"
#include "OSDefines.h"
#include "Helpers.h"
#include "Exception.h"
#include "PT2026TypeConversions.h"
#include "SCPIParsing.h"
// External tools includes
#include "date.h"

//----------------------------------------------------------------------//
//								Definitions								//
//----------------------------------------------------------------------//
#define DEBUG_MTL_INSTRUMENT_PT2026					1
#define DEBUG_MTL_INSTRUMENT_PT2026_ERRORS_ONLY		1
#if (defined(_DEBUG) && defined(DEBUG_MTL_INSTRUMENT_PT2026) && DEBUG_MTL_INSTRUMENT_PT2026)
#if (defined(DEBUG_MTL_INSTRUMENT_PT2026_ERRORS_ONLY) && DEBUG_MTL_INSTRUMENT_PT2026_ERRORS_ONLY)
#define MTL_INSTRUMENT_PT2026_DEBUG_COUT(__X__)
#else
#define MTL_INSTRUMENT_PT2026_DEBUG_COUT(__X__)		COUT(__X__)
#endif
#define MTL_INSTRUMENT_PT2026_DEBUG_CERR(__X__)		CERR(__X__)
#else
#define MTL_INSTRUMENT_PT2026_DEBUG_COUT(__X__)
#define MTL_INSTRUMENT_PT2026_DEBUG_CERR(__X__)
#endif

#define MTL_WRITE_AND_READ_STB_POLLING_PERIOD_MS	50
#define MTL_ERROR_LIST_HISTORY_LEN					20
#define MTL_ERROR_BUFFER_LEN						4096
#if defined(_WIN32)
	#define MTL_F64_CONVERSION_LOCALE				"en-US"
#elif defined(__APPLE__)
	#define MTL_F64_CONVERSION_LOCALE				"en_US"
#else
	#error "Selecting the locale format: unknown OS environment"
#endif
#define MTL_F64_CONVERSION_PRECISION				15

using namespace MTL::Instrument;
using namespace MTL::SCPI;
using namespace MTL::Instrument::PT2026Types;

//----------------------------------------------------------------------//
//								Exceptions								//
//----------------------------------------------------------------------//
typedef MTL::CException< CPT2026Instrument>		CPT2026InsException;

//----------------------------------------------------------------------//
//								Utilities								//
//----------------------------------------------------------------------//
void l_ParseErrorString(std::string & rErrStr, const std::string & rContext, sError & rError)
{
	// Get code
	size_t l_Coma	= rErrStr.find_first_of(',');
	rError.Code = std::stoi(rErrStr.substr(0, l_Coma));
	// Get description
	size_t l_OpenQuote	= rErrStr.find_first_of('"', l_Coma + 1);
	size_t l_CloseQuote	= rErrStr.find_last_of('"');
	rError.Description	= rErrStr.substr(l_OpenQuote + 1, l_CloseQuote - (l_OpenQuote + 1));
	rError.Context		= rContext;
}
std::string uIPAddress_to_string(uIPAddress Addr)
{
	return	std::to_string(Addr.Member.A) + '.' +
			std::to_string(Addr.Member.B) + '.' +
			std::to_string(Addr.Member.C) + '.' +
			std::to_string(Addr.Member.D);
}
std::string l_ToString(F64 number, std::streamsize precision = MTL_F64_CONVERSION_PRECISION, const char * locale = MTL_F64_CONVERSION_LOCALE)
{
	std::ostringstream l_oss;

	l_oss.imbue(std::locale(locale));
	l_oss << std::scientific << std::setprecision(precision);
	l_oss << number;

	return l_oss.str();
}
bool CPT2026Instrument::l_GetErrorList(const std::string & rContext)
{
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		uStatusByte l_STB;
		// Read STB
		if (!ReadSTB(l_STB.RawSTB))
			throw CPT2026InsException("Failed reading STB", MTL__LOCATION__);

		// While there is an error
		CVISABuffer l_Error(MTL_ERROR_BUFFER_LEN);
		while (l_STB.StatusByte.EAV)
		{
			// Try to read error string
			enum eRetry { kRetryErrorRetrieving };
			try {
				if (!Write(":SYST:ERR?"))
					throw kRetryErrorRetrieving;
				if (!Read(l_Error))
					throw kRetryErrorRetrieving;
			}
			catch (eRetry & rRetry)
			{
				if (rRetry == kRetryErrorRetrieving)
				{
					// Clear
					if (!Clear())
						throw CPT2026InsException("Could not clear device", MTL__LOCATION__);
					// Retry reading error
					if (!Write(":SYST:ERR?"))
						throw CPT2026InsException("Failed getting error list", MTL__LOCATION__);
					if (!Read(l_Error))
						throw CPT2026InsException("Failed getting error list", MTL__LOCATION__);
				}
			}
			// Parse Error
			std::string l_ErrStr = std::string(l_Error.data(), l_Error.size());
			sError l_Err;
			l_ParseErrorString(l_ErrStr, rContext, l_Err);
#if (defined(_DEBUG))
			CERR("PT2026Error " << l_Err.Code << " : " << l_Err.Description << " in " << l_Err.Context << std::endl);
#endif
			// Append to the error list
			m_ErrorList.push(l_Err);
			while (m_ErrorList.size() > MTL_ERROR_LIST_HISTORY_LEN)
				m_ErrorList.pop();
			// Read STB
			if (!ReadSTB(l_STB.RawSTB))
				throw CPT2026InsException("Failed reading STB", MTL__LOCATION__);
		}
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}

bool CPT2026Instrument::l_WriteAndRead(const std::string & rWriteStr, CVISABuffer & rReadBuffer, bool WaitForMAV)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	// Note about Status byte
	// MAV (Message Available): As long as the Output Queue contains one or more bytes, MAV shall be TRUE.MAV shall be FALSE when the Output Queue is empty.
	// EAV (Error/Event Available): As long as the Error/Event Queue contains one or more Event/Error, EAV shall be TRUE. EAV shall be FALSE when the Error/Event Queue is empty.

	try
	{
		rReadBuffer.clear();

		// Acquire VISA lock
		if (!LockExclusive(m_Timeout_ms))
			throw CPT2026InsException("Failed acquiring lock", MTL__LOCATION__);

		// Append *OPC ? if command does not include "*IDN?" (which is the only command to return an indefinite - length response, and can therefore not be followed by another query).
		//												Match a string which contains "*IDN?"
		bool l_OPCAppended = !std::regex_match(rWriteStr, std::regex("\\*IDN\\?"));
		std::string l_Query;
		if (l_OPCAppended)
			l_Query = rWriteStr + ";*OPC?";
		else
			l_Query = rWriteStr;

		// Write
		if (!Write(reinterpret_cast<ViBuf>(const_cast<char *>(l_Query.data())), static_cast<ViUInt32>(l_Query.size())))
			throw CPT2026InsException("Could not write", MTL__LOCATION__);

		// If preparing the answer can take time, we can request to wait for MAV being set before reading answer.
		// This prevents from getting a read timeout is answer is not ready in the expected amount of time.
		// We could always check for MAV before reading, but this involves getting the status byte for every query
		// and greatly increases execution time.
		if (WaitForMAV)
		{
			uStatusByte l_STB;
			if (!ReadSTB(l_STB.RawSTB))
				throw CPT2026InsException("Failed reading STB", MTL__LOCATION__);
			while (!l_STB.StatusByte.MAV)
			{
				MTL_SleepMs(MTL_WRITE_AND_READ_STB_POLLING_PERIOD_MS);
				if (!ReadSTB(l_STB.RawSTB))
					throw CPT2026InsException("Failed reading STB", MTL__LOCATION__);
			}
		}

		// There is always at least the *IDN? or the *OPC? answer
		// Ensure the read buffer has the capacity to guarantee reading the whole answer in one read (otherwise, read multiple times and check MAV bit)
		// Read
		if (!Read(rReadBuffer, true))
			throw CPT2026InsException("Could not read", MTL__LOCATION__);

		// Remove *OPC? result if given
		if (l_OPCAppended)
		{
			size_t l_MaxSizeToMatch = sizeof(";1\n") - 1;	// It is unnecessary to perform a regex on more characters than this
			std::match_results<std::vector<char>::iterator> m;
			// Match a string that ends with an optional ';', a '1', and an optional '\n'
			if (std::regex_match((rReadBuffer.size() <= l_MaxSizeToMatch) ? rReadBuffer.begin() : rReadBuffer.end() - static_cast<std::vector<char>::difference_type>(l_MaxSizeToMatch), rReadBuffer.end(), m, std::regex(";?1\\n?$")))
				rReadBuffer.resize(rReadBuffer.size() - static_cast<size_t>(m.length()));
		}

		// Check status
		uStatusByte l_STB;
		if (!ReadSTB(l_STB.RawSTB))
			throw CPT2026InsException("Failed reading STB", MTL__LOCATION__);
		if (l_STB.StatusByte.EAV)
			throw CPT2026InsException("Error available", MTL__LOCATION__);

		// Release VISA lock
		if (!Unlock())
			throw CPT2026InsException("Could not unlock", MTL__LOCATION__);
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);

		// Try to read whatever data is available.
		uStatusByte l_STB;
		while (ReadSTB(l_STB.RawSTB) && l_STB.StatusByte.MAV)
			Read(rReadBuffer, true);

		// Try to get errors
		l_GetErrorList(rWriteStr);

		// Try to release the lock
		Unlock();

		return false;
	}
	return true;
}


//----------------------------------------------------------------------//
//						Constructors / destructors						//
//----------------------------------------------------------------------//
CPT2026Instrument::CPT2026Instrument(CVISAResourceManager & rRM, tResourceName Rsrc)
	: CVISAInstrument(rRM, Rsrc), m_Timeout_ms(5000), m_ReadBuffer(65535)
{

}
CPT2026Instrument::~CPT2026Instrument()
{
	Abort(true);
	UnlockInterface();
	Disconnect();
}

//----------------------------------------------------------------------//
//							General Information							//
//----------------------------------------------------------------------//
const std::vector<sError> CPT2026Instrument::ErrorList()
{
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);
	std::vector<sError> l_ErrList;
	while (!m_ErrorList.empty())
	{
		l_ErrList.push_back(m_ErrorList.front());
		m_ErrorList.pop();
	}
	return l_ErrList;
}

//----------------------------------------------------------------------//
//								Open / Close							//
//----------------------------------------------------------------------//
bool CPT2026Instrument::Connect(U32 InitialTimeout)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);
	try
	{
		m_Timeout_ms = InitialTimeout;
		if (!CVISAInstrument::Open(eOpenAccessMode::ExclusiveLock, m_Timeout_ms))
			throw CPT2026InsException("Could not open", MTL__LOCATION__);
		if (!CVISAInstrument::SetTimeout(m_Timeout_ms))
			throw CPT2026InsException("Could not set timeout", MTL__LOCATION__);
		if (!CVISAInstrument::Clear())
			throw CPT2026InsException("Could not clear", MTL__LOCATION__);
		if (!l_WriteAndRead("*CLS", m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		// Try to close resource (in case Open succeeded)
		CVISAInstrument::Close();
		return false;
	}
	return true;
}

void CPT2026Instrument::Disconnect()
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	CVISAInstrument::Close();
}

//----------------------------------------------------------------------//
//							Interface Lock								//
//----------------------------------------------------------------------//
bool CPT2026Instrument::LockInterface()
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(":SYST:LOCK:REQ?", m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
		if (m_ReadBuffer[0] != '1')
			throw CPT2026InsException("Failed Locking interface", MTL__LOCATION__);
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}

bool CPT2026Instrument::UnlockInterface()
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try{
		if (!l_WriteAndRead(":SYST:LOCK:REL", m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}

//----------------------------------------------------------------------//
//								Channels								//
//----------------------------------------------------------------------//
bool CPT2026Instrument::ChannelsGetList(tChannelList & rChanList)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(":ROUT:PROB:SCAN?", m_ReadBuffer, true))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		size_t l_Start, l_Len;
		if (IsArbitraryBlock(m_ReadBuffer.begin(), m_ReadBuffer.end(), l_Start, l_Len))
			FromBinaryChannelList(m_ReadBuffer.data() + l_Start, l_Len, rChanList);
		else
			FromStringChannelList(m_ReadBuffer.data(), m_ReadBuffer.data() + m_ReadBuffer.size(), rChanList);
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}

bool CPT2026Instrument::ChannelsGetList(CExtendedChannelList & rChanList)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(":ROUT:ESCAN?", m_ReadBuffer, true))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		if (!StringToExtendedChannelList(std::string(m_ReadBuffer.begin(), m_ReadBuffer.end()), rChanList))
			throw CPT2026InsException(std::string("Invalid answer format in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}

bool CPT2026Instrument::ChannelsGetState(tChannelList & rSelChanList, tChannelList & rActiveChan)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(":ROUT:CLOS:STAT?;ACT?", m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		size_t l_Start, l_Len;
		if (IsArbitraryBlock(m_ReadBuffer.begin(), m_ReadBuffer.end(), l_Start, l_Len))
		{
			FromBinaryChannelList(m_ReadBuffer.data() + l_Start, l_Len, rSelChanList);
			std::vector<char>::iterator Separator = std::find(m_ReadBuffer.begin(), m_ReadBuffer.end(), ';');
			FromBinaryChannelList(&*Separator + 1, static_cast<size_t>(m_ReadBuffer.end() - Separator), rActiveChan);
		}
		else
		{
			CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
			CVISABufferParser::tTokens l_Tokens = l_BP.Tokenize();
			if (l_Tokens.size() != 2)
				throw CPT2026InsException("Invalid number of tokens in answer", MTL__LOCATION__);
			FromStringChannelList(l_Tokens[0].begin, l_Tokens[0].end, rSelChanList);
			FromStringChannelList(l_Tokens[1].begin, l_Tokens[1].end, rActiveChan);
		}
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}

bool CPT2026Instrument::ChannelsSetLimitedListState(bool State)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command = ":ROUT:CLOS:LLIS";
		l_Command += (State ? " 1" : " 0");
		if (!l_WriteAndRead(l_Command, m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}

bool CPT2026Instrument::ChannelsGetLimitedListState(bool & rState)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(":ROUT:CLOS:LLIS?", m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		std::string l_Answer(m_ReadBuffer.begin(), m_ReadBuffer.end());
		if (l_Answer == "1")
			rState = true;
		else if (l_Answer == "0")
			rState = false;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}

bool CPT2026Instrument::ChannelsSelect(const tChannelList & rChanList)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_ChanList;
		ToStringChannelList(rChanList, l_ChanList);
		std::string l_Command = ":ROUT:CLOS " + l_ChanList;
		if (!l_WriteAndRead(l_Command, m_ReadBuffer, true))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}

bool CPT2026Instrument::ChannelsGetInfo(const tChannelList & rChanList, std::vector<sChannelIformation> & rChanInfo)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	rChanInfo.clear();

	try {
		std::string l_ChanList;
		ToStringChannelList(rChanList, l_ChanList);
		std::string l_Command =	":ROUT:PROB:MOD? " + l_ChanList +
								";SER? " + l_ChanList +
								";DES? " + l_ChanList +
								";FWV? " + l_ChanList;
		if (!l_WriteAndRead(l_Command, m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		CVISABufferParser::tTokens l_Tokens = l_BP.Tokenize();
		if (l_Tokens.size() != 4)
			throw CPT2026InsException("Invalid number of tokens in answer", MTL__LOCATION__);

		// Models
		std::vector<std::string> l_Models;
		SplitString(std::string(l_Tokens[0].begin, l_Tokens[0].end), l_Models, ',');
		if (l_Models.size() != rChanList.size())
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Serial Numbers
		std::vector<std::string> l_SNs;
		SplitString(std::string(l_Tokens[1].begin, l_Tokens[1].end), l_SNs, ',');
		if (l_SNs.size() != rChanList.size())
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Designations
		std::vector<std::string> l_Des;
		SplitString(std::string(l_Tokens[2].begin, l_Tokens[2].end), l_Des, ',');
		if (l_Des.size() != rChanList.size())
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Firmware versions
		std::vector<std::string> l_FWVs;
		SplitString(std::string(l_Tokens[3].begin, l_Tokens[3].end), l_FWVs, ',');
		if (l_FWVs.size() != rChanList.size())
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);

		// Fill information
		for (size_t l_ChanNo = 0; l_ChanNo < rChanList.size(); l_ChanNo++)
		{
			sChannelIformation ChanInfo(static_cast<tModel>(std::stoul(l_Models[l_ChanNo])), static_cast<tSerialNumber>(std::stoul(l_SNs[l_ChanNo])), l_Des[l_ChanNo], static_cast<tFirmwareVersion>(std::stoul(l_FWVs[l_ChanNo], nullptr, 16)));
			rChanInfo.push_back(ChanInfo);
		}
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ChannelsGetLimits(const MTL::SCPI::tChannelList & rChanList, std::vector<F64> & rLowLimit_UNITS, std::vector<F64> & rHighLimit_UNITS)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	rLowLimit_UNITS.clear();
	rHighLimit_UNITS.clear();

	try {
		std::string l_ChanList;
		ToStringChannelList(rChanList, l_ChanList);
		std::string l_Command =	":ROUT:PROB:MIN? " + l_ChanList +
								";MAX? " + l_ChanList;
		if (!l_WriteAndRead(l_Command, m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		CVISABufferParser::tTokens l_Tokens = l_BP.Tokenize();
		if (l_Tokens.size() != 2)
			throw CPT2026InsException("Invalid number of tokens in answer", MTL__LOCATION__);

		// Low Lomits
		std::vector<std::string> l_LowLimits;
		SplitString(std::string(l_Tokens[0].begin, l_Tokens[0].end), l_LowLimits, ',');
		if (l_LowLimits.size() != rChanList.size())
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// High Limits
		std::vector<std::string> l_HighLimits;
		SplitString(std::string(l_Tokens[1].begin, l_Tokens[1].end), l_HighLimits, ',');
		if (l_HighLimits.size() != rChanList.size())
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);

		// Fill information
		for (size_t l_ChanNo = 0; l_ChanNo < rChanList.size(); l_ChanNo++)
		{
			rLowLimit_UNITS.push_back(std::stod(l_LowLimits[l_ChanNo]));
			rHighLimit_UNITS.push_back(std::stod(l_HighLimits[l_ChanNo]));
		}
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ChannelGetRawHall(sProbeRawHall & rRawHall)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	rRawHall.clear();

	try {
		if (!l_WriteAndRead(":ROUT:PROB:HALL:RAW?", m_ReadBuffer, true))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		CVISABufferParser::tTokens l_Tokens = l_BP.Tokenize(',');
		if (l_Tokens.size() != 4)
			throw CPT2026InsException("Invalid number of tokens in answer", MTL__LOCATION__);

		rRawHall.Bx	= static_cast<U8>(std::stoul(std::string(l_Tokens[0].begin, l_Tokens[0].end)));
		rRawHall.Bz	= static_cast<U8>(std::stoul(std::string(l_Tokens[2].begin, l_Tokens[2].end)));
		rRawHall.By	= static_cast<U8>(std::stoul(std::string(l_Tokens[1].begin, l_Tokens[1].end)));
		rRawHall.T	= static_cast<U8>(std::stoul(std::string(l_Tokens[3].begin, l_Tokens[3].end)));
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ChannelGetHall(sProbeHall & rHall)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	rHall.clear();

	try {
		if (!l_WriteAndRead(":ROUT:PROB:HALL:CAL?", m_ReadBuffer, true))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		CVISABufferParser::tTokens l_Tokens = l_BP.Tokenize(',');
		if (l_Tokens.size() != 4)
			throw CPT2026InsException("Invalid number of tokens in answer", MTL__LOCATION__);

		rHall.Bx	= std::stod(std::string(l_Tokens[0].begin, l_Tokens[0].end));
		rHall.By	= std::stod(std::string(l_Tokens[1].begin, l_Tokens[1].end));
		rHall.Bz	= std::stod(std::string(l_Tokens[2].begin, l_Tokens[2].end));
		rHall.B		= std::stod(std::string(l_Tokens[3].begin, l_Tokens[3].end));
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ChannelGetRemoteLEDMode(eRemoteBusyLEDmode & rMode)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(":ROUT:PROB:REM:LEDM?", m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		U8 l_ModeNumber = static_cast<U8>(std::stoul(std::string(m_ReadBuffer.begin(), m_ReadBuffer.end())));
		if (l_ModeNumber > MAX_BUSY_LED_MODE)
			throw CPT2026InsException("Unexpected busy LED mode has been returned", MTL__LOCATION__);

		rMode = static_cast<eRemoteBusyLEDmode>(l_ModeNumber);
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ChannelSetRemoteLEDMode(eRemoteBusyLEDmode Mode)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command = ":ROUT:PROB:REM:LEDM " + std::to_string(Mode);
	return l_WriteAndRead(l_Command, m_ReadBuffer);
}

//----------------------------------------------------------------------//
//								Parameters								//
//----------------------------------------------------------------------//
bool CPT2026Instrument::ParmSearchSet(const sSearch<uParm> & rSearch)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command;
	switch (rSearch.Mode)
	{
	case kSearchAuto:
		l_Command = ":CONF:SEAR:MODE AUTO";											// Mode
		break;
	case kSearchCustom:
		l_Command = ":CONF:SEAR:MODE CUST";											// Mode
		l_Command += ";LEV " + l_ToString(rSearch.DetectionLevel_V) + "V";			// Detection Level
		l_Command += ";FSTE " + l_ToString(rSearch.FrequencyStep_Hz) + "HZ";		// Frequency Step
		break;
	case kSearchManual:
		l_Command = ":CONF:SEAR:MODE MAN";											// Mode
		l_Command += ";LEV " + l_ToString(rSearch.DetectionLevel_V) + "V";			// Detection Level
		l_Command += ";VAL " + l_ToString(rSearch.ManualValue_UNITS);				// Manual Value
		break;
	default:
		MTL_Assert(false);
		break;
	}
	l_Command += ";LOW " + l_ToString(rSearch.LowLimit_UNITS);						// Low Limit
	l_Command += ";HIGH " + l_ToString(rSearch.HighLimit_UNITS);					// High Limit

	return l_WriteAndRead(l_Command, m_ReadBuffer);
}
bool CPT2026Instrument::ParmSearchGet(sSearch<uParm> & rSearch)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(	":CONF:SEAR:MODE?"
								";LEV?"
								";FSTE?"
								";HIGH?"
								";LOW?"
								";VAL?",
								m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		CVISABufferParser::tTokens l_Tokens = l_BP.Tokenize();
		if (l_Tokens.size() != 6)
			throw CPT2026InsException("Invalid number of tokens in answer", MTL__LOCATION__);

		// Mode
		std::string l_Mode(l_Tokens[0].begin, l_Tokens[0].end);
		if (l_Mode == "AUTO")
			rSearch.Mode = kSearchAuto;
		else if (l_Mode == "CUSTom")
			rSearch.Mode = kSearchCustom;
		else if (l_Mode == "MANual")
			rSearch.Mode = kSearchManual;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Detection Level
		rSearch.DetectionLevel_V = std::stod(std::string(l_Tokens[1].begin, l_Tokens[1].end));
		// Frequency Step
		rSearch.FrequencyStep_Hz = std::stod(std::string(l_Tokens[2].begin, l_Tokens[2].end));
		// High Limit
		rSearch.HighLimit_UNITS = std::stod(std::string(l_Tokens[3].begin, l_Tokens[3].end));
		// Low Limit
		rSearch.LowLimit_UNITS = std::stod(std::string(l_Tokens[4].begin, l_Tokens[4].end));
		// Manual Value
		rSearch.ManualValue_UNITS = std::stod(std::string(l_Tokens[5].begin, l_Tokens[5].end));
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmSearchGet(sSearch<sBoundedParm> & rSearch)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(	":CONF:SEAR:MODE?"
								";LEV?;LEV? MIN;LEV? MAX;LEV? DEF"
								";FSTE?;FSTE? MIN;FSTE? MAX;FSTE? DEF"
								";HIGH?;HIGH? MIN;HIGH? MAX;HIGH? DEF"
								";LOW?;LOW? MIN;LOW? MAX;LOW? DEF"
								";VAL?;VAL? MIN;VAL? MAX;VAL? DEF",
								m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		CVISABufferParser::tTokens l_Tokens = l_BP.Tokenize();
		if (l_Tokens.size() != 21)
			throw CPT2026InsException("Invalid number of tokens in answer", MTL__LOCATION__);

		// Mode
		std::string l_Mode(l_Tokens[0].begin, l_Tokens[0].end);
		if (l_Mode == "AUTO")
			rSearch.Mode = kSearchAuto;
		else if (l_Mode == "CUSTom")
			rSearch.Mode = kSearchCustom;
		else if (l_Mode == "MANual")
			rSearch.Mode = kSearchManual;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Detection Level
		rSearch.DetectionLevel_V.Val = std::stod(std::string(l_Tokens[1].begin, l_Tokens[1].end));
		rSearch.DetectionLevel_V.Min = std::stod(std::string(l_Tokens[2].begin, l_Tokens[2].end));
		rSearch.DetectionLevel_V.Max = std::stod(std::string(l_Tokens[3].begin, l_Tokens[3].end));
		rSearch.DetectionLevel_V.Def = std::stod(std::string(l_Tokens[4].begin, l_Tokens[4].end));
		// Frequency Step
		rSearch.FrequencyStep_Hz.Val = std::stod(std::string(l_Tokens[5].begin, l_Tokens[5].end));
		rSearch.FrequencyStep_Hz.Min = std::stod(std::string(l_Tokens[6].begin, l_Tokens[6].end));
		rSearch.FrequencyStep_Hz.Max = std::stod(std::string(l_Tokens[7].begin, l_Tokens[7].end));
		rSearch.FrequencyStep_Hz.Def = std::stod(std::string(l_Tokens[8].begin, l_Tokens[8].end));
		// High Limit
		rSearch.HighLimit_UNITS.Val = std::stod(std::string(l_Tokens[9].begin, l_Tokens[9].end));
		rSearch.HighLimit_UNITS.Min = std::stod(std::string(l_Tokens[10].begin, l_Tokens[10].end));
		rSearch.HighLimit_UNITS.Max = std::stod(std::string(l_Tokens[11].begin, l_Tokens[11].end));
		rSearch.HighLimit_UNITS.Def = std::stod(std::string(l_Tokens[12].begin, l_Tokens[12].end));
		// Low Limit
		rSearch.LowLimit_UNITS.Val = std::stod(std::string(l_Tokens[13].begin, l_Tokens[13].end));
		rSearch.LowLimit_UNITS.Min = std::stod(std::string(l_Tokens[14].begin, l_Tokens[14].end));
		rSearch.LowLimit_UNITS.Max = std::stod(std::string(l_Tokens[15].begin, l_Tokens[15].end));
		rSearch.LowLimit_UNITS.Def = std::stod(std::string(l_Tokens[16].begin, l_Tokens[16].end));
		// Manual Value
		rSearch.ManualValue_UNITS.Val = std::stod(std::string(l_Tokens[17].begin, l_Tokens[17].end));
		rSearch.ManualValue_UNITS.Min = std::stod(std::string(l_Tokens[18].begin, l_Tokens[18].end));
		rSearch.ManualValue_UNITS.Max = std::stod(std::string(l_Tokens[19].begin, l_Tokens[19].end));
		rSearch.ManualValue_UNITS.Def = std::stod(std::string(l_Tokens[20].begin, l_Tokens[20].end));
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmSearchSetManualValue(F64 ManualValue_UNITS)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command = ":CONF:SEAR:VAL " + l_ToString(ManualValue_UNITS);

	return l_WriteAndRead(l_Command, m_ReadBuffer);
}
bool CPT2026Instrument::ParmSearchSetLimits(F64 LowLimit_UNITS, F64 HighLimit_UNITS)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command = ":CONF:SEAR:LIM";
	l_Command += ":LOW " + l_ToString(LowLimit_UNITS);
	l_Command += ";HIGH " + l_ToString(HighLimit_UNITS);

	return l_WriteAndRead(l_Command, m_ReadBuffer);
}
bool CPT2026Instrument::ParmSearchSetHallEnable(bool Enabled)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command = std::string(":CONF:SEAR:HALL ") + (Enabled ? "1" : "0");

	return l_WriteAndRead(l_Command, m_ReadBuffer);
}
bool CPT2026Instrument::ParmSearchGetHallEnable(bool & rEnabled)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(":CONF:SEAR:HALL?", m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		std::string l_Answer(m_ReadBuffer.begin(), m_ReadBuffer.end());
		if (l_Answer == "1")
			rEnabled = true;
		else if (l_Answer == "0")
			rEnabled = false;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmMeasureSet(const sMeasure<uParm> & rMeasure)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command;
	switch (rMeasure.Mode)
	{
	case kMeasureAuto:
		l_Command = ":CONF:MEAS:MODE AUTO";														// Mode
		break;
	case kMeasureManual:
		l_Command = ":CONF:MEAS:MODE MAN";														// Mode
		l_Command += ";LEV " + l_ToString(rMeasure.DetectionLevel_V);							// Detection Level
		l_Command += ";HYST " + std::to_string(rMeasure.AllowedMissMeas);						// Allowed Missing Measurements
		l_Command += rMeasure.Reject ? ";REJ ON" : ";REJ OFF";									// Spurious Rejection
		l_Command += ";BAND " + l_ToString(rMeasure.Bandwidth_Hz);								// Spectrum Bandwidth
		l_Command += ";POIN " + std::to_string(rMeasure.NoFitPoints);							// Spectrum Points
		l_Command += ";:CONF:TRAC:HYST " + std::to_string(rMeasure.Tracking.AllowedMissMeas);	// Tracking Allowed Lost
		l_Command += ";LOW " + l_ToString(rMeasure.Tracking.MinIF_Hz);							// Tracking Low IF
		l_Command += ";HIGH " + l_ToString(rMeasure.Tracking.MaxIF_Hz);							// Tracking High IF
		break;
	default:
		MTL_Assert(false);
		break;
	}

	return l_WriteAndRead(l_Command, m_ReadBuffer);
}
bool CPT2026Instrument::ParmMeasureGet(sMeasure<uParm> & rMeasure)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(	":CONF:MEAS:MODE?"
								";REJ?"
								";LEV?"
								";BAND?"
								";POIN?"
								";HYST?"
								";:CONF:TRAC:HYST?"
								";LOW?"
								";HIGH?",
								m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		CVISABufferParser::tTokens l_Tokens = l_BP.Tokenize();
		if (l_Tokens.size() != 9)
			throw CPT2026InsException("Invalid number of tokens in answer", MTL__LOCATION__);

		// Mode
		std::string l_Mode(l_Tokens[0].begin, l_Tokens[0].end);
		if (l_Mode == "AUTO")
			rMeasure.Mode = kMeasureAuto;
		else if (l_Mode == "MANual")
			rMeasure.Mode = kMeasureManual;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Rejection
		std::string l_Rejection(l_Tokens[1].begin, l_Tokens[1].end);
		if (l_Rejection == "1")
			rMeasure.Reject = true;
		else if (l_Rejection == "0")
			rMeasure.Reject = false;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Detection Level
		rMeasure.DetectionLevel_V = std::stod(std::string(l_Tokens[2].begin, l_Tokens[2].end));
		// Bandwidth
		rMeasure.Bandwidth_Hz = std::stod(std::string(l_Tokens[3].begin, l_Tokens[3].end));
		// Fitting Points
		rMeasure.NoFitPoints = static_cast<U32>(std::stol(std::string(l_Tokens[4].begin, l_Tokens[4].end)));
		// Allowed Missing Measurements
		rMeasure.AllowedMissMeas = static_cast<U32>(std::stol(std::string(l_Tokens[5].begin, l_Tokens[5].end)));
		// Tracking Missing Measurements
		rMeasure.Tracking.AllowedMissMeas = static_cast<U32>(std::stol(std::string(l_Tokens[6].begin, l_Tokens[6].end)));
		// Tracking Min IF
		rMeasure.Tracking.MinIF_Hz = std::stod(std::string(l_Tokens[7].begin, l_Tokens[7].end));
		// Tracking Max IF
		rMeasure.Tracking.MaxIF_Hz = std::stod(std::string(l_Tokens[8].begin, l_Tokens[8].end));
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmMeasureGet(sMeasure<sBoundedParm> & rMeasure)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(	":CONF:MEAS:MODE?"
								";REJ?"
								";LEV?;LEV? MIN;LEV? MAX;LEV? DEF"
								";BAND?;BAND? MIN;BAND? MAX;BAND? DEF"
								";POIN?;POIN? MIN;POIN? MAX;POIN? DEF"
								";HYST?;HYST? MIN;HYST? MAX;HYST? DEF"
								";:CONF:TRAC:HYST?;HYST? MIN;HYST? MAX;HYST? DEF"
								";LOW?;LOW? MIN;LOW? MAX;LOW? DEF"
								";HIGH?;HIGH? MIN;HIGH? MAX;HIGH? DEF",
								m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		CVISABufferParser::tTokens l_Tokens = l_BP.Tokenize();
		if (l_Tokens.size() != 30)
			throw CPT2026InsException("Invalid number of tokens in answer", MTL__LOCATION__);

		// Mode
		std::string l_Mode(l_Tokens[0].begin, l_Tokens[0].end);
		if (l_Mode == "AUTO")
			rMeasure.Mode = kMeasureAuto;
		else if (l_Mode == "MANual")
			rMeasure.Mode = kMeasureManual;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Rejection
		std::string l_Rejection(l_Tokens[1].begin, l_Tokens[1].end);
		if (l_Rejection == "1")
			rMeasure.Reject = true;
		else if (l_Rejection == "0")
			rMeasure.Reject = false;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Detection Level
		rMeasure.DetectionLevel_V.Val = std::stod(std::string(l_Tokens[2].begin, l_Tokens[2].end));
		rMeasure.DetectionLevel_V.Min = std::stod(std::string(l_Tokens[3].begin, l_Tokens[3].end));
		rMeasure.DetectionLevel_V.Max = std::stod(std::string(l_Tokens[4].begin, l_Tokens[4].end));
		rMeasure.DetectionLevel_V.Def = std::stod(std::string(l_Tokens[5].begin, l_Tokens[5].end));
		// Bandwidth
		rMeasure.Bandwidth_Hz.Val = std::stod(std::string(l_Tokens[6].begin, l_Tokens[6].end));
		rMeasure.Bandwidth_Hz.Min = std::stod(std::string(l_Tokens[7].begin, l_Tokens[7].end));
		rMeasure.Bandwidth_Hz.Max = std::stod(std::string(l_Tokens[8].begin, l_Tokens[8].end));
		rMeasure.Bandwidth_Hz.Def = std::stod(std::string(l_Tokens[9].begin, l_Tokens[9].end));
		// Fitting Points
		rMeasure.NoFitPoints.Val = static_cast<U32>(std::stol(std::string(l_Tokens[10].begin, l_Tokens[10].end)));
		rMeasure.NoFitPoints.Min = static_cast<U32>(std::stol(std::string(l_Tokens[11].begin, l_Tokens[11].end)));
		rMeasure.NoFitPoints.Max = static_cast<U32>(std::stol(std::string(l_Tokens[12].begin, l_Tokens[12].end)));
		rMeasure.NoFitPoints.Def = static_cast<U32>(std::stol(std::string(l_Tokens[13].begin, l_Tokens[13].end)));
		// Allowed Missing Measurements
		rMeasure.AllowedMissMeas.Val = static_cast<U32>(std::stol(std::string(l_Tokens[14].begin, l_Tokens[14].end)));
		rMeasure.AllowedMissMeas.Min = static_cast<U32>(std::stol(std::string(l_Tokens[15].begin, l_Tokens[15].end)));
		rMeasure.AllowedMissMeas.Max = static_cast<U32>(std::stol(std::string(l_Tokens[16].begin, l_Tokens[16].end)));
		rMeasure.AllowedMissMeas.Def = static_cast<U32>(std::stol(std::string(l_Tokens[17].begin, l_Tokens[17].end)));
		// Tracking Missing Measurements
		rMeasure.Tracking.AllowedMissMeas.Val = static_cast<U32>(std::stol(std::string(l_Tokens[18].begin, l_Tokens[18].end)));
		rMeasure.Tracking.AllowedMissMeas.Min = static_cast<U32>(std::stol(std::string(l_Tokens[19].begin, l_Tokens[19].end)));
		rMeasure.Tracking.AllowedMissMeas.Max = static_cast<U32>(std::stol(std::string(l_Tokens[20].begin, l_Tokens[20].end)));
		rMeasure.Tracking.AllowedMissMeas.Def = static_cast<U32>(std::stol(std::string(l_Tokens[21].begin, l_Tokens[21].end)));
		// Tracking Min IF
		rMeasure.Tracking.MinIF_Hz.Val = std::stod(std::string(l_Tokens[22].begin, l_Tokens[22].end));
		rMeasure.Tracking.MinIF_Hz.Min = std::stod(std::string(l_Tokens[23].begin, l_Tokens[23].end));
		rMeasure.Tracking.MinIF_Hz.Max = std::stod(std::string(l_Tokens[24].begin, l_Tokens[24].end));
		rMeasure.Tracking.MinIF_Hz.Def = std::stod(std::string(l_Tokens[25].begin, l_Tokens[25].end));
		// Tracking Max IF
		rMeasure.Tracking.MaxIF_Hz.Val = std::stod(std::string(l_Tokens[26].begin, l_Tokens[26].end));
		rMeasure.Tracking.MaxIF_Hz.Min = std::stod(std::string(l_Tokens[27].begin, l_Tokens[27].end));
		rMeasure.Tracking.MaxIF_Hz.Max = std::stod(std::string(l_Tokens[28].begin, l_Tokens[28].end));
		rMeasure.Tracking.MaxIF_Hz.Def = std::stod(std::string(l_Tokens[29].begin, l_Tokens[29].end));
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmAveragingSignalSet(const sSignalAveraging<uParm> & rSigAvg)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command = ":AVERage1:STAT ";
	switch (rSigAvg.Type)
	{
	case kSigNone:
		l_Command += "OFF";
		break;
	case kSigExponential:
		l_Command += "ON" ";:AVERage1:TCON EXP";
		l_Command += ";:AVERage1:COUN " + std::to_string(rSigAvg.NoPoints);
		break;
	case kSigRepeat:
		l_Command += "ON" ";:AVERage1:TCON REP";
		l_Command += ";:AVERage1:COUN " + std::to_string(rSigAvg.NoPoints);
		break;
	default:
		MTL_Assert(false);
		break;
	}

	return l_WriteAndRead(l_Command, m_ReadBuffer);
}
bool CPT2026Instrument::ParmAveragingSignalGet(sSignalAveraging<uParm> & rSigAvg)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(	":AVERage1:STAT?"
								";:AVERage1:TCON?"
								";:AVERage1:COUN?",
								m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		CVISABufferParser::tTokens l_Tokens = l_BP.Tokenize();
		if (l_Tokens.size() != 3)
			throw CPT2026InsException("Invalid number of tokens in answer", MTL__LOCATION__);

		// State
		std::string l_State(l_Tokens[0].begin, l_Tokens[0].end);
		enum { kInactive, kActive } l_eState;
		if (l_State == "0")
			l_eState = kInactive;
		else if (l_State == "1")
			l_eState = kActive;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Type
		std::string l_Type(l_Tokens[1].begin, l_Tokens[1].end);
		enum { kExponential, kRepeat } l_eType;
		if (l_Type == "EXPonential")
			l_eType = kExponential;
		else if (l_Type == "REPeat")
			l_eType = kRepeat;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);

		if (l_eState == kInactive)
			rSigAvg.Type = kSigNone;
		else if (l_eState == kActive)
		{
			if (l_eType == kExponential)
				rSigAvg.Type = kSigExponential;
			else if (l_eType == kRepeat)
				rSigAvg.Type = kSigRepeat;
			else
				MTL_Assert(false);
		}
		else
			MTL_Assert(false);
		// Count
		rSigAvg.NoPoints = static_cast<U32>(std::stoi(std::string(l_Tokens[2].begin, l_Tokens[2].end)));
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmAveragingSignalGet(sSignalAveraging<sBoundedParm> & rSigAvg)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(	":AVERage1:STAT?"
								";:AVERage1:TCON?"
								";:AVERage1:COUN?;:AVERage1:COUN? MIN;:AVERage1:COUN? MAX;:AVERage1:COUN? DEF",
								m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		CVISABufferParser::tTokens l_Tokens = l_BP.Tokenize();
		if (l_Tokens.size() != 6)
			throw CPT2026InsException("Invalid number of tokens in answer", MTL__LOCATION__);

		// State
		std::string l_State(l_Tokens[0].begin, l_Tokens[0].end);
		enum { kInactive, kActive } l_eState;
		if (l_State == "0")
			l_eState = kInactive;
		else if (l_State == "1")
			l_eState = kActive;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Type
		std::string l_Type(l_Tokens[1].begin, l_Tokens[1].end);
		enum { kExponential, kRepeat } l_eType;
		if (l_Type == "EXPonential")
			l_eType = kExponential;
		else if (l_Type == "REPeat")
			l_eType = kRepeat;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);

		if (l_eState == kInactive)
			rSigAvg.Type = kSigNone;
		else if (l_eState == kActive)
		{
			if (l_eType == kExponential)
				rSigAvg.Type = kSigExponential;
			else if (l_eType == kRepeat)
				rSigAvg.Type = kSigRepeat;
			else
				MTL_Assert(false);
		}
		else
			MTL_Assert(false);
		// Count
		rSigAvg.NoPoints.Val = static_cast<U32>(std::stoi(std::string(l_Tokens[2].begin, l_Tokens[2].end)));
		rSigAvg.NoPoints.Min = static_cast<U32>(std::stoi(std::string(l_Tokens[3].begin, l_Tokens[3].end)));
		rSigAvg.NoPoints.Max = static_cast<U32>(std::stoi(std::string(l_Tokens[4].begin, l_Tokens[4].end)));
		rSigAvg.NoPoints.Def = static_cast<U32>(std::stoi(std::string(l_Tokens[5].begin, l_Tokens[5].end)));
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmAveragingMeasurementSet(const sMeasurementAveraging<uParm> & rMeasAvg)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command = ":AVERage2:STAT ";
	switch (rMeasAvg.Type)
	{
	case kMeasNone:
		l_Command += "OFF";
		break;
	case kMeasExponential:
		l_Command += "ON" ";:AVERage2:TCON EXP";
		l_Command += ";:AVERage2:COUN " + std::to_string(rMeasAvg.NoPoints);
		break;
	case kMeasMoving:
		l_Command += "ON" ";:AVERage2:TCON EXP";
		l_Command += ";:AVERage2:COUN " + std::to_string(rMeasAvg.NoPoints);
		break;
	case kMeasRepeat:
		l_Command += "ON" ";:AVERage2:TCON REP";
		l_Command += ";:AVERage2:COUN " + std::to_string(rMeasAvg.NoPoints);
		break;
	default:
		MTL_Assert(false);
		break;
	}

	return l_WriteAndRead(l_Command, m_ReadBuffer);
}
bool CPT2026Instrument::ParmAveragingMeasurementGet(sMeasurementAveraging<uParm> & rMeasAvg)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(	":AVERage2:STAT?"
								";:AVERage2:TCON?"
								";:AVERage2:COUN?",
								m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);


		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		CVISABufferParser::tTokens l_Tokens = l_BP.Tokenize();
		if (l_Tokens.size() != 3)
			throw CPT2026InsException("Invalid number of tokens in answer", MTL__LOCATION__);

		// State
		std::string l_State(l_Tokens[0].begin, l_Tokens[0].end);
		enum { kInactive, kActive } l_eState;
		if (l_State == "0")
			l_eState = kInactive;
		else if (l_State == "1")
			l_eState = kActive;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Type
		std::string l_Type(l_Tokens[1].begin, l_Tokens[1].end);
		enum { kExponential, kMoving, kRepeat } l_eType;
		if (l_Type == "EXPonential")
			l_eType = kExponential;
		else if (l_Type == "MOVing")
			l_eType = kMoving;
		else if (l_Type == "REPeat")
			l_eType = kRepeat;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);

		if (l_eState == kInactive)
			rMeasAvg.Type = kMeasNone;
		else if (l_eState == kActive)
		{
			if (l_eType == kExponential)
				rMeasAvg.Type = kMeasExponential;
			else if (l_eType == kMoving)
				rMeasAvg.Type = kMeasMoving;
			else if (l_eType == kRepeat)
				rMeasAvg.Type = kMeasRepeat;
			else
				MTL_Assert(false);
		}
		else
			MTL_Assert(false);
		// Count
		rMeasAvg.NoPoints = std::stoi(std::string(l_Tokens[2].begin, l_Tokens[2].end));
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmAveragingMeasurementGet(sMeasurementAveraging<sBoundedParm> & rMeasAvg)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(	":AVERage2:STAT?"
								";:AVERage2:TCON?"
								";:AVERage2:COUN?;:AVERage2:COUN? MIN;:AVERage2:COUN? MAX;:AVERage2:COUN? DEF",
								m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		CVISABufferParser::tTokens l_Tokens = l_BP.Tokenize();
		if (l_Tokens.size() != 6)
			throw CPT2026InsException("Invalid number of tokens in answer", MTL__LOCATION__);

		// State
		std::string l_State(l_Tokens[0].begin, l_Tokens[0].end);
		enum { kInactive, kActive } l_eState;
		if (l_State == "0")
			l_eState = kInactive;
		else if (l_State == "1")
			l_eState = kActive;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Type
		std::string l_Type(l_Tokens[1].begin, l_Tokens[1].end);
		enum { kExponential, kMoving, kRepeat } l_eType;
		if (l_Type == "EXPonential")
			l_eType = kExponential;
		else if (l_Type == "MOVing")
			l_eType = kMoving;
		else if (l_Type == "REPeat")
			l_eType = kRepeat;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);

		if (l_eState == kInactive)
			rMeasAvg.Type = kMeasNone;
		else if (l_eState == kActive)
		{
			if (l_eType == kExponential)
				rMeasAvg.Type = kMeasExponential;
			else if (l_eType == kMoving)
				rMeasAvg.Type = kMeasMoving;
			else if (l_eType == kRepeat)
				rMeasAvg.Type = kMeasRepeat;
			else
				MTL_Assert(false);
		}
		else
			MTL_Assert(false);
		// Count
		rMeasAvg.NoPoints.Val = std::stoi(std::string(l_Tokens[2].begin, l_Tokens[2].end));
		rMeasAvg.NoPoints.Min = std::stoi(std::string(l_Tokens[3].begin, l_Tokens[3].end));
		rMeasAvg.NoPoints.Max = std::stoi(std::string(l_Tokens[4].begin, l_Tokens[4].end));
		rMeasAvg.NoPoints.Def = std::stoi(std::string(l_Tokens[5].begin, l_Tokens[5].end));
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmTriggerInputSet(const sInputTrigger<uParm> & rInputTrig)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command = ":TRIG:SEQuence1:SOUR ";
	switch (rInputTrig.Source)
	{
	case kInputTrigSrceImmediate:
		l_Command += "IMM";
		l_Command += ";COUN " + std::to_string(rInputTrig.Count);
		break;
	case kInputTrigSrceTimer:
		l_Command += "TIM";
		l_Command += ";COUN " + std::to_string(rInputTrig.Count);
		l_Command += ";TIM " + l_ToString(rInputTrig.Period_s);
		break;
	case kInputTrigSrceBus:
		l_Command += "BUS";
		l_Command += ";COUN " + std::to_string(rInputTrig.Count);
		break;
	case kInputTrigSrceExternal:
		l_Command += "EXT";
		l_Command += ";COUN " + std::to_string(rInputTrig.Count);
		switch (rInputTrig.Edge)
		{
		case kInputTrigEdgeRising:
			l_Command += ";SLOP POS";
			break;
		case kInputTrigEdgeFalling:
			l_Command += ";SLOP NEG";
			break;
		case kInputTrigEdgeBoth:
			l_Command += ";SLOP EITH";
			break;
		default:
			MTL_Assert(false);
			break;
		}
		break;
	default:
		MTL_Assert(false);
		break;
	}

	return l_WriteAndRead(l_Command, m_ReadBuffer);
}
bool CPT2026Instrument::ParmTriggerInputGet(sInputTrigger<uParm> & rInputTrig)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(	":TRIG:SEQuence1:SOUR?"
								";TIM?"
								";COUN?"
								";SLOP?",
								m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		CVISABufferParser::tTokens l_Tokens = l_BP.Tokenize();
		if (l_Tokens.size() != 4)
			throw CPT2026InsException("Invalid number of tokens in answer", MTL__LOCATION__);

		// Source
		std::string l_Source(l_Tokens[0].begin, l_Tokens[0].end);
		if (l_Source == "IMMediate")
			rInputTrig.Source = kInputTrigSrceImmediate;
		else if (l_Source == "TIMer")
			rInputTrig.Source = kInputTrigSrceTimer;
		else if (l_Source == "BUS")
			rInputTrig.Source = kInputTrigSrceBus;
		else if (l_Source == "EXTernal")
			rInputTrig.Source = kInputTrigSrceExternal;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Period
		rInputTrig.Period_s = std::stod(std::string(l_Tokens[1].begin, l_Tokens[1].end));
		// Trigger Count
		rInputTrig.Count = std::stoi(std::string(l_Tokens[2].begin, l_Tokens[2].end));
		// Slope
		std::string l_Slope(l_Tokens[3].begin, l_Tokens[3].end);
		if (l_Slope == "POSitive")
			rInputTrig.Edge = kInputTrigEdgeRising;
		else if (l_Slope == "NEGative")
			rInputTrig.Edge = kInputTrigEdgeFalling;
		else if (l_Slope == "EITHer")
			rInputTrig.Edge = kInputTrigEdgeBoth;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmTriggerInputGet(sInputTrigger<sBoundedParm> & rInputTrig)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(	":TRIG:SEQuence1:SOUR?"
								";TIM?;TIM? MIN;TIM? MAX;TIM? DEF"
								";COUN?;COUN? MIN;COUN? MAX;COUN? DEF"
								";SLOP?",
								m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		CVISABufferParser::tTokens l_Tokens = l_BP.Tokenize();
		if (l_Tokens.size() != 10)
			throw CPT2026InsException("Invalid number of tokens in answer", MTL__LOCATION__);

		// Source
		std::string l_Source(l_Tokens[0].begin, l_Tokens[0].end);
		if (l_Source == "IMMediate")
			rInputTrig.Source = kInputTrigSrceImmediate;
		else if (l_Source == "TIMer")
			rInputTrig.Source = kInputTrigSrceTimer;
		else if (l_Source == "BUS")
			rInputTrig.Source = kInputTrigSrceBus;
		else if (l_Source == "EXTernal")
			rInputTrig.Source = kInputTrigSrceExternal;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Period
		rInputTrig.Period_s.Val = std::stod(std::string(l_Tokens[1].begin, l_Tokens[1].end));
		rInputTrig.Period_s.Min = std::stod(std::string(l_Tokens[2].begin, l_Tokens[2].end));
		rInputTrig.Period_s.Max = std::stod(std::string(l_Tokens[3].begin, l_Tokens[3].end));
		rInputTrig.Period_s.Def = std::stod(std::string(l_Tokens[4].begin, l_Tokens[4].end));
		// Trigger Count
		rInputTrig.Count.Val = std::stoi(std::string(l_Tokens[5].begin, l_Tokens[5].end));
		rInputTrig.Count.Min = std::stoi(std::string(l_Tokens[6].begin, l_Tokens[6].end));
		rInputTrig.Count.Max = std::stoi(std::string(l_Tokens[7].begin, l_Tokens[7].end));
		rInputTrig.Count.Def = std::stoi(std::string(l_Tokens[8].begin, l_Tokens[8].end));
		// Slope
		std::string l_Slope(l_Tokens[9].begin, l_Tokens[9].end);
		if (l_Slope == "POSitive")
			rInputTrig.Edge = kInputTrigEdgeRising;
		else if (l_Slope == "NEGative")
			rInputTrig.Edge = kInputTrigEdgeFalling;
		else if (l_Slope == "EITHer")
			rInputTrig.Edge = kInputTrigEdgeBoth;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmTriggerOutputSet(const sOutputTrigger<uParm> & rOutputTrig)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command;
	switch (rOutputTrig.Mode)
	{
	case kOutputTrigShapeOFF:
		l_Command = ":OUTP:TRIG:STAT OFF";
		break;
	case kOutputTrigShapeDCBGreaterThan:
		// Set trigerring conditions
		l_Command = ":TRIG:SEQuence2:LEV " + l_ToString(rOutputTrig.Level_UNITS);
		l_Command += ";:SLOP POS";
		// Set output signal parameters
		l_Command += ";:OUTP:TRIG:STAT ON";										// State
		l_Command += ";SHAP DC";												// Shape
		l_Command += (rOutputTrig.InvertedPolarity ? ";POL INV" : ";POL NORM");	// Polarity
																				// Pulse Width
		l_Command += ";DEL " + l_ToString(rOutputTrig.OutputDelay_s);			// Delay
		break;
	case kOutputTrigShapeDCBLowerThan:
		// Set trigerring conditions
		l_Command = ":TRIG:SEQuence2:LEV " + l_ToString(rOutputTrig.Level_UNITS);
		l_Command += ";:SLOP NEG";
		// Set output signal parameters
		l_Command += ";:OUTP:TRIG:STAT ON";										// State
		l_Command += ";SHAP DC";												// Shape
		l_Command += (rOutputTrig.InvertedPolarity ? ";POL INV" : ";POL NORM");	// Polarity
																				// Pulse Width
		l_Command += ";DEL " + l_ToString(rOutputTrig.OutputDelay_s);			// Delay
		break;
	case kOutputTrigShapePulseBRising:
		// Set trigerring conditions
		l_Command = ":TRIG:SEQuence2:LEV " + l_ToString(rOutputTrig.Level_UNITS);
		l_Command += ";:SLOP POS";
		// Set output signal parameters
		l_Command += ";:OUTP:TRIG:STAT ON";										// State
		l_Command += ";SHAP PULS";												// Shape
		l_Command += (rOutputTrig.InvertedPolarity ? ";POL INV" : ";POL NORM");	// Polarity
		l_Command += ";WIDT " + l_ToString(rOutputTrig.PulseWidth_s);			// Pulse Width
		l_Command += ";DEL " + l_ToString(rOutputTrig.OutputDelay_s);			// Delay
		break;
	case kOutputTrigShapePulseBFalling:
		// Set trigerring conditions
		l_Command = ":TRIG:SEQuence2:LEV " + l_ToString(rOutputTrig.Level_UNITS);
		l_Command += ";:SLOP NEG";
		// Set output signal parameters
		l_Command += ";:OUTP:TRIG:STAT ON";										// State
		l_Command += ";SHAP PULS";												// Shape
		l_Command += (rOutputTrig.InvertedPolarity ? ";POL INV" : ";POL NORM");	// Polarity
		l_Command += ";WIDT " + l_ToString(rOutputTrig.PulseWidth_s);			// Pulse Width
		l_Command += ";DEL " + l_ToString(rOutputTrig.OutputDelay_s);			// Delay
		break;
	case kOutputTrigShapePulseBCrossing:
		// Set trigerring conditions
		l_Command = ":TRIG:SEQuence2:LEV " + l_ToString(rOutputTrig.Level_UNITS);
		l_Command += ";:SLOP EITH";
		// Set output signal parameters
		l_Command += ";:OUTP:TRIG:STAT ON";										// State
		l_Command += ";SHAP PULS";												// Shape
		l_Command += (rOutputTrig.InvertedPolarity ? ";POL INV" : ";POL NORM");	// Polarity
		l_Command += ";WIDT " + l_ToString(rOutputTrig.PulseWidth_s);			// Pulse Width
		l_Command += ";DEL " + l_ToString(rOutputTrig.OutputDelay_s);			// Delay
		break;
	default:
		MTL_Assert(false);
		break;
	}

	return l_WriteAndRead(l_Command, m_ReadBuffer);
}
bool CPT2026Instrument::ParmTriggerOutputGet(sOutputTrigger<uParm> & rOutputTrig)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(	":OUTP:TRIG:STAT?"
								";POL?"
								";WIDT?"
								";DEL?"
								";SHAP?"
								";:TRIG:SEQuence2:SLOP?"
								";LEV?",
								m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		CVISABufferParser::tTokens l_Tokens = l_BP.Tokenize();
		if (l_Tokens.size() != 7)
			throw CPT2026InsException("Invalid number of tokens in answer", MTL__LOCATION__);

		// State
		std::string l_State(l_Tokens[0].begin, l_Tokens[0].end);
		enum {kInactive, kActive} l_eActive;
		if (l_State == "0")
			l_eActive = kInactive;
		else if (l_State == "1")
			l_eActive = kActive;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Polarity
		std::string l_Polarity(l_Tokens[1].begin, l_Tokens[1].end);
		if (l_Polarity == "INVerted")
			rOutputTrig.InvertedPolarity = true;
		else if (l_Polarity == "NORMal")
			rOutputTrig.InvertedPolarity = false;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Pulse Width
		rOutputTrig.PulseWidth_s = std::stod(std::string(l_Tokens[2].begin, l_Tokens[2].end));
		// Output Delay
		rOutputTrig.OutputDelay_s = std::stod(std::string(l_Tokens[3].begin, l_Tokens[3].end));
		// Shape
		std::string l_Shape(l_Tokens[4].begin, l_Tokens[4].end);
		enum { kDC, kPulse } l_eShape;
		if (l_Shape == "DC")
			l_eShape = kDC;
		else if (l_Shape == "PULSe")
			l_eShape = kPulse;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Slope
		std::string l_Slope(l_Tokens[5].begin, l_Tokens[5].end);
		enum { kRising, kFalling, kEither } l_eSlope;
		if (l_Slope == "POSitive")
			l_eSlope = kRising;
		else if (l_Slope == "NEGative")
			l_eSlope = kFalling;
		else if (l_Slope == "EITHer")
			l_eSlope = kEither;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Level
		rOutputTrig.Level_UNITS = std::stod(std::string(l_Tokens[6].begin, l_Tokens[6].end));
		// Mode
		if (l_eActive == kActive)
		{
			if (l_eShape == kDC)
			{
				if (l_eSlope == kRising || l_eSlope == kEither)
					rOutputTrig.Mode = kOutputTrigShapeDCBGreaterThan;
				else
					rOutputTrig.Mode = kOutputTrigShapeDCBLowerThan;
			}
			else
			{
				if (l_eSlope == kRising)
					rOutputTrig.Mode = kOutputTrigShapePulseBRising;
				else if (l_eSlope == kFalling)
					rOutputTrig.Mode = kOutputTrigShapePulseBFalling;
				else
					rOutputTrig.Mode = kOutputTrigShapePulseBCrossing;
			}
		}
		else
			rOutputTrig.Mode = kOutputTrigShapeOFF;
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmTriggerOutputGet(sOutputTrigger<sBoundedParm> & rOutputTrig)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(	":OUTP:TRIG:STAT?"
								";POL?"
								";WIDT?;WIDT? MIN;WIDT? MAX;WIDT? DEF"
								";DEL?;DEL? MIN;DEL? MAX;DEL? DEF"
								";SHAP?"
								";:TRIG:SEQuence2:SLOP?"
								";LEV?;LEV? MIN ;LEV? MAX;LEV? DEF",
								m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		CVISABufferParser::tTokens l_Tokens = l_BP.Tokenize();
		if (l_Tokens.size() != 16)
			throw CPT2026InsException("Invalid number of tokens in answer", MTL__LOCATION__);

		// State
		std::string l_State(l_Tokens[0].begin, l_Tokens[0].end);
		enum { kInactive, kActive } l_eActive;
		if (l_State == "0")
			l_eActive = kInactive;
		else if (l_State == "1")
			l_eActive = kActive;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Polarity
		std::string l_Polarity(l_Tokens[1].begin, l_Tokens[1].end);
		if (l_Polarity == "INVerted")
			rOutputTrig.InvertedPolarity = true;
		else if (l_Polarity == "NORMal")
			rOutputTrig.InvertedPolarity = false;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Pulse Width
		rOutputTrig.PulseWidth_s.Val = std::stod(std::string(l_Tokens[2].begin, l_Tokens[2].end));
		rOutputTrig.PulseWidth_s.Min = std::stod(std::string(l_Tokens[3].begin, l_Tokens[3].end));
		rOutputTrig.PulseWidth_s.Max = std::stod(std::string(l_Tokens[4].begin, l_Tokens[4].end));
		rOutputTrig.PulseWidth_s.Def = std::stod(std::string(l_Tokens[5].begin, l_Tokens[5].end));
		// Output Delay
		rOutputTrig.OutputDelay_s.Val = std::stod(std::string(l_Tokens[6].begin, l_Tokens[6].end));
		rOutputTrig.OutputDelay_s.Min = std::stod(std::string(l_Tokens[7].begin, l_Tokens[7].end));
		rOutputTrig.OutputDelay_s.Max = std::stod(std::string(l_Tokens[8].begin, l_Tokens[8].end));
		rOutputTrig.OutputDelay_s.Def = std::stod(std::string(l_Tokens[9].begin, l_Tokens[9].end));
		// Shape
		std::string l_Shape(l_Tokens[10].begin, l_Tokens[10].end);
		enum { kDC, kPulse } l_eShape;
		if (l_Shape == "DC")
			l_eShape = kDC;
		else if (l_Shape == "PULSe")
			l_eShape = kPulse;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Slope
		std::string l_Slope(l_Tokens[11].begin, l_Tokens[11].end);
		enum { kRising, kFalling, kEither } l_eSlope;
		if (l_Slope == "POSitive")
			l_eSlope = kRising;
		else if (l_Slope == "NEGative")
			l_eSlope = kFalling;
		else if (l_Slope == "EITHer")
			l_eSlope = kEither;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Level
		rOutputTrig.Level_UNITS.Val = std::stod(std::string(l_Tokens[12].begin, l_Tokens[12].end));
		rOutputTrig.Level_UNITS.Min = std::stod(std::string(l_Tokens[13].begin, l_Tokens[13].end));
		rOutputTrig.Level_UNITS.Max = std::stod(std::string(l_Tokens[14].begin, l_Tokens[14].end));
		rOutputTrig.Level_UNITS.Def = std::stod(std::string(l_Tokens[15].begin, l_Tokens[15].end));
		// Mode
		if (l_eActive == kActive)
		{
			if (l_eShape == kDC)
			{
				if (l_eSlope == kRising || l_eSlope == kEither)
					rOutputTrig.Mode = kOutputTrigShapeDCBGreaterThan;
				else
					rOutputTrig.Mode = kOutputTrigShapeDCBLowerThan;
			}
			else
			{
				if (l_eSlope == kRising)
					rOutputTrig.Mode = kOutputTrigShapePulseBRising;
				else if (l_eSlope == kFalling)
					rOutputTrig.Mode = kOutputTrigShapePulseBFalling;
				else
					rOutputTrig.Mode = kOutputTrigShapePulseBCrossing;
			}
		}
		else
			rOutputTrig.Mode = kOutputTrigShapeOFF;
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmRFPulseSet(const sPulse<uParm> & rPulse)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command;
	switch (rPulse.Mode)
	{
	case kPulseAuto:
		l_Command = ":SOUR:PULS:MODE AUTO";
		break;
	case kPulseManual:
		l_Command = ":SOUR:PULS:MODE MAN";
		l_Command += ";PER " + l_ToString(rPulse.Period_s);	// Period
		l_Command += ";WIDT " + l_ToString(rPulse.Width_s);	// Width
		break;
	default:
		MTL_Assert(false);
		break;
	}

	return l_WriteAndRead(l_Command, m_ReadBuffer);
}
bool CPT2026Instrument::ParmRFPulseGet(sPulse<uParm> & rPulse)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(	":SOUR:PULS:MODE?"
								";PER?"
								";WIDT?",
								m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		CVISABufferParser::tTokens l_Tokens = l_BP.Tokenize();
		if (l_Tokens.size() != 3)
			throw CPT2026InsException("Invalid number of tokens in answer", MTL__LOCATION__);

		// Mode
		std::string l_Mode(l_Tokens[0].begin, l_Tokens[0].end);
		if (l_Mode == "AUTO")
			rPulse.Mode = kPulseAuto;
		else if (l_Mode == "MANual")
			rPulse.Mode = kPulseManual;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Period
		rPulse.Period_s = std::stod(std::string(l_Tokens[1].begin, l_Tokens[1].end));
		// Width
		rPulse.Width_s = std::stod(std::string(l_Tokens[2].begin, l_Tokens[2].end));
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmRFPulseGet(sPulse<sBoundedParm> & rPulse)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(	":SOUR:PULS:MODE?"
								";PER?;PER? MIN;PER? MAX;PER? DEF"
								";WIDT?;WIDT? MIN;WIDT? MAX;WIDT? DEF",
								m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		CVISABufferParser::tTokens l_Tokens = l_BP.Tokenize();
		if (l_Tokens.size() != 9)
			throw CPT2026InsException("Invalid number of tokens in answer", MTL__LOCATION__);

		// Mode
		std::string l_Mode(l_Tokens[0].begin, l_Tokens[0].end);
		if (l_Mode == "AUTO")
			rPulse.Mode = kPulseAuto;
		else if (l_Mode == "MANual")
			rPulse.Mode = kPulseManual;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Period
		rPulse.Period_s.Val = std::stod(std::string(l_Tokens[1].begin, l_Tokens[1].end));
		rPulse.Period_s.Min = std::stod(std::string(l_Tokens[2].begin, l_Tokens[2].end));
		rPulse.Period_s.Max = std::stod(std::string(l_Tokens[3].begin, l_Tokens[3].end));
		rPulse.Period_s.Def = std::stod(std::string(l_Tokens[4].begin, l_Tokens[4].end));
		// Width
		rPulse.Width_s.Val = std::stod(std::string(l_Tokens[5].begin, l_Tokens[5].end));
		rPulse.Width_s.Min = std::stod(std::string(l_Tokens[6].begin, l_Tokens[6].end));
		rPulse.Width_s.Max = std::stod(std::string(l_Tokens[7].begin, l_Tokens[7].end));
		rPulse.Width_s.Def = std::stod(std::string(l_Tokens[8].begin, l_Tokens[8].end));
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmDigitizationSet(const sDigitization<uParm> & rDigitization)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command;
	switch (rDigitization.Mode)
	{
	case kDigitizationAuto:
		l_Command = ":SENS:SWE:MODE AUTO";										// Mode
		break;
	case kDigitizationManual:
		l_Command = ":SENS:SWE:MODE MAN";										// Mode
		l_Command += "TIME " + l_ToString(rDigitization.SweepTime_s);			// Sweep Time
		l_Command += "FREQ " + l_ToString(rDigitization.Rate_Hz);				// Dead Time// Rate
		l_Command += "OFFS:TIME " + l_ToString(rDigitization.DeadTime_s);		// Dead Time
		break;
	default:
		MTL_Assert(false);
		break;
	}

	return l_WriteAndRead(l_Command, m_ReadBuffer);
}
bool CPT2026Instrument::ParmDigitizationGet(sDigitization<uParm> & rDigitization)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(	":SENS:SWE:MODE?"
								";TIME?"
								";FREQ?"
								";OFFS:TIME?",
								m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		CVISABufferParser::tTokens l_Tokens = l_BP.Tokenize();
		if (l_Tokens.size() != 4)
			throw CPT2026InsException("Invalid number of tokens in answer", MTL__LOCATION__);

		// Mode
		std::string l_Mode(l_Tokens[0].begin, l_Tokens[0].end);
		if (l_Mode == "AUTO")
			rDigitization.Mode = kDigitizationAuto;
		else if (l_Mode == "MANual")
			rDigitization.Mode = kDigitizationManual;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Sweep Time
		rDigitization.SweepTime_s = std::stod(std::string(l_Tokens[1].begin, l_Tokens[1].end));
		// Digitization rate
		rDigitization.Rate_Hz = std::stod(std::string(l_Tokens[2].begin, l_Tokens[2].end));
		// Hold Off
		rDigitization.DeadTime_s = std::stod(std::string(l_Tokens[3].begin, l_Tokens[3].end));
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmDigitizationGet(sDigitization<sBoundedParm> & rDigitization)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(	":SENS:SWE:MODE?"
								";TIME?;TIME? MIN;TIME? MAX;TIME? DEF"
								";FREQ?;FREQ? MIN;FREQ? MAX;FREQ? DEF"
								";OFFS:TIME?;TIME? MIN;TIME? MAX;TIME? DEF",
								m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		CVISABufferParser::tTokens l_Tokens = l_BP.Tokenize();
		if (l_Tokens.size() != 13)
			throw CPT2026InsException("Invalid number of tokens in answer", MTL__LOCATION__);

		// Mode
		std::string l_Mode(l_Tokens[0].begin, l_Tokens[0].end);
		if (l_Mode == "AUTO")
			rDigitization.Mode = kDigitizationAuto;
		else if (l_Mode == "MANual")
			rDigitization.Mode = kDigitizationManual;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Sweep Time
		rDigitization.SweepTime_s.Val = std::stod(std::string(l_Tokens[1].begin, l_Tokens[1].end));
		rDigitization.SweepTime_s.Min = std::stod(std::string(l_Tokens[2].begin, l_Tokens[2].end));
		rDigitization.SweepTime_s.Max = std::stod(std::string(l_Tokens[3].begin, l_Tokens[3].end));
		rDigitization.SweepTime_s.Def = std::stod(std::string(l_Tokens[4].begin, l_Tokens[4].end));
		// Digitization rate
		rDigitization.Rate_Hz.Val = std::stod(std::string(l_Tokens[5].begin, l_Tokens[5].end));
		rDigitization.Rate_Hz.Min = std::stod(std::string(l_Tokens[6].begin, l_Tokens[6].end));
		rDigitization.Rate_Hz.Max = std::stod(std::string(l_Tokens[7].begin, l_Tokens[7].end));
		rDigitization.Rate_Hz.Def = std::stod(std::string(l_Tokens[8].begin, l_Tokens[8].end));
		// Hold Off
		rDigitization.DeadTime_s.Val = std::stod(std::string(l_Tokens[9].begin, l_Tokens[9].end));
		rDigitization.DeadTime_s.Min = std::stod(std::string(l_Tokens[10].begin, l_Tokens[10].end));
		rDigitization.DeadTime_s.Max = std::stod(std::string(l_Tokens[11].begin, l_Tokens[11].end));
		rDigitization.DeadTime_s.Def = std::stod(std::string(l_Tokens[12].begin, l_Tokens[12].end));
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmMatchingTuningSet(const sMatchingTuning<uParm> & rMatchingTuning)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command = ":CONF:PROB:MODE ";
	switch (rMatchingTuning.Mode)
	{
	case kMatchTunAuto:
		l_Command += "AUTO";
		break;
	case kMatchTunManual:
		l_Command += "MAN;MATC " + l_ToString(rMatchingTuning.Matching) + ";TUN " + l_ToString(rMatchingTuning.Tuning) + " V";
		break;
	default:
		MTL_Assert(false);
		break;
	}

	return l_WriteAndRead(l_Command, m_ReadBuffer);
}
bool CPT2026Instrument::ParmMatchingTuningGet(sMatchingTuning<uParm> & rMatchingTuning)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(	":CONF:PROB:MODE?"
								";MATC?"
								";TUN?",
								m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		CVISABufferParser::tTokens l_Tokens = l_BP.Tokenize();
		if (l_Tokens.size() != 3)
			throw CPT2026InsException("Invalid number of tokens in answer", MTL__LOCATION__);

		// Mode
		std::string l_Mode(l_Tokens[0].begin, l_Tokens[0].end);
		if (l_Mode == "AUTO")
			rMatchingTuning.Mode = kMatchTunAuto;
		else if (l_Mode == "MANual")
			rMatchingTuning.Mode = kMatchTunManual;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Matching
		rMatchingTuning.Matching = std::stod(std::string(l_Tokens[1].begin, l_Tokens[1].end));
		// Tuning
		rMatchingTuning.Tuning = std::stod(std::string(l_Tokens[2].begin, l_Tokens[2].end));
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmMatchingTuningGet(sMatchingTuning<sBoundedParm> & rMatchingTuning)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(	":CONF:PROB:MODE?"
								";MATC?;MATC? MIN;MATC? MAX;MATC? DEF"
								";TUN?;TUN? MIN;TUN? MAX;TUN? DEF",
								m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		CVISABufferParser::tTokens l_Tokens = l_BP.Tokenize();
		if (l_Tokens.size() != 9)
			throw CPT2026InsException("Invalid number of tokens in answer", MTL__LOCATION__);

		// Mode
		std::string l_Mode(l_Tokens[0].begin, l_Tokens[0].end);
		if (l_Mode == "AUTO")
			rMatchingTuning.Mode = kMatchTunAuto;
		else if (l_Mode == "MANual")
			rMatchingTuning.Mode = kMatchTunManual;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Matching
		rMatchingTuning.Matching.Val = std::stod(std::string(l_Tokens[1].begin, l_Tokens[1].end));
		rMatchingTuning.Matching.Min = std::stod(std::string(l_Tokens[2].begin, l_Tokens[2].end));
		rMatchingTuning.Matching.Max = std::stod(std::string(l_Tokens[3].begin, l_Tokens[3].end));
		rMatchingTuning.Matching.Def = std::stod(std::string(l_Tokens[4].begin, l_Tokens[4].end));
		// Tuning
		rMatchingTuning.Tuning.Val = std::stod(std::string(l_Tokens[5].begin, l_Tokens[5].end));
		rMatchingTuning.Tuning.Min = std::stod(std::string(l_Tokens[6].begin, l_Tokens[6].end));
		rMatchingTuning.Tuning.Max = std::stod(std::string(l_Tokens[7].begin, l_Tokens[7].end));
		rMatchingTuning.Tuning.Def = std::stod(std::string(l_Tokens[8].begin, l_Tokens[8].end));
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmUnitsSet(eUnits Units)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command = ":UNIT " + UnitsToString(Units);

	return l_WriteAndRead(l_Command, m_ReadBuffer);
}
bool CPT2026Instrument::ParmUnitsSetPPMRef(tPPMReference_UNITS Ref)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command = ":UNIT:PPMR " + l_ToString(Ref);

	return l_WriteAndRead(l_Command, m_ReadBuffer);
}
bool CPT2026Instrument::ParmUnitsGet(eUnits & rUnits, tPPMReference_UNITS & rRef)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(":UNIT?;:UNIT:PPMR?", m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		std::vector<char>::const_iterator l_tokenbeg, l_tokenend;
		// Units
		if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
			throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
		std::string l_Answer(l_tokenbeg, l_tokenend);
		if (!StringToUnits(l_Answer, rUnits))
			throw CPT2026InsException("Invalid response format", MTL__LOCATION__);
		// PPM Reference
		if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
			throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
		l_Answer.assign(l_tokenbeg, l_tokenend);
		rRef = std::stod(l_Answer);
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmRefClockSet(eReferenceClock RefClock)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command = ":INP:CLOC ";
	switch (RefClock)
	{
	case kRefClockInternal:
		l_Command += "INT";
		break;
	case kRefClockExternal:
		l_Command += "EXT";
		break;
	default:
		MTL_Assert(false);
		break;
	}

	return l_WriteAndRead(l_Command, m_ReadBuffer);
}
bool CPT2026Instrument::ParmRefClockGet(eReferenceClock & rRefClock)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(":INP:CLOC:SOUR?",m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		// Clock source
		std::string l_Answer(m_ReadBuffer.begin(), m_ReadBuffer.end());
		if (l_Answer == "INTernal")
			rRefClock = kRefClockInternal;
		else if (l_Answer == "EXTernal")
			rRefClock = kRefClockExternal;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmComUsbSet(bool Enabled)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command = std::string(":SYST:COMM:USBT:ENAB ") + (Enabled ? "1" : "0");
		if (!l_WriteAndRead(l_Command , m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmComUsbGet(bool & rEnabled)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(":SYST:COMM:USBT:ENAB?", m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		std::string l_Answer(m_ReadBuffer.begin(), m_ReadBuffer.end());
		if (l_Answer.empty())
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		else if (l_Answer[0] == '0')
			rEnabled = false;
		else if (l_Answer[0] == '1')
			rEnabled = true;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmComEthernetSet(const sEthernet & rSettings)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command(":SYST:COMM:ETH");
		l_Command += rSettings.Enabled ? ":ENAB 1" : ":ENAB 0";
		l_Command += ";ADDR " + uIPAddress_to_string(rSettings.IP);
		l_Command += ";MASK " + uIPAddress_to_string(rSettings.Mask);
		l_Command += ";DOM " + rSettings.Domain;
		l_Command += ";BRO " + uIPAddress_to_string(rSettings.Broadcast);
		l_Command += ";GAT " + uIPAddress_to_string(rSettings.Gateway);
		l_Command += ";DNSS " + uIPAddress_to_string(rSettings.DNSServer);
		l_Command += std::string(";NBNS ") + (rSettings.NBNSEnabled ? "1" : "0");
		l_Command += ";IDEN " + rSettings.Identifier;

		if (!l_WriteAndRead(l_Command, m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmComEthernetGet(sEthernet & rSettings)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(	":SYST:COMM:ETH:ENAB?"
								";ADDR?"
								";MASK?"
								";DOM?"
								";BRO?"
								";GAT?"
								";DNSS?"
								";NBNS?"
								";IDEN?"
								, m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		std::vector<char>::const_iterator l_tokenbeg, l_tokenend;
		// Enable
		if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
			throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
		std::string l_Enable(l_tokenbeg, l_tokenend);
		if (l_Enable.empty())
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		else if (l_Enable[0] == '0')
			rSettings.Enabled = false;
		else if (l_Enable[0] == '1')
			rSettings.Enabled = true;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// IP Address
		if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
			throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
		std::string l_IPAddress(l_tokenbeg, l_tokenend);
		std::match_results<std::string::const_iterator> l_IPmatch;
		if (!std::regex_match(l_IPAddress, l_IPmatch, std::regex("^([0-9]{1,3})\\.([0-9]{1,3})\\.([0-9]{1,3})\\.([0-9]{1,3})")))
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		rSettings.IP.Member.A = static_cast<U8>(std::stoul(l_IPmatch[1]));
		rSettings.IP.Member.B = static_cast<U8>(std::stoul(l_IPmatch[2]));
		rSettings.IP.Member.C = static_cast<U8>(std::stoul(l_IPmatch[3]));
		rSettings.IP.Member.D = static_cast<U8>(std::stoul(l_IPmatch[4]));
		// Mask
		if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
			throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
		std::string l_Mask(l_tokenbeg, l_tokenend);
		std::match_results<std::string::const_iterator> l_Maskmatch;
		if (!std::regex_match(l_Mask, l_Maskmatch, std::regex("^([0-9]{1,3})\\.([0-9]{1,3})\\.([0-9]{1,3})\\.([0-9]{1,3})")))
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		rSettings.Mask.Member.A = static_cast<U8>(std::stoul(l_Maskmatch[1]));
		rSettings.Mask.Member.B = static_cast<U8>(std::stoul(l_Maskmatch[2]));
		rSettings.Mask.Member.C = static_cast<U8>(std::stoul(l_Maskmatch[3]));
		rSettings.Mask.Member.D = static_cast<U8>(std::stoul(l_Maskmatch[4]));
		// Domain
		if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
			throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
		rSettings.Domain = std::string(l_tokenbeg, l_tokenend);
		// Broadcast
		if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
			throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
		std::string l_Broadcast(l_tokenbeg, l_tokenend);
		std::match_results<std::string::const_iterator> l_Broadcastmatch;
		if (!std::regex_match(l_Broadcast, l_Broadcastmatch, std::regex("^([0-9]{1,3})\\.([0-9]{1,3})\\.([0-9]{1,3})\\.([0-9]{1,3})")))
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		rSettings.Broadcast.Member.A = static_cast<U8>(std::stoul(l_Broadcastmatch[1]));
		rSettings.Broadcast.Member.B = static_cast<U8>(std::stoul(l_Broadcastmatch[2]));
		rSettings.Broadcast.Member.C = static_cast<U8>(std::stoul(l_Broadcastmatch[3]));
		rSettings.Broadcast.Member.D = static_cast<U8>(std::stoul(l_Broadcastmatch[4]));
		// Gateway
		if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
			throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
		std::string l_Gateway(l_tokenbeg, l_tokenend);
		std::match_results<std::string::const_iterator> l_Gatewaymatch;
		if (!std::regex_match(l_Gateway, l_Gatewaymatch, std::regex("^([0-9]{1,3})\\.([0-9]{1,3})\\.([0-9]{1,3})\\.([0-9]{1,3})")))
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		rSettings.Gateway.Member.A = static_cast<U8>(std::stoul(l_Gatewaymatch[1]));
		rSettings.Gateway.Member.B = static_cast<U8>(std::stoul(l_Gatewaymatch[2]));
		rSettings.Gateway.Member.C = static_cast<U8>(std::stoul(l_Gatewaymatch[3]));
		rSettings.Gateway.Member.D = static_cast<U8>(std::stoul(l_Gatewaymatch[4]));
		// DNS Server
		if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
			throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
		std::string l_DNS(l_tokenbeg, l_tokenend);
		std::match_results<std::string::const_iterator> l_DNSmatch;
		if (!std::regex_match(l_DNS, l_DNSmatch, std::regex("^([0-9]{1,3})\\.([0-9]{1,3})\\.([0-9]{1,3})\\.([0-9]{1,3})")))
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		rSettings.DNSServer.Member.A = static_cast<U8>(std::stoul(l_DNSmatch[1]));
		rSettings.DNSServer.Member.B = static_cast<U8>(std::stoul(l_DNSmatch[2]));
		rSettings.DNSServer.Member.C = static_cast<U8>(std::stoul(l_DNSmatch[3]));
		rSettings.DNSServer.Member.D = static_cast<U8>(std::stoul(l_DNSmatch[4]));
		// NBNS Enable
		if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
			throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
		std::string l_NBNSEnab(l_tokenbeg, l_tokenend);
		if (l_NBNSEnab.empty())
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		else if (l_NBNSEnab[0] == '0')
			rSettings.NBNSEnabled = false;
		else if (l_NBNSEnab[0] == '1')
			rSettings.NBNSEnabled = true;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// Identifier
		if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
			throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
		rSettings.Identifier = std::string(l_tokenbeg, l_tokenend);
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmComVXI11Set(bool Enabled)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command = std::string(":SYST:COMM:VXI11:ENAB ") + (Enabled ? "1" : "0");
		if (!l_WriteAndRead(l_Command, m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ParmComVXI11Get(bool & rEnabled)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(":SYST:COMM:VXI11:ENAB?", m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		std::string l_Answer(m_ReadBuffer.begin(), m_ReadBuffer.end());
		if (l_Answer.empty())
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		else if (l_Answer[0] == '0')
			rEnabled = false;
		else if (l_Answer[0] == '1')
			rEnabled = true;
		else
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}

//----------------------------------------------------------------------//
//							Settings / Files							//
//----------------------------------------------------------------------//
bool CPT2026Instrument::ReadFileDirectory(U32 & rUsedBytes, U32 & rAvailableBytes, tFileList & rFileList)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	rFileList.clear();

	try {
		if (!l_WriteAndRead(":MMEM:CAT?", m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		// Parse results
		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		std::vector<char>::const_iterator l_tokenbeg, l_tokenend;

		// Used Bytes
		if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ','))
			throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
		rUsedBytes = static_cast<U32>(std::stoul(std::string(l_tokenbeg, l_tokenend)));
		// Available Bytes
		if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ','))
			throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
		rAvailableBytes = static_cast<U32>(std::stoul(std::string(l_tokenbeg, l_tokenend)));
		// File List
		while (l_BP.GetNext(l_tokenbeg, l_tokenend, ','))
		{
			sFile l_File;
			// Path
			l_File.Path = std::string(l_tokenbeg, l_tokenend);
			// Type
			if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ','))
				throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
			l_File.Type = std::string(l_tokenbeg, l_tokenend);
			// Size
			if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ','))
				throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
			l_File.Size = std::stoul(std::string(l_tokenbeg, l_tokenend));
			// Add file to the list
			rFileList.push_back(l_File);
		}
	} // try
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ReadFile(std::string Path, std::string & rContent)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command;
		l_Command = ":MMEM:DATA? " "\"" + Path + "\"";
		if (!l_WriteAndRead(l_Command, m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		rContent = std::string(m_ReadBuffer.begin(), m_ReadBuffer.end());
	} // try
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::WriteFile(std::string Path, const std::string & rContent)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_BinaryFileContent;
		ToArbitraryBlock(rContent, l_BinaryFileContent);
		std::string l_Command;
		l_Command = ":MMEM:DATA " "\"" + Path + "\"" "," + l_BinaryFileContent;
		if (!l_WriteAndRead(l_Command, m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	} // try
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::DeleteFile(std::string Path)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command;
		l_Command = ":MMEM:DEL " "\"" + Path + "\"";
		if (!l_WriteAndRead(l_Command, m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	} // try
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::GetSettingFileList(std::vector<std::string> & rSettingFileList)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	rSettingFileList.clear();

	try {
		U32 l_Dummy;
		tFileList l_FileList;
		if (!ReadFileDirectory(l_Dummy, l_Dummy, l_FileList))
			throw CPT2026InsException("Could not retrieve file list", MTL__LOCATION__);
		// For each file, check if it is a setting
		for (tFileList::const_iterator it = l_FileList.begin(); it != l_FileList.end(); it++)
		{
			// Match a string that starts with "/yaffs/settings/"
			if (std::regex_match(it->Path, std::regex("^/yaffs/settings/.+")))
				rSettingFileList.push_back(it->Path);
		}
	} // try
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::SaveSettings(eSettingType Type, std::string Name)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command;
		l_Command = ":MMEM:STOR 0" ",\"" + Name + "\"";
		switch (Type)
		{
		case kAll:
			l_Command += ",ALL";
			break;
		case kCommunication:
			l_Command += ",COMM";
			break;
		case kMeasure:
			l_Command += ",MEAS";
			break;
		case kTriggers:
			l_Command += ",TRIG";
			break;
		case kReferenceClock:
			l_Command += ",REFC";
			break;
		case kUnits:
			l_Command += ",UNIT";
			break;
		default:
			MTL_Assert("Unhandled setting type");
			throw CPT2026InsException(std::string("Unhandled setting type in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
			break;
		}
		if (!l_WriteAndRead(l_Command, m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	} // try
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::RecallSettings(std::string Name)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command;
		l_Command = ":MMEM:LOAD 0," "\"" + Name + "\"";
		if (!l_WriteAndRead(l_Command, m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	} // try
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}

//----------------------------------------------------------------------//
//						Initiate / Abort / Trigger						//
//----------------------------------------------------------------------//
bool CPT2026Instrument::Initiate(bool Continuous)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command = Continuous ? ":INIT:CONT 1" : ":INIT:CONT 0;:INIT";
	return l_WriteAndRead(l_Command, m_ReadBuffer);
}

bool CPT2026Instrument::Abort(bool CancelContinuous)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command = CancelContinuous ? ":INIT:CONT 0;:ABOR" : ":ABOR";
	return l_WriteAndRead(l_Command, m_ReadBuffer);
}

bool CPT2026Instrument::SendBusTrigger()
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	return AssertTrigger(eTriggerProtocol::Default);
}

bool CPT2026Instrument::ForceOutputTrigger()
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command = ":OUTP:TRIG:IMM";
	return l_WriteAndRead(l_Command, m_ReadBuffer);
}

//----------------------------------------------------------------------//
//							Measurements / Data							//
//----------------------------------------------------------------------//
bool CPT2026Instrument::MeasurementsGet(U32 NoMeasurements, std::vector<tFlux> & rFlux, eUnits & rUnits)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		std::string l_Command = ":UNIT?;:FETC:ARR:FLUX? " + std::to_string(NoMeasurements);
		if (!l_WriteAndRead(l_Command, m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		std::vector<char>::const_iterator l_tokenbeg, l_tokenend;
		// Units
		if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
			throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
		if (!StringToUnits(std::string(l_tokenbeg, l_tokenend), rUnits))
			throw CPT2026InsException("Could not parse units", MTL__LOCATION__);
		// Flux
		if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
			throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
		size_t l_Start, l_Len;
		if (IsArbitraryBlock(l_tokenbeg, l_BP.end(), l_Start, l_Len))
		{
			for (auto l_it = l_tokenbeg + l_Start; l_it < l_tokenbeg + l_Start + l_Len; l_it += sizeof(F64))
				rFlux.push_back(BinaryToF64(&*l_it));
		}
		else
		{
			std::vector<char>::const_iterator fluxbeg, fluxend;
			CVISABufferParser l_FluxParser(l_tokenbeg, l_tokenend);
			bool l_GetNextRet;
			for (l_GetNextRet = l_FluxParser.GetNext(fluxbeg, fluxend, ','); l_GetNextRet && fluxbeg != l_tokenend; l_GetNextRet = l_FluxParser.GetNext(fluxbeg, fluxend, ','))
				rFlux.push_back(std::stof(std::string(fluxbeg, fluxend)));
		}

		if (rFlux.size() != NoMeasurements)
			throw CPT2026InsException("Parsed size doesn't match required size", MTL__LOCATION__);
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}

bool CPT2026Instrument::MeasurementsGet(U32 NoMeasurements, sArbitraryMeasurements ArbSelect, sAllMeasurements & rMeas, eUnits & rUnits)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	rMeas.clear();
	try
	{
		// Build command
		std::string l_Command = ":UNIT?";
		if (ArbSelect.Flux)					l_Command += ";:FETC:ARR:FLUX? " + std::to_string(NoMeasurements);
		if (ArbSelect.Deviation)			l_Command += ";:FETC:ARR:SIGM? " + std::to_string(NoMeasurements);
		if (ArbSelect.Uniformity)			l_Command += ";:FETC:ARR:UNIF? " + std::to_string(NoMeasurements);
		if (ArbSelect.Channel)				l_Command += ";:FETC:ARR:CHAN? " + std::to_string(NoMeasurements);
		if (ArbSelect.Timestamp)			l_Command += ";:FETC:ARR:TIM? " + std::to_string(NoMeasurements);
		if (ArbSelect.Status)				l_Command += ";:FETC:ARR:STAT? " + std::to_string(NoMeasurements);
		if (ArbSelect.NbValidMeas)			l_Command += ";:FETC:ARR:VLM? " + std::to_string(NoMeasurements);
		if (ArbSelect.IF)					l_Command += ";:FETC:SCAL:IFR?";
		if (ArbSelect.RelaxationTime)		l_Command += ";:FETC:SCAL:REL?";
		if (ArbSelect.RadioFrequency)		l_Command += ";:FETC:SCAL:RFFR?";
		if (ArbSelect.NoNMRSamples)			l_Command += ";:FETC:ARR:NMRS? " + std::to_string(ArbSelect.NoNMRSamples);
		if (ArbSelect.NoFFTSamples)			l_Command += ";:FETC:ARR:FFTB? " + std::to_string(ArbSelect.NoFFTSamples);
		if (ArbSelect.NoSpectrumSamples)	l_Command += ";:FETC:ARR:SPEC? " + std::to_string(ArbSelect.NoSpectrumSamples);
		if (ArbSelect.PolynomDegree)		l_Command += ";:FETC:ARR:FIT? " + std::to_string(ArbSelect.PolynomDegree);

		if (!l_WriteAndRead(l_Command, m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		// Parse results
		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		std::vector<char>::const_iterator l_tokenbeg, l_tokenend;
		// Units
		if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
			throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
		if (!StringToUnits(std::string(l_tokenbeg, l_tokenend), rUnits))
			throw CPT2026InsException("Could not parse units", MTL__LOCATION__);

		// Reserve memory for the new basic results (if any)
		if (ArbSelect.Flux ||
			ArbSelect.Deviation ||
			ArbSelect.Uniformity ||
			ArbSelect.Channel ||
			ArbSelect.Timestamp ||
			ArbSelect.Status ||
			ArbSelect.NbValidMeas)
		rMeas.AdvMeas.Basic.resize(NoMeasurements);

		// Get basic results (if any)
		if (ArbSelect.Flux)
		{
			if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
				throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
			size_t Start, Len;
			if (IsArbitraryBlock(l_tokenbeg, l_BP.end(), Start, Len))
			{
				l_BP.SetNextOffset(l_tokenbeg + Start + Len + 1);	// Set next offset after the end of the arbitrary block
				for (size_t index = 0; Len >= sizeof(F64); Len -= sizeof(F64), Start += sizeof(F64), index++)
					rMeas.AdvMeas.Basic[index].Flux_UNITS = BinaryToF64(&*l_tokenbeg + Start);
			}
			else
			{
				std::vector<char>::const_iterator fluxbeg, fluxend;
				CVISABufferParser l_FluxParser(l_tokenbeg, l_tokenend);
				if (!l_FluxParser.GetNext(fluxbeg, fluxend, ','))
					throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
				bool l_GetNextRet = true;
				for (size_t index = 0; l_GetNextRet && fluxbeg != l_tokenend; l_GetNextRet = l_FluxParser.GetNext(fluxbeg, fluxend, ','), index++)
					rMeas.AdvMeas.Basic[index].Flux_UNITS = std::stod(std::string(fluxbeg, fluxend));
			}
		}
		if (ArbSelect.Deviation)
		{
			if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
				throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
			size_t Start, Len;
			if (IsArbitraryBlock(l_tokenbeg, l_BP.end(), Start, Len))
			{
				l_BP.SetNextOffset(l_tokenbeg + Start + Len + 1);	// Set next offset after the end of the arbitrary block
				for (size_t index = 0; Len >= sizeof(F32); Len -= sizeof(F32), Start += sizeof(F32), index++)
					rMeas.AdvMeas.Basic[index].Deviation_ppm = BinaryToF32(&*l_tokenbeg + Start);
			}
			else
			{
				std::vector<char>::const_iterator devbeg, devend;
				CVISABufferParser l_DevParser(l_tokenbeg, l_tokenend);
				if (!l_DevParser.GetNext(devbeg, devend, ','))
					throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
				bool l_GetNextRet = true;
				for (size_t index = 0; l_GetNextRet && devbeg != l_tokenend; l_GetNextRet = l_DevParser.GetNext(devbeg, devend, ','), index++)
					rMeas.AdvMeas.Basic[index].Deviation_ppm = std::stof(std::string(devbeg, devend));
			}
		}
		if (ArbSelect.Uniformity)
		{
			if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
				throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
			size_t Start, Len;
			if (IsArbitraryBlock(l_tokenbeg, l_BP.end(), Start, Len))
			{
				l_BP.SetNextOffset(l_tokenbeg + Start + Len + 1);	// Set next offset after the end of the arbitrary block
				for (size_t index = 0; Len >= sizeof(F32); Len -= sizeof(F32), Start += sizeof(F32), index++)
					rMeas.AdvMeas.Basic[index].Uniformity = BinaryToF32(&*l_tokenbeg + Start);
			}
			else
			{
				std::vector<char>::const_iterator unifbeg, unifend;
				CVISABufferParser l_UnifParser(l_tokenbeg, l_tokenend);
				if (!l_UnifParser.GetNext(unifbeg, unifend, ','))
					throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
				bool l_GetNextRet = true;
				for (size_t index = 0; l_GetNextRet && unifbeg != l_tokenend; l_GetNextRet = l_UnifParser.GetNext(unifbeg, unifend, ','), index++)
					rMeas.AdvMeas.Basic[index].Uniformity = std::stof(std::string(unifbeg, unifend));

			}
		}
		if (ArbSelect.Channel)
		{
			if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
				throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
			size_t Start, Len;
			tChannelList l_ChanList;
			if (IsArbitraryBlock(l_tokenbeg, l_BP.end(), Start, Len))
				FromBinaryChannelList(&*l_tokenbeg + Start, Len, l_ChanList);
			else
				FromStringChannelList(&*l_tokenbeg, &*l_tokenend, l_ChanList);

			for (size_t index = 0; index < l_ChanList.size() && index < rMeas.AdvMeas.Basic.size(); index++)
				rMeas.AdvMeas.Basic[index].Channel = l_ChanList[index];
		}
		if (ArbSelect.Timestamp)
		{
			if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
				throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
			size_t Start, Len;
			if (IsArbitraryBlock(l_tokenbeg, l_BP.end(), Start, Len))
			{
				l_BP.SetNextOffset(l_tokenbeg + Start + Len + 1);	// Set next offset after the end of the arbitrary block
				for (size_t index = 0; Len >= sizeof(U64); Len -= sizeof(U64), Start += sizeof(U64), index++)
					rMeas.AdvMeas.Basic[index].Timestamp = BinaryToU64(&*l_tokenbeg + Start);
			}
			else
			{
				std::vector<char>::const_iterator timbeg, timend;
				CVISABufferParser l_TimParser(l_tokenbeg, l_tokenend);
				if (!l_TimParser.GetNext(timbeg, timend, ','))
					throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
				bool l_GetNextRet = true;
				for (size_t index = 0; l_GetNextRet && timbeg != l_tokenend; l_GetNextRet = l_TimParser.GetNext(timbeg, timend, ','), index++)
					rMeas.AdvMeas.Basic[index].Timestamp = std::stoull(std::string(timbeg, timend));
			}
		}
		if (ArbSelect.Status)
		{
			if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
				throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
			size_t Start, Len;
			if (IsArbitraryBlock(l_tokenbeg, l_BP.end(), Start, Len))
			{
				l_BP.SetNextOffset(l_tokenbeg + Start + Len + 1);	// Set next offset after the end of the arbitrary block
				for (size_t index = 0; Len >= sizeof(U8); Len -= sizeof(U8), Start += sizeof(U8), index++)
					rMeas.AdvMeas.Basic[index].Status.StatusByte = *(&*l_tokenbeg + Start);
			}
			else
			{
				std::vector<char>::const_iterator statbeg, statend;
				CVISABufferParser l_StatParser(l_tokenbeg, l_tokenend);
				if (!l_StatParser.GetNext(statbeg, statend, ','))
					throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
				bool l_GetNextRet = true;
				for (size_t index = 0; l_GetNextRet && statbeg != l_tokenend; l_GetNextRet = l_StatParser.GetNext(statbeg, statend, ','), index++)
					rMeas.AdvMeas.Basic[index].Status.StatusByte = static_cast<U8>(std::stoul(std::string(statbeg, statend)));
			}
		}
		if (ArbSelect.NbValidMeas)
		{
			if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
				throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
			size_t Start, Len;
			if (IsArbitraryBlock(l_tokenbeg, l_BP.end(), Start, Len))
			{
				l_BP.SetNextOffset(l_tokenbeg + Start + Len + 1);	// Set next offset after the end of the arbitrary block
				for (size_t index = 0; Len >= sizeof(U32); Len -= sizeof(U32), Start += sizeof(U32), index++)
					rMeas.AdvMeas.Basic[index].NbValidMeas = BinaryToU32(&*l_tokenbeg + Start);
			}
			else
			{
				std::vector<char>::const_iterator timbeg, timend;
				CVISABufferParser l_NbValidParser(l_tokenbeg, l_tokenend);
				if (!l_NbValidParser.GetNext(timbeg, timend, ','))
					throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
				bool l_GetNextRet = true;
				for (size_t index = 0; l_GetNextRet && timbeg != l_tokenend; l_GetNextRet = l_NbValidParser.GetNext(timbeg, timend, ','), index++)
					rMeas.AdvMeas.Basic[index].NbValidMeas = std::stoul(std::string(timbeg, timend));
			}
		}
		if (ArbSelect.IF)
		{
			if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
				throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
			size_t Start, Len;
			if (IsArbitraryBlock(l_tokenbeg, l_BP.end(), Start, Len))
			{
				l_BP.SetNextOffset(l_tokenbeg + Start + Len + 1);	// Set next offset after the end of the arbitrary block
				rMeas.AdvMeas.IntermediateFrequency_Hz = BinaryToF64(&*l_tokenbeg + Start);
			}
			else
			{
				rMeas.AdvMeas.IntermediateFrequency_Hz = std::stod(std::string(l_tokenbeg, l_tokenend));
			}
		}
		if (ArbSelect.RelaxationTime)
		{
			if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
				throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
			size_t Start, Len;
			if (IsArbitraryBlock(l_tokenbeg, l_BP.end(), Start, Len))
			{
				l_BP.SetNextOffset(l_tokenbeg + Start + Len + 1);	// Set next offset after the end of the arbitrary block
				rMeas.AdvMeas.RelaxationTime_s = BinaryToF32(&*l_tokenbeg + Start);
			}
			else
			{
				rMeas.AdvMeas.RelaxationTime_s = std::stof(std::string(l_tokenbeg, l_tokenend));
			}
		}
		if (ArbSelect.RadioFrequency)
		{
			if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
				throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
			size_t Start, Len;
			if (IsArbitraryBlock(l_tokenbeg, l_BP.end(), Start, Len))
			{
				l_BP.SetNextOffset(l_tokenbeg + Start + Len + 1);	// Set next offset after the end of the arbitrary block
				rMeas.AdvMeas.RadioFrequency_Hz = BinaryToF64(&*l_tokenbeg + Start);
			}
			else
			{
				rMeas.AdvMeas.RadioFrequency_Hz = std::stod(std::string(l_tokenbeg, l_tokenend));
			}
		}
		if (ArbSelect.NoNMRSamples)
		{
			if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
				throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
			size_t Start, Len;
			if (IsArbitraryBlock(l_tokenbeg, l_BP.end(), Start, Len))
			{
				l_BP.SetNextOffset(l_tokenbeg + Start + Len + 1);	// Set next offset after the end of the arbitrary block
				// Get Sample Period
				rMeas.NMR.SamplePeriod_s = BinaryToF32(&*l_tokenbeg + Start);
				std::vector<char>::const_iterator NmrSampbeg = l_tokenbeg + Start + sizeof(rMeas.NMR.SamplePeriod_s);
				size_t NmrSampLen = Len - sizeof(rMeas.NMR.SamplePeriod_s);
				size_t NmrSampStart = 0;
				// Get Samples
				for (size_t index = 0; NmrSampLen >= sizeof(F32); NmrSampLen -= sizeof(F32), NmrSampStart += sizeof(F32), index++)
					rMeas.NMR.Samples_V.push_back(BinaryToF32(&*NmrSampbeg + NmrSampStart));
			}
			else
			{
				std::vector<char>::const_iterator nmrbeg, nmrend;
				CVISABufferParser l_NmrParser(l_tokenbeg, l_tokenend);
				if (!l_NmrParser.GetNext(nmrbeg, nmrend, ','))
					throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
				bool l_GetNextRes = true;
				for (size_t index = 0; l_GetNextRes && nmrbeg != l_tokenend; l_GetNextRes = l_NmrParser.GetNext(nmrbeg, nmrend, ','), index++)
					rMeas.NMR.Samples_V.push_back(std::stof(std::string(nmrbeg, nmrend)));
			}
		}
		if (ArbSelect.NoFFTSamples)
		{
			if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
				throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
			size_t Start, Len;
			if (IsArbitraryBlock(l_tokenbeg, l_BP.end(), Start, Len))
			{
				l_BP.SetNextOffset(l_tokenbeg + Start + Len + 1);	// Set next offset after the end of the arbitrary block
				// Get Frequency Resolution
				rMeas.FFT.FrequencyResolution_Hz = BinaryToF32(&*l_tokenbeg + Start);
				std::vector<char>::const_iterator FFTSampbeg = l_tokenbeg + Start + sizeof(rMeas.FFT.FrequencyResolution_Hz);
				size_t FFTSampLen = Len - sizeof(rMeas.FFT.FrequencyResolution_Hz);
				size_t FFTSampStart = 0;
				// Get Samples
				for (size_t index = 0; FFTSampLen >= sizeof(F32); FFTSampLen -= sizeof(F32), FFTSampStart += sizeof(F32), index++)
					rMeas.FFT.Samples_V.push_back(BinaryToF32(&*FFTSampbeg + FFTSampStart));
			}
			else
			{
				std::vector<char>::const_iterator fftbeg, fftend;
				CVISABufferParser l_FftParser(l_tokenbeg, l_tokenend);
				if (!l_FftParser.GetNext(fftbeg, fftend, ','))
				throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
				bool l_GetNextRes = true;
				for (size_t index = 0; l_GetNextRes && fftbeg != l_tokenend; l_GetNextRes = l_FftParser.GetNext(fftbeg, fftend, ','), index++)
					rMeas.FFT.Samples_V.push_back(std::stof(std::string(fftbeg, fftend)));
			}
		}
		if (ArbSelect.NoSpectrumSamples)
		{
			if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
				throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
			size_t Start, Len;
			if (IsArbitraryBlock(l_tokenbeg, l_BP.end(), Start, Len))
			{
				l_BP.SetNextOffset(l_tokenbeg + Start + Len + 1);	// Set next offset after the end of the arbitrary block
				// Get Center Frequency and Frequency Resolution
				rMeas.Spectrum.CenterFrequency_Hz = BinaryToF32(&*l_tokenbeg + Start);
				rMeas.Spectrum.FrequencyResolution_Hz = BinaryToF32(&*l_tokenbeg + Start + sizeof(rMeas.Spectrum.CenterFrequency_Hz));
				std::vector<char>::const_iterator SpecSampbeg = l_tokenbeg + Start + sizeof(rMeas.Spectrum.CenterFrequency_Hz) + sizeof(rMeas.Spectrum.FrequencyResolution_Hz);
				size_t SpecSampLen = Len - sizeof(rMeas.Spectrum.CenterFrequency_Hz) + sizeof(rMeas.Spectrum.FrequencyResolution_Hz);
				size_t SpecSampStart = 0;
				// Get Samples
				for (size_t index = 0; SpecSampLen >= sizeof(F32); SpecSampLen -= sizeof(F32), SpecSampStart += sizeof(F32), index++)
					rMeas.Spectrum.Samples_V.push_back(BinaryToF32(&*SpecSampbeg + SpecSampStart));
			}
			else
			{
				std::vector<char>::const_iterator specbeg, specend;
				CVISABufferParser l_SpecParser(l_tokenbeg, l_tokenend);
				if (!l_SpecParser.GetNext(specbeg, specend, ','))
					throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
				bool l_GetNextRes = true;
				for (size_t index = 0; l_GetNextRes && specbeg != l_tokenend; l_GetNextRes = l_SpecParser.GetNext(specbeg, specend, ','), index++)
					rMeas.Spectrum.Samples_V.push_back(std::stof(std::string(specbeg, specend)));
			}
		}
		if (ArbSelect.PolynomDegree)
		{
			if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
				throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
			size_t Start, Len;
			if (IsArbitraryBlock(l_tokenbeg, l_BP.end(), Start, Len))
			{
				l_BP.SetNextOffset(l_tokenbeg + Start + Len + 1);	// Set next offset after the end of the arbitrary block
				// Get Residue
				rMeas.FitPolynom.Residue = BinaryToF32(&*l_tokenbeg + Start);
				std::vector<char>::const_iterator PolyCoefbeg = l_tokenbeg + Start + sizeof(rMeas.FitPolynom.Residue);
				size_t PolyCoefLen = Len - sizeof(rMeas.FitPolynom.Residue);
				size_t PolyCoefStart = 0;
				// Get Samples
				for (size_t index = 0; PolyCoefLen >= sizeof(F32); PolyCoefLen -= sizeof(F32), PolyCoefStart += sizeof(F32), index++)
					rMeas.FitPolynom.Coeffs.push_back(BinaryToF32(&*PolyCoefbeg + PolyCoefStart));
			}
			else
			{
				std::vector<char>::const_iterator polybeg, polyend;
				CVISABufferParser l_PolyParser(l_tokenbeg, l_tokenend);
				if (!l_PolyParser.GetNext(polybeg, polyend, ','))
					throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
				rMeas.FitPolynom.Residue = std::stof(std::string(polybeg, polyend));
				if (!l_PolyParser.GetNext(polybeg, polyend, ','))
					throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
				bool l_GetNextRes = true;
				for (size_t index = 0; l_GetNextRes && polybeg != l_tokenend; l_GetNextRes = l_PolyParser.GetNext(polybeg, polyend, ','), index++)
					rMeas.FitPolynom.Coeffs.push_back(std::stof(std::string(polybeg, polyend)));
			}
		}
	} // try
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}

bool CPT2026Instrument::GetSearchProgress(U8 & rSPR, F64 & rHallMeasurement_UNITS)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(":FETC:SPR?", m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		size_t l_Start, l_Len;
		if (IsArbitraryBlock(m_ReadBuffer.begin(), m_ReadBuffer.end(), l_Start, l_Len))
		{
			rSPR = *(m_ReadBuffer.data() + l_Start);
			rHallMeasurement_UNITS = BinaryToF64(m_ReadBuffer.data() + l_Start + 1);
		}
		else
		{
			// Parse results
			CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
			std::vector<char>::const_iterator l_tokenbeg, l_tokenend;
			// Search Progress
			if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ','))
				throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
			rSPR = std::stoi(std::string(l_tokenbeg, l_tokenend));
			// Hall Measurement
			if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ','))
				throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
			rHallMeasurement_UNITS = std::stod(std::string(l_tokenbeg, l_tokenend));
		}
	}
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}

bool CPT2026Instrument::SetFormat(eCommunicationFormat Format)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command;

	switch (Format)
	{
	case kComFormatAscii:
		l_Command += ":FORM:DATA ASC";
		break;
	case kComFormatInteger:
		l_Command += ":FORM:DATA INT";
		break;
	default:
		MTL_Assert(false);
		break;
	}
	return l_WriteAndRead(l_Command, m_ReadBuffer);
}

//----------------------------------------------------------------------//
//							Status Handling								//
//----------------------------------------------------------------------//
bool CPT2026Instrument::StatusPreset()
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	return l_WriteAndRead(":STAT:PRES", m_ReadBuffer);
}

bool CPT2026Instrument::StatusGet(sStatusRegister Reg, U16 & rStatus)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command;

	// Prepare register set
	switch (Reg.Set)
	{
	case kStatusQestionable:
		l_Command += ":STAT:QUES";
		break;
	case kStatusOperation:
		l_Command += ":STAT:OPER";
		break;
	case kStatusOperationBit11:
		l_Command += ":STAT:OPER:BIT11";
		break;
	case kStatusOperationBit12:
		l_Command += ":STAT:OPER:BIT12";
		break;
	case kStatusQuestionableBit12:
		l_Command += ":STAT:QUES:BIT12";
		break;
	default:
		MTL_Assert(false);
		break;
	}
	// Prepare register type
	switch (Reg.Type)
	{
	case kStatusEvent:
		l_Command += ":EVEN?";
		break;
	case kStatusCondition:
		l_Command += ":COND?";
		break;
	case kStatusEnable:
		l_Command += ":ENAB?";
		break;
	default:
		MTL_Assert(false);
		break;
	}

	try {
		if (!l_WriteAndRead(l_Command, m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		rStatus = std::stoi(std::string(m_ReadBuffer.begin(), m_ReadBuffer.end()));
	} // try
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		rStatus = 0;
		return false;
	}
	return true;
}

bool CPT2026Instrument::StatusGet(std::vector<sStatusRegister> Regs, std::vector<U16> & rStatus)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	rStatus.clear();
	std::string l_Command;

	for (std::vector<sStatusRegister>::const_iterator it = Regs.begin(); it != Regs.end(); it++)
	{
		if (it != Regs.begin())
			l_Command += ";";
		switch (it->Set)
		{
		case kStatusQestionable:
			l_Command += ":STAT:QUES";
			break;
		case kStatusOperation:
			l_Command += ":STAT:OPER";
			break;
		case kStatusOperationBit11:
			l_Command += ":STAT:OPER:BIT11";
			break;
		case kStatusOperationBit12:
			l_Command += ":STAT:OPER:BIT12";
			break;
		case kStatusQuestionableBit12:
			l_Command += ":STAT:QUES:BIT12";
			break;
		default:
			MTL_Assert(false);
			break;
		}
		switch (it->Type)
		{
		case kStatusEvent:
			l_Command += ":EVEN?";
			break;
		case kStatusCondition:
			l_Command += ":COND?";
			break;
		case kStatusEnable:
			l_Command += ":ENAB?";
			break;
		default:
			MTL_Assert(false);
			break;
		}
	}

	try {
		if (!l_WriteAndRead(l_Command, m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		// Parse results
		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		std::vector<char>::const_iterator l_tokenbeg, l_tokenend;

		for (size_t i = Regs.size(); i > 0; i--)
		{
			if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
				throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
			rStatus.push_back(std::stoi(std::string(l_tokenbeg, l_tokenend)));
		}
	} // try
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}

bool CPT2026Instrument::StatusSetEnableRegister(eStatusRegisterSet Set, U16 DisableMask, U16 EnableMask)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_RegSet, l_Command;

	switch (Set)
	{
	case kStatusQestionable:
		l_RegSet += ":STAT:QUES";
		break;
	case kStatusOperation:
		l_RegSet += ":STAT:OPER";
		break;
	case kStatusOperationBit11:
		l_RegSet += ":STAT:OPER:BIT11";
		break;
	case kStatusOperationBit12:
		l_RegSet += ":STAT:OPER:BIT12";
		break;
	case kStatusQuestionableBit12:
		l_RegSet += ":STAT:QUES:BIT12";
		break;
	default:
		MTL_Assert(false);
		break;
	}

	try {
		// Read current enable register
		l_Command = l_RegSet + ":ENAB?";
		if (!l_WriteAndRead(l_Command, m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		U16 l_StatusEnab = static_cast<U16>(std::stoi(std::string(m_ReadBuffer.begin(), m_ReadBuffer.end())));

		// Compute new enable register
		l_StatusEnab |= EnableMask;
		l_StatusEnab &= ~DisableMask;

		// Set new enable register
		l_Command = l_RegSet + ":ENAB" + " " + std::to_string(l_StatusEnab);
		if (!l_WriteAndRead(l_Command, m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	} // try
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}

bool CPT2026Instrument::StatusSetTransitionFilter(eStatusRegisterSet Set, eTransition Transition, U16 ClearMask, U16 SetMask)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(
		FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_RegSet, l_Command;

	switch (Set)
	{
	case kStatusQestionable:
		l_RegSet += ":STAT:QUES";
		break;
	case kStatusOperation:
		l_RegSet += ":STAT:OPER";
		break;
	case kStatusOperationBit11:
		l_RegSet += ":STAT:OPER:BIT11";
		break;
	case kStatusOperationBit12:
		l_RegSet += ":STAT:OPER:BIT12";
		break;
	case kStatusQuestionableBit12:
		l_RegSet += ":STAT:QUES:BIT12";
		break;
	default:
		MTL_Assert(false);
		break;
	}

	try {
		// Read current transition filter
		switch (Transition)
		{
		case kPositive:
			l_Command = l_RegSet + ":PTR?";
			break;
		case kNegative:
			l_Command = l_RegSet + ":NTR?";
			break;
		default:
			MTL_Assert(false);
			break;
		}
		if (!l_WriteAndRead(l_Command, m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		U16 l_TransFilter = std::stoi(std::string(m_ReadBuffer.begin(), m_ReadBuffer.end()));

		// Compute new transition filter
		l_TransFilter |= SetMask;
		l_TransFilter &= ~ClearMask;

		// Set new transition filter
		switch (Transition)
		{
		case kPositive:
			l_Command = l_RegSet + ":PTR";
			break;
		case kNegative:
			l_Command = l_RegSet + ":NTR";
			break;
		default:
			MTL_Assert(false);
			break;
		}
		l_Command += " " + std::to_string(l_TransFilter);
		if (!l_WriteAndRead(l_Command, m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	} // try
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}

bool CPT2026Instrument::StandardStatusSetEnableRegister(eStandardStatusRegister Set, U16 DisableMask, U16 EnableMask)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command;

	switch (Set)
	{
	case kStandardEventStatusRegister:
		l_Command += "*ESE";
		break;
	case kServiceRequestRegister:
		l_Command += "*SRE";
		break;
	default:
		MTL_Assert(false);
		break;
	}

	try {
		// Read current enable register
		if (!l_WriteAndRead(l_Command + "?", m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		U16 l_StatusEnab = std::stoi(std::string(m_ReadBuffer.begin(), m_ReadBuffer.end()));

		// Compute new enable register
		l_StatusEnab |= EnableMask;
		l_StatusEnab &= ~DisableMask;

		// Set new enable register
		l_Command += " " + std::to_string(l_StatusEnab);
		if (!l_WriteAndRead(l_Command, m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);
	} // try
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}

//----------------------------------------------------------------------//
//								Utilities								//
//----------------------------------------------------------------------//
bool CPT2026Instrument::GetIdentification(std::string & rIdentification)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead("*IDN?", m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		rIdentification = std::string(m_ReadBuffer.begin(), m_ReadBuffer.end());
	} // try
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		rIdentification.clear();
		return false;
	}
	return true;
}
bool CPT2026Instrument::PowerOff(bool Reboot)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command(":SYST:POFF ");
	l_Command += (Reboot ? "1" : "0");

	return l_WriteAndRead(l_Command, m_ReadBuffer);
}
bool CPT2026Instrument::Reset()
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	return Write("*RST");
}
bool CPT2026Instrument::ReadCaseTemperature(F32 & rTemperatureDegC)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(":SYST:TEMP?", m_ReadBuffer, true))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		rTemperatureDegC = std::stof(std::string(m_ReadBuffer.begin(), m_ReadBuffer.end()));
	} // try
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		rTemperatureDegC = SPqNaN;
		return false;
	}
	return true;
}
bool CPT2026Instrument::SwitchToDFUMode()
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	return Write(":DIAG:UPGR");
}
bool CPT2026Instrument::ReadDebugLog(std::string & rDebugLog)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	rDebugLog.clear();
	try {
		if (!l_WriteAndRead(":DIAG:LOG?", m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		rDebugLog = std::string(m_ReadBuffer.begin(), m_ReadBuffer.end());
	} // try
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::ReadInformationDates(std::string & rSManufacturingDate, std::time_t & rManufacturingDate, std::string & rSCalibrationDate, std::time_t & rCalibrationDate)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	try {
		if (!l_WriteAndRead(":SYST:MDAT?;CDAT?", m_ReadBuffer))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		// Parse results
		CVISABufferParser l_BP(m_ReadBuffer.begin(), m_ReadBuffer.end());
		std::vector<char>::const_iterator l_tokenbeg, l_tokenend;

		// Get Manufacturing Date (UTC)
		if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
			throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
		rSManufacturingDate = std::string(l_tokenbeg, l_tokenend);
		std::match_results<std::string::const_iterator> l_Mm;
		// Match 3 groups of 2 digits
		if (!std::regex_match(rSManufacturingDate, l_Mm, std::regex("^([0-9]{2})([0-9]{2})([0-9]{2})")))
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		std::stringstream l_DateStream;
		std::chrono::system_clock::time_point l_TimePointUTC;
		// we consider date in years 2000 to 2099
		l_DateStream << "20" + l_Mm[1].str() + "-" + l_Mm[2].str() + "-" + l_Mm[3].str() + " 00:00:00";
		l_DateStream >> date::parse("%F %T", l_TimePointUTC); // parse string (UTC) date to time_point UTC
		rManufacturingDate = std::chrono::system_clock::to_time_t(l_TimePointUTC); // convert to time_t (UTC)

		// Get Calibration Date (UTC)
		if (!l_BP.GetNext(l_tokenbeg, l_tokenend, ';'))
			throw CPT2026InsException("Could not find next token", MTL__LOCATION__);
		rSCalibrationDate = std::string(l_tokenbeg, l_tokenend);
		std::match_results<std::string::const_iterator> l_Cm;
		// Match 3 groups of 2 digits
		if (!std::regex_match(rSCalibrationDate, l_Cm, std::regex("^([0-9]{2})([0-9]{2})([0-9]{2})")))
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		// we consider date in years 2000 to 2099
		l_DateStream.clear();
		std::chrono::system_clock::time_point l_TimePointCalUTC;
		l_DateStream << "20" + l_Cm[1].str() + "-" + l_Cm[2].str() + "-" + l_Cm[3].str() + " 00:00:00";
		l_DateStream >> date::parse("%F %T", l_TimePointCalUTC); // parse string (UTC) date to time_point UTC
		rCalibrationDate = std::chrono::system_clock::to_time_t(l_TimePointCalUTC); // convert to time_t (UTC)
	} // try
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		rSManufacturingDate.clear();
		rManufacturingDate = -1;
		rSCalibrationDate.clear();
		rCalibrationDate = -1;
		return false;
	}
	return true;
}

//----------------------------------------------------------------------//
//							Low Level Utilities							//
//----------------------------------------------------------------------//
bool CPT2026Instrument::DeviceROMRead(const tChannel & rChannel, std::vector<U8> & rROMContent)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	// Convert channel to a string channel list
	tChannelList l_SingleChannelList;
	l_SingleChannelList.push_back(rChannel);
	std::string l_StrChannelList;
	ToStringChannelList(l_SingleChannelList, l_StrChannelList);

	std::string l_Command = ":ROUT:MMEM:DATA? " + l_StrChannelList;

	try {
		if (!l_WriteAndRead(l_Command, m_ReadBuffer, true))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		size_t l_Start, l_Len;
		if (!IsArbitraryBlock(m_ReadBuffer.begin(), m_ReadBuffer.end(), l_Start, l_Len))
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);
		
		rROMContent.assign(m_ReadBuffer.data() + l_Start, m_ReadBuffer.data() + l_Start + l_Len);
	} // try
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::DeviceROMWrite(const tChannel & rChannel, const std::vector<U8> & rROMContent)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	// Convert channel to a string channel list
	tChannelList l_SingleChannelList;
	l_SingleChannelList.push_back(rChannel);
	std::string l_StrChannelList;
	ToStringChannelList(l_SingleChannelList, l_StrChannelList);
	// Convert ROM block to arbitrary block
	std::string l_StrROMContent;
	l_StrROMContent.assign(rROMContent.begin(), rROMContent.end());
	std::string l_StrROMArbBlock;
	ToArbitraryBlock(l_StrROMContent, l_StrROMArbBlock);

	std::string l_Command = ":ROUT:MMEM:DATA " + l_StrChannelList + "," + l_StrROMArbBlock;

	return l_WriteAndRead(l_Command, m_ReadBuffer, true);
}
bool CPT2026Instrument::NormalizationGet(const MTL::SCPI::tChannel & rChannel, tNormalizationTable & rNormalizationTable)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	// Convert channel to a string channel list
	tChannelList l_SingleChannelList;
	l_SingleChannelList.push_back(rChannel);
	std::string l_StrChannelList;
	ToStringChannelList(l_SingleChannelList, l_StrChannelList);

	std::string l_Command = ":ROUT:MMEM:NORM? " + l_StrChannelList;

	try {
		if (!l_WriteAndRead(l_Command, m_ReadBuffer, true))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		size_t l_Start, l_Len;
		if (!IsArbitraryBlock(m_ReadBuffer.begin(), m_ReadBuffer.end(), l_Start, l_Len))
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);

		for (auto l_it = m_ReadBuffer.begin() + l_Start; l_it < m_ReadBuffer.begin() + l_Start + l_Len; l_it += sizeof(tNormalizationFactor_Hz))
			rNormalizationTable.push_back(BinaryToF32(&*l_it));
	} // try
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		return false;
	}
	return true;
}
bool CPT2026Instrument::NormalizationSet(const MTL::SCPI::tChannel & rChannel, const tNormalizationTable & rNormalizationTable)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	// Convert channel to a string channel list
	tChannelList l_SingleChannelList;
	l_SingleChannelList.push_back(rChannel);
	std::string l_StrChannelList;
	ToStringChannelList(l_SingleChannelList, l_StrChannelList);
	// Convert Normalization table to arbitrary block
	std::vector<U8> l_RawNormalizationTable;
	l_RawNormalizationTable.resize(rNormalizationTable.size() * sizeof(tNormalizationFactor_Hz));
	std::memcpy(l_RawNormalizationTable.data(), rNormalizationTable.data(), l_RawNormalizationTable.size());
	std::string l_StrNormTableArbBlock;
	ToArbitraryBlock(std::string(l_RawNormalizationTable.begin(), l_RawNormalizationTable.end()), l_StrNormTableArbBlock);

	std::string l_Command = ":ROUT:MMEM:NORM " + l_StrChannelList + "," + l_StrNormTableArbBlock;

	return l_WriteAndRead(l_Command, m_ReadBuffer, true);
}
bool CPT2026Instrument::LowLevelTransfer(const std::vector<U8> & rDataToSend, std::vector<U8> & rRcvdData, U8 & rReturnedStatus)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	std::string l_Command = ":ROUT:LOWL? ";
	std::string l_DataToSend;
	l_DataToSend.assign(rDataToSend.begin(), rDataToSend.end());
	l_Command += l_DataToSend;

	try {
		if (!l_WriteAndRead(l_Command, m_ReadBuffer, true))
			throw CPT2026InsException(std::string("Failed W&R in ") + MTL__FUNCTION_NAME__, MTL__LOCATION__);

		size_t l_Start, l_Len;
		if (!IsArbitraryBlock(m_ReadBuffer.begin(), m_ReadBuffer.end(), l_Start, l_Len))
			throw CPT2026InsException("Invalid answer format", MTL__LOCATION__);

		rReturnedStatus = *(m_ReadBuffer.data() + l_Start);
		rRcvdData.assign(m_ReadBuffer.data() + l_Start + 1, m_ReadBuffer.data() + l_Start + l_Len);
	} // try
	catch (CPT2026InsException & rE)
	{
		MTL_INSTRUMENT_PT2026_DEBUG_CERR(rE.what());
		MTL_Unused(rE);
		rReturnedStatus = 0;
		rRcvdData.clear();
		return false;
	}
	return true;
}
bool CPT2026Instrument::WriteCustomCommand(const std::string & rWriteStr)
{
	MTL_INSTRUMENT_PT2026_DEBUG_COUT(MTL__FUNCTION_NAME__ << std::endl);
	CLockGuard<CRecursiveMutex> l_LockGuard(m_Lock);

	return l_WriteAndRead(rWriteStr, m_ReadBuffer, true);
}
