#pragma once

#include <iterator>

#include "VISAInstrumentTypes.h"
#include "Synchronization.h"

using namespace MTL::Synchronization;

namespace MTL {
	namespace Instrument {

	//----------------------------------------------------------------------//
	//								VISA Buffer								//
	//----------------------------------------------------------------------//
	// This class is intended to provide an efficient way of reading data
	// through VISA and easy aswer parsing.
	// VISA Read requires an allocated buffer pointer and its length.
	// Parsing methods (regex, find etc...) require std::string object or
	// iterators.
	//  - std::string doesn't ensure to store data contiguously: cannot write
	// to it through a pointer.
	//  - std::vector ensures contiguous storage and it is allowed to write to
	// its allocated storage (capacity) through a pointer. But it is
	// impossible to set its size. It is thus not possible to indicate how
	// many bytes have been stored into the vector after a read.
	//
	// This class holds data into an std::vector. Storage allocation in thus
	// stored into contiguous memory and memory allocation is auto handled.
	// A size member is managed by this class itself and can be set by a
	// method. We must ensure that this class size is always lower or equal
	// to the contained vector size (the class capacity is equivalent to the
	// contained vector size).
	//----------------------------------------------------------------------//
	//#define MTL_VISA_BUFFER_TYPE		ViByte
	#define MTL_VISA_BUFFER_TYPE		char
	class CVISABuffer
	{
	private:
		mutable CRecursiveMutex				m_Lock;
		std::vector<MTL_VISA_BUFFER_TYPE>	m_Vector;	// Used to store the data contiguously
		size_t								m_Size;		// Used to handle the real size
	public:
		CVISABuffer(size_t InitialCapacity = 2048)
			: m_Vector(), m_Size(0)
		{
			CLockGuard<CRecursiveMutex> Lock(m_Lock);

			reserve(InitialCapacity);
		}
		CVISABuffer(const MTL_VISA_BUFFER_TYPE * pData, size_t DataLen)
			: m_Vector(), m_Size(0)
		{
			CLockGuard<CRecursiveMutex> Lock(m_Lock);

			assign(pData, pData + DataLen);
		}
		virtual ~CVISABuffer()
		{}

		void clear()
		{
			CLockGuard<CRecursiveMutex> Lock(m_Lock);

			m_Size = 0;
		}
		void reserve(size_t capacity)
		{
			CLockGuard<CRecursiveMutex> Lock(m_Lock);

			if (capacity > m_Vector.size())
				m_Vector.resize(capacity);
		}
		void resize(size_t size)
		{
			CLockGuard<CRecursiveMutex> Lock(m_Lock);

			if (size > m_Vector.size())
			{
				// Ensure the vector can contain at least the required size
				m_Vector.resize(size);
			}
			m_Size = size;
		}

		size_t size() const
		{
			CLockGuard<CRecursiveMutex> Lock(m_Lock);

			return m_Size;
		}
		size_t capacity() const
		{
			CLockGuard<CRecursiveMutex> Lock(m_Lock);

			return m_Vector.size();
		}

		std::vector<MTL_VISA_BUFFER_TYPE>::iterator begin()
		{
			CLockGuard<CRecursiveMutex> Lock(m_Lock);

			return m_Vector.begin();
		}
		std::vector<MTL_VISA_BUFFER_TYPE>::const_iterator begin() const
		{
			CLockGuard<CRecursiveMutex> Lock(m_Lock);

			return m_Vector.begin();
		}
		std::vector<MTL_VISA_BUFFER_TYPE>::iterator end()
		{
			CLockGuard<CRecursiveMutex> Lock(m_Lock);

			return m_Vector.begin() + m_Size;
		}
		std::vector<MTL_VISA_BUFFER_TYPE>::const_iterator end() const
		{
			CLockGuard<CRecursiveMutex> Lock(m_Lock);

			return m_Vector.begin() + m_Size;
		}

		MTL_VISA_BUFFER_TYPE* data() noexcept
		{
			CLockGuard<CRecursiveMutex> Lock(m_Lock);

			return m_Vector.data();
		}
		const MTL_VISA_BUFFER_TYPE* data() const noexcept
		{
			CLockGuard<CRecursiveMutex> Lock(m_Lock);

			return m_Vector.data();
		}
		MTL_VISA_BUFFER_TYPE& operator[] (size_t n)
		{
			CLockGuard<CRecursiveMutex> Lock(m_Lock);

			return m_Vector[n];
		}
		const MTL_VISA_BUFFER_TYPE& operator[] (size_t n) const
		{
			CLockGuard<CRecursiveMutex> Lock(m_Lock);

			return m_Vector[n];
		}
		void assign(const MTL_VISA_BUFFER_TYPE* First, const MTL_VISA_BUFFER_TYPE* Last)
		{
			CLockGuard<CRecursiveMutex> Lock(m_Lock);

			if (capacity() < (size_t)(Last - First))
				reserve(Last - First);
			std::copy(First, Last, m_Vector.begin());
			m_Size = Last - First;
		}
		void assign(const MTL_VISA_BUFFER_TYPE* pData, size_t n)
		{
			CLockGuard<CRecursiveMutex> Lock(m_Lock);

			if (capacity() < n)
				reserve(n);
			std::copy(pData, pData + n, m_Vector.begin());
			m_Size = n;
		}
	};

	//----------------------------------------------------------------------//
	//								Buffer Parser							//
	//----------------------------------------------------------------------//
	class CVISABufferParser
	{
	private:
		std::vector<char>::const_iterator	m_Bbegin;
		std::vector<char>::const_iterator	m_Bend;
		std::vector<char>::const_iterator	m_SplitOffset;
	public:
		struct sToken {
			std::vector<char>::const_iterator	begin;
			std::vector<char>::const_iterator	end;
			sToken(std::vector<char>::const_iterator b, std::vector<char>::const_iterator e)
				: begin(b), end(e) {}
		};
		typedef std::vector<sToken>				tTokens;

	public:
		CVISABufferParser(std::vector<char>::const_iterator begin, std::vector<char>::const_iterator end)
			: m_Bbegin(begin), m_Bend(end), m_SplitOffset(begin) {}

		std::vector<char>::const_iterator begin()
		{
			return m_Bbegin;
		}
		std::vector<char>::const_iterator end()
		{
			return m_Bend;
		}

		const tTokens Tokenize(const char Separator = ';', size_t Offset = 0)
		{
			tTokens l_Tokens;

			// Check if there is data to split
			if (m_Bbegin == m_Bend)
				return l_Tokens;

			// Find first separator (if any).
			std::vector<char>::const_iterator l_Start = m_Bbegin + Offset;
			std::vector<char>::const_iterator l_Next = std::find(l_Start, m_Bend, Separator);
			l_Tokens.push_back(sToken(l_Start, l_Next));

			// Find all separators
			while (l_Next != m_Bend)
			{
				l_Start = l_Next + 1;
				l_Next = std::find(l_Start, m_Bend, Separator);
				l_Tokens.push_back(sToken(l_Start, l_Next));
			}
			
			// Strip white space off front.
			const std::string l_Whitespace(" \t\n\r");
			for (auto l_Token : l_Tokens)
			{
				while (std::string::npos != l_Whitespace.find(*l_Token.begin))
					l_Token.begin++;
			}
			
			return l_Tokens;
		}

		// /!\ When next token is an arbitrary block, it may contain a ';', which causes rNextEnd to be wrong
		bool GetNext(std::vector<char>::const_iterator & rNextBegin, std::vector<char>::const_iterator & rNextEnd, const char Separator = ';')
		{
			// Strip white space off front.
			const std::string l_Whitespace(" \t\n\r");
			while (m_SplitOffset < m_Bend &&
				   std::string::npos != l_Whitespace.find(*m_SplitOffset))
				m_SplitOffset++;
			
			// Did we reach the last token?
			if (m_SplitOffset == m_Bend)
			{
				rNextBegin = m_Bend;
				rNextEnd = m_Bend;
				return false;
			}
			
			// Find next
			rNextBegin = m_SplitOffset;
			rNextEnd = std::find(m_SplitOffset, m_Bend, Separator);
			
			// Did we reach the last token?
			if (rNextEnd == m_Bend)
				m_SplitOffset = m_Bend;			// Set offset to the ending interator
			else
				m_SplitOffset = rNextEnd + 1;	// Skip separator for next call

			return true;
		}

		// When parsing arbitrary block, it is necessary to force offset to skip ';' chars that are not separators
		void SetNextOffset(std::vector<char>::const_iterator Offset)
		{
			m_SplitOffset = Offset;
			// Check bounds
			if (m_SplitOffset < m_Bbegin)
				m_SplitOffset = m_Bbegin;
			if (m_SplitOffset > m_Bend)
				m_SplitOffset = m_Bbegin;
		}
	};

}}	// namespace MTL::Instrument
